#include "adaptiveRule.h"
#include <stdio.h>
#include <strings.h>
#include "attributesInfo.h"
#include "timerADI.h"
#include "timerGlobals.h"
#include "timerMutation.h"
#include "timerMDL.h"
#include "random.h"
#include "instanceSet.h"
#include "sampling.h"

extern attributesInfo ai;
extern instanceSet *is;
extern timerADI *tADI;
extern timerGlobals *tGlobals;
extern timerMDL *tMDL;
extern timerMutation *tMut;
extern Random rnd;

#define max(a,b) (a)>(b)?(a):(b)

void adaptiveRuleConstructor(int *rule)
{
	int i;
	int *ptr = &rule[2];

	rule[0] = 1;
	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		adaptiveAttributeConstructor(ptr, i);
		rule[0] += ptr[0];
		ptr += tADI->sizeAdaptiveAttribute[i];
	}
}

double adaptiveRuleTheoryLength(int *rule)
{
	int i, j, k;
	int *ptr = &rule[2];
	double length = 0;
	int numAttr = tGlobals->numAttributesMC;

	for (i = 0; i < numAttr; i++) {
		if (tADI->typeOfAttribute[i] == REAL) {
			double intervalCount = 0;
			int previousValue = ptr[3];
			int numInt = ptr[0];
			if (previousValue)
				intervalCount++;
			for (j = 5, k = 1; k < numInt; k++, j += 2) {
				if (ptr[j] != previousValue)
					intervalCount++;
				previousValue = ptr[j];
			}
			if (!previousValue)
				intervalCount--;
			length += numInt + intervalCount;
		} else {
			double countFalses = 0;
			int numValues = tADI->sizeAdaptiveAttribute[i];
			for (j = 2; j < numValues; j++) {
				if (!ptr[j])
					countFalses++;
			}
			length += numValues - 2.0 + countFalses;
		}
		ptr += tADI->sizeAdaptiveAttribute[i];
	}

	return length;
}

void adaptiveRuleMutation(int *rule, int defaultClass)
{
	if (!tGlobals->smartClassAssignment && tGlobals->numClasses > 1 && !rnd < 0.10) {
		int newClass;
		do {
			newClass = rnd(0, ai.getNumClasses()-1);
		} while (newClass == rule[1] || tGlobals->defaultClassPolicy!=DISABLED && newClass==defaultClass);
		rule[1] = newClass;
	} else {
		int attribute = rnd(0, tGlobals->numAttributesMC - 1);
		int *ptr =
		    &rule[tADI->offsetAdaptiveAttribute[attribute]];
		adaptiveAttributeMutation(ptr, attribute);
	}
}

void adaptiveRuleCrossover(int *in1, int *in2, int *out1, int *out2,
			   int cutPoint)
{
	int i;

	int *pOrig1 = in1 + 2;
	int *pDest1 = out1 + 2;
	int *pOrig2 = in2 + 2;
	int *pDest2 = out2 + 2;
	// Lengths of output rules 
	out1[0] = 1;
	out2[0] = 1;

	for (i = 0; i < cutPoint && i < tGlobals->numAttributesMC; i++) {
		int inc = tADI->sizeAdaptiveAttribute[i];
		bcopy(pOrig1, pDest1, sizeof(int) * inc);
		bcopy(pOrig2, pDest2, sizeof(int) * inc);
		out1[0] += pOrig1[0];
		out2[0] += pOrig2[0];
		pOrig1 += inc;
		pOrig2 += inc;
		pDest1 += inc;
		pDest2 += inc;
	}

	for (i = cutPoint; i < tGlobals->numAttributesMC; i++) {
		int inc = tADI->sizeAdaptiveAttribute[i];
		bcopy(pOrig1, pDest2, sizeof(int) * inc);
		bcopy(pOrig2, pDest1, sizeof(int) * inc);
		out1[0] += pOrig2[0];
		out2[0] += pOrig1[0];
		pOrig1 += inc;
		pOrig2 += inc;
		pDest1 += inc;
		pDest2 += inc;
	}

	if (cutPoint == tGlobals->numAttributes) {
		out1[1] = in1[1];
		out2[1] = in2[1];
	} else {
		out1[1] = in2[1];
		out2[1] = in1[1];
	}
}

void adaptiveRuleInitialize(int *rule, int defaultClass)
{
	instance *ins = NULL;

	if (tGlobals->smartInit) {
		if (tGlobals->defaultClassPolicy != DISABLED) {
			ins = is->getInstanceInit(defaultClass);
		} else {
			ins = is->getInstanceInit(ai.getNumClasses());
		}
	}

	int i;
	int *ptr = &rule[2];

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		adaptiveAttributeInitializeIntervalValues(ptr, i, ins);
		ptr += tADI->sizeAdaptiveAttribute[i];
	}

	if (ins) {
		rule[1] = ins->getClass();
	} else {
		int cl;
		do {
			cl=rnd(0,ai.getNumClasses()-1);
		} while(tGlobals->defaultClassPolicy!=DISABLED && cl==defaultClass);
		rule[1] = cl;
	}
}

void adaptiveRuleDumpPhenotype(int *rule, char *string, int defaultClass)
{
	int *ptr = &rule[2];
	char temp[10000];
	int i;

	strcpy(string, "");
	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		adaptiveAttributeDumpPhenotype(ptr, i, temp);
		if (strlen(temp)) {
			strcat(string, temp);
			strcat(string, "|");
		}
		ptr += tADI->sizeAdaptiveAttribute[i];
	}
	int cls = rule[1];
	sprintf(temp, "%s\n",
		ai.getNominalValue(tGlobals->numAttributesMC, cls)->cstr());
	strcat(string, temp);
}

int adaptiveRuleDoMerge(int *rule)
{
	int i;
	int *ptr = &rule[2];
	int modif = 0;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		if (!rnd < tADI->probMergeInterval) {
			modif = 1;
			int pos = rnd(0, ptr[0] - 1);
			rule[0] -= ptr[0];
			adaptiveAttributeDoMerge(ptr, i, pos);
			rule[0] += ptr[0];
		}
		ptr += tADI->sizeAdaptiveAttribute[i];
	}
	return modif;

}

int adaptiveRuleDoSplit(int *rule)
{
	int i;
	int *ptr = &rule[2];
	int modif = 0;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		if (!rnd < tADI->probSplitInterval) {
			modif = 1;
			int pos = rnd(0, ptr[0] - 1);
			rule[0] -= ptr[0];
			adaptiveAttributeDoSplit(ptr, i, pos);
			rule[0] += ptr[0];
		}
		ptr += tADI->sizeAdaptiveAttribute[i];
	}
	return modif;
}

int adaptiveRuleDoReinitialize(int *rule)
{
	int i;
	int *ptr = &rule[2];
	int modif = 0;

	if (!tADI->probReinitializeAttribute)
		return 0;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		if (!rnd < tADI->probReinitializeAttribute) {
			modif = 1;
			rule[0] -= ptr[0];
			adaptiveAttributeDoReinitialize(ptr, i);
			rule[0] += ptr[0];
		}
		ptr += tADI->sizeAdaptiveAttribute[i];
	}
	return modif;
}

int adaptiveRuleCountDiscretizers(int *rule, int attribute, int *counters)
{
	int *ptr = &rule[tADI->offsetAdaptiveAttribute[attribute]];
	int count = 0;

	if (ai.getTypeOfAttribute(attribute) == REAL) {
		count++;
		counters[ptr[1]]++;
	}
	return count;
}

void adaptiveRuleCover(int *rule, instance *ins)
{
	int i;
	int *ptr = &rule[2];

	adaptiveRuleConstructor(rule);
	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		adaptiveAttributeInitializeIntervalValues(ptr, i, ins);
		ptr += tADI->sizeAdaptiveAttribute[i];
	}
	rule[1] = ins->getClass();
}


matchProfileAgent *evaluateRuleCX(int *rule)
{
	unsigned long long i;
	int cl = rule[1];
	unsigned long long max = is->getNumInstancesOfIteration();
	instance **instances=is->getInstancesOfIteration();

	matchProfileAgent *mpa=new matchProfileAgent(max,cl);

	for (i = 0; i < max; i++) {
		if (adaptiveRuleMatches(rule, instances[i])) {
			if (instances[i]->getClass() == cl) {
				mpa->addOK(i);
			} else {
				mpa->addKO(i);
			}
		}
	}
	mpa->generateProfiles();

	return mpa;
}

void adaptiveRuleDisable(int *rule) 
{
	int attribute = rnd(0, tGlobals->numAttributesMC - 1);
	int *ptr = &rule[tADI->offsetAdaptiveAttribute[attribute]];
	adaptiveAttributeDisable(ptr,attribute);
}

