#ifndef _ATTRIBUTES_INFO_H_
#define _ATTRIBUTES_INFO_H_

#include "JString.h"
#include "JVector.h"
#include "instance.h"

#define NOMINAL 1
#define ENTER 2
#define REAL 3

class attributesInfo {
	int numAttributes;
	int numAttributesMC;
	int *typeOfAttributes;
	double *minDomain,*maxDomain,*sizeDomain;
	double **averageOfAttribute;
	double **deviationOfAttribute;
	int **countNumValuesForRealAttributes;
	int *classOfInstances;
	double ***valueFrequenciesForNominalAttributes;
	int **mostFrequentValueForNominalAttributes;
	JVector<JString *> *valuesOfNominalAttributes;
	JVector<JString *> attributeNames;
	int numExamples;
	int mostFrequentClass;
	int leastFrequentClass;
	int thereAreNominal;
	int thereAreRealValued;
	int allBinary;
	int recodeBlocks;

public:
	attributesInfo();
	void setNumAttributes(int num);
	inline int getNumAttributes() {return numAttributes;}
	inline int getNumAttributesMC() {return numAttributesMC;}
	void setTypeOfAttribute(int attribute,int type);
	void insertNominalValue(int attribute,JString *value);
	JString *getNominalValue(int attribute,int value);
	
	inline void insertAttributeName(JString *name) {
		attributeNames.addElement(name);
	}

	inline JString *getAttributeName(int attr) {
		return attributeNames.elementAt(attr);
	}

	inline void finish() {
		if(allBinary) {
			recodeBlocks=(4*(2*numAttributesMC/(8*4*sizeof(unsigned int))+1));
			//recodeBlocks=(2*numAttributesMC/(8*sizeof(unsigned int))+1);		
		}
	}
	inline int getRecodeBlocks(){return recodeBlocks;}


	int getNumValuesAttribute(int attribute);
	inline int getInstancesOfClass(int pClass) {return classOfInstances[pClass];}
	inline int thereAreNominalAttributes() {return thereAreNominal;}
	inline int thereAreRealValuedAttributes() {return thereAreRealValued;}

	inline int allAttributesBinary() {return allBinary;}


	

	inline int getMostFrequentClass() {return mostFrequentClass;}
	inline int getLeastFrequentClass() {return leastFrequentClass;}
	
	inline int getNumClasses() 
	{
		return valuesOfNominalAttributes[numAttributes-1].size();
	}

	void insertInstance(instance *i);

	inline int getTypeOfAttribute(int attribute)
	{
	#ifdef DEBUG
		if(attribute<0 || attribute>=numAttributes ) {
			fprintf(stderr,"Incorrect values at attributesInfo::getTypeOfAttribute %d\n",attribute);
			exit(1);
		}
	#endif
		return typeOfAttributes[attribute];
	}


	void calculateAverages();
	double getAverageOfAttribute(int whichClass,int attribute);
	double getDeviationOfAttribute(int whichClass,int attribute);
	int getMostFrequentValueOfAttribute(int whichClass,int attribute);
	double getFrequencyOfValueOfAttribute(int whichClass,int attribute,int value);
	void setBounds(double *min,double *max);
	inline double getMinDomain(int attribute) {return minDomain[attribute];}
	inline double getMaxDomain(int attribute) {return maxDomain[attribute];}
	inline double getSizeDomain(int attribute) {return sizeDomain[attribute];}

	int valueOfNominalAttribute(int attribute,char *def);
};

#endif
