/************************************************************************
 * ChiMerge Discretizer					    		*
 *						 Pere Miralles Ribas	*
 ************************************************************************/

#include "chimergeDiscretizer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "messageBuffer.h"

extern messageBuffer mb;

void chimergeDiscretizer::results()
{
	double numpttotals=0;
	double numAttr=0;
	for(int i=0;i<numatributs;i++)
	{ // Per a tots els atributs
		if(tipusatributs[i]==3 || tipusatributs[i]==2)
		{
			numAttr++;
			numpttotals=numpttotals+numCutPoints[i];
			for(int j=0;j<numCutPoints[i];j++) {
				mb.printf("Cut point %d of attribute %d : %f\n",j,i,cutPoints[i][j]);
			}
			mb.printf("Number of cut points of attribute %d : %d\n",i,numCutPoints[i]);
		}
	}
	mb.printf("Numero de punts de tall Discretitzacio : %g\n",numpttotals/numAttr);
}


double chimergeDiscretizer::calcullimitxi2(double siglevel,int grausllib)
{ // Retorna el valor de la taula xi2 (la llegeix de fitxer)
	 FILE *fp;
	 int files,columnes;
	 double *sl;
	 int *gll;
	 double **taula;
	 float tmp;
	 int x, y;
	 char fileName[256];
	
	sprintf(fileName, "%s/doctorat/proves/taulaxi2.txt",getenv("HOME"));
        fp = fopen(fileName, "r");
						 
	 if (!fp)
	 {
		 fprintf(stderr, "Can't open taulaxi2.txt\n");
		 exit(1);
	 }
	fscanf(fp,"%d",&columnes);
	fscanf(fp,"%d",&files);
	demanaMatriuDouble(taula,files,columnes);
	demanaVectorInt(gll,files);
	demanaVectorDouble(sl,columnes);
	for(int i=0;i<columnes;i++)
	{  
		fscanf(fp,"%f",&tmp);
		sl[i]=(double)tmp;
	}
	y=0;
	while(y<files)
	{
		fscanf(fp,"%d",&gll[y]);
		for(int x=0;x<columnes;x++)
		{
			fscanf(fp,"%f",&tmp);
			taula[y][x]=(double)tmp;
		}
		y++;		
	}
	
	x=0;
	while(x<columnes && siglevel!=sl[x]) x++;
	y=0;
	while(y<files && grausllib!=gll[y]) y++;
	
	if(x>=columnes || y>=files)
	{
		fprintf(stderr, "Valor no trobat a la taula\n");
		exit(1);
	}
	return taula[y][x];
}

void chimergeDiscretizer::inicialitzapt(double *ptpos,int *espt,int atribut,int limit)
{
	for(int i=0;i<limit;i++)
	{
		ptpos[i]=(dadesc[atribut][i]+dadesc[atribut][i+1])/2;
		espt[i]=1; // primer tots els possibels pt son considertats
	}
}

double chimergeDiscretizer::calculxi2(int inici,int final,double pt,int atr)
{
	// Calcul de Xi²
	// Xi² = SUM(j=1,p, (A1j-E1j)²/ E1j) + SUM(j=1,P, (A2j-E2j)²/ E2j) 
		
	int P,A,R,C,N,pivot;
	double E;
	double Xi2;
	int limitinf,limitsup;

	Xi2=0;
	// Calcul de P 
	P=numclasses;
	// Calcul de N
	N=final-inici+1;
	// Calcul de pivot
	pivot=inici;
	while(pivot<final && dades[atr][pivot]<pt) //
		pivot++;
	pivot--;
	// Intervals entre [inici..pivot] i [pivot+1,final]
	
	for(int particio=1;particio<=2;particio++)
	{
		switch(particio)
		{
			case 1:
				limitinf=inici;
				limitsup=pivot;
				break;
			case 2:
				limitinf=pivot+1;
				limitsup=final;
				break;
		}
	
		for(int classe=0;classe<P;classe++)
		{
			// Calcul de A
			A=0;
			for(int j=limitinf;j<=limitsup;j++)
			{
				if(!strcmp(nomclasses[classe],classes[atr][j]))
					A++;
			}
			// Calcul de R
			R=limitsup-limitinf+1;
			//Calcul de C
			C=0;
			for(int j=inici;j<=final;j++)
			{
				if(!strcmp(nomclasses[classe],classes[atr][j]))
					C++;
			}
			//Calcul de E
			if(R<=0 || C<=0){
				E=1.0/10.0;		
			}
			else
			{
				E=((double)R*(double)C)/(double)N;
			}
			Xi2=Xi2+(pow(A-E,2)/E);
		}
	}

	return Xi2;
}
int chimergeDiscretizer::aturada(int atr,double *ptpos,int *espt,double *xi2pt,int limit)
{ 
	int numpt;
	int totsmenorsxi2;
	int algunmenorxi2;

	numpt=0;
	// Calcul dels pt existents
	for(int i=0;i<limit;i++)
		if(espt[i])
			numpt++;
	
	totsmenorsxi2=1;
	algunmenorxi2=0;
	for(int i=0;i<limit;i++)
	{
		if(espt[i] && xi2pt[i]>xi2limit)
			totsmenorsxi2=0;
		if(espt[i] && xi2pt[i]<=xi2limit)
			algunmenorxi2=1;
	}
	
	if(numpt>10) // Num de talls >10
	{ // No hem arribat a MAX_INTERVALS (seguim)
		return 0;
	}
	else 
	{ // Num talls <= MAX_INTERVALS (podem acabar o seguir)
		if(totsmenorsxi2)
		{ // Tots pt menors que xi2limit (acabem)
			return 1;
		}
		else 
		{
			if(algunmenorxi2)
			{ // Algun pt amb xi2 menor que xi2limit (seguim)
				return 0;
			}
			else
			{ // Cap pt amb xi2 menor que xi2limit (acabem)
				return 1;
			}
		}		
	}
}

void chimergeDiscretizer::discretitzador()
{
	double *ptpos; 	// Possible pt
	int *espt;	// Indica si el possible pt es pt (1) o no (0)
	double *xi2pt;	// Valor de la xi2 associat al pt
	int limit;
	int inici,fi;
	int k;
	double minim;
	int posminim;
	
	grausllib=numclasses-1;
	xi2limit=calcullimitxi2(siglevel,grausllib);
	
	for(int i=0;i<numatributs;i++)
	{ // Per a tots els atributs
		if(tipusatributs[i]==3 || tipusatributs[i]==2)
		{ // Nomes per als atributs numerics
			limit=contavalors(i);
			demanaVectorDouble(ptpos,limit-1);
			demanaVectorInt(espt,limit-1);
			demanaVectorDouble(xi2pt,limit-1);
			// Entre tots els valors reals pot existit un punt de tall (valreals-1)
			inicialitzapt(ptpos,espt,i,limit-1);
			do{ // Bucle mentre no condició d'aturada
				for(int j=0;j<limit-1;j++)
				{ // Evaluem el possibles punts de tall
					if(espt[j])
					{ // Nomes aquells possibles
						k=j-1;
						while(k>=0 && !espt[k]) k--;
					
						if(k<0)
							inici=0;
						else
						{
							inici=numdades-1;
					     	  	while(inici>0 && dades[i][inici]>ptpos[k]) inici--;
					        	inici++;
						}
						k=j+1;
						while(k<limit-1 && !espt[k]) k++;
						if(k==limit-1)
							fi=numdades-1;
						else
						{
							fi=0;
							while(fi<numdades && dades[i][fi]<ptpos[k]) 
								fi++;
							fi--;						
						}
						xi2pt[j]=calculxi2(inici,fi,ptpos[j],i);
					}	
				}		
				minim=MAX_DOUBLE;	
				for(int j=0;j<limit-1;j++)
				{
					if(espt[j] && xi2pt[j]<minim)
					{
						minim=xi2pt[j];
						posminim=j;
					}	
				}
				espt[posminim]=0;
			}while(!aturada(i,ptpos,espt,xi2pt,limit-1));
								
			int numpt=0;
			for(int j=0;j<limit-1;j++)
			{
				if(espt[j])
					numpt++;
			}
			numCutPoints[i]=numpt;
			demanaVectorDouble(cutPoints[i],numpt);
			numpt=0;
			
			for(int j=0;j<limit-1;j++)
			{
				if(espt[j])
				{
					cutPoints[i][numpt]=ptpos[j];
					numpt++;
				}				
			}		
		}
	}
}

chimergeDiscretizer::chimergeDiscretizer(double sig, int attributes)
{
	siglevel=sig;						// Nivell de significació		
	numAttributes = attributes;                             // Nombre d'atributs
	demanaVectorInt(numCutPoints,attributes);               // Numero de punts de tall per atribut
	cutPoints = new double *[numAttributes];                // Punts de tall per atribut
	demanaVectorDouble(minDomain,numAttributes);		// Valor mínim
	demanaVectorDouble(maxDomain,numAttributes);		// Valor màxim
}

void chimergeDiscretizer::setBounds(double *min, double *max)
{
	int i;
	for (i = 0; i < numAttributes; i++)
	{
		minDomain[i] = min[i];
		maxDomain[i] = max[i];
	}
}

void chimergeDiscretizer::updateDiscretizer(char *dataFile)
{
	parserFitxer(dataFile);
	carregarDades(dataFile);
	compactardades();
	discretitzador();
	results();
							
}
int chimergeDiscretizer::getNumIntervals(int attr)
{
	return numCutPoints[attr]+1;
}

int chimergeDiscretizer::discretize(int attribute, double value)
{
	if (numCutPoints[attribute] == 0)
		return 0;
	for(int i=0;i<numCutPoints[attribute];i++)
	{
		if(value<cutPoints[attribute][i])
			return i;
	}
	return numCutPoints[attribute];
}
