#ifndef _CLASSIFIER_H_
#define _CLASSIFIER_H_

#include "configCodes.h"
#include "attributesInfo.h"
#include "classifierFitness.h"
#include "timerHierar.h"
#include "timerGlobals.h"
#include "timerMDL.h"
#include "random.h"
#include <math.h>

extern attributesInfo ai;
extern timerHierar *tHierar;
extern timerGlobals *tGlobals;
extern timerMDL *tMDL;
extern Random rnd;

class agentPerformanceTraining;

class classifier {
      protected:
	int length;		// Length of the individual in genes
	double fitness;
	double scaledFitness;
	int modif;
	int front;
	double exceptionsLength;	//For MDL fitness function
	double accuracy;
	int numAttributes;
	int numAttributesMC;
	int aliveClassifiers;
	double theoryLength;
	int defaultClass;
	double averageActivation;
	double accDiff;
	double cwAcc;

      public:

	 inline classifier() {
		length = 0;
		modif = 1;
		front = 0;
	} 

	inline ~classifier() {
	}

	inline void fitnessComputation() {
		modif = 0;
		fitness = classifierFitness(*this);
	}
	inline void setScaledFitness(double pFitness) {
		scaledFitness = pFitness;
	}
	inline double getFitness(void) {
		return fitness;
	}
	inline void setFitness(double pFit) {
		fitness=pFit;
	}
	inline double getScaledFitness(void) {
		return scaledFitness;
	}
	inline int getFront() {
		return front;
	}
	inline void setFront(int pFront) {
		front = pFront;
	}
	inline void setAliveClassifiers(int num) {
		aliveClassifiers=num;
	}
	inline int getAliveClassifiers() {
		return aliveClassifiers;
	}
	inline void setAccuracy(double acc) {
		accuracy = acc;
	}
	inline double getAccuracy() {
		return accuracy;
	}
	inline void setCWAcc(double pCWAcc) {
		cwAcc = pCWAcc;
	}
	inline double getCWAcc() {
		return cwAcc;
	}

	inline double getExceptionsLength() {
		return exceptionsLength;
	}
	inline void setExceptionsLength(double excep) {
		exceptionsLength = excep;
	}
	inline int getNumObjectives() {
		return 2;
	}
	inline double getObjective(int objective) {
		if (objective == 0)
			return 1 - accuracy;
		return getNumClassifiers();
	}
	inline int isModified() {
		return modif;
	}
	inline void activateModified() {
		modif = 1;
	}
	inline double getTheoryLength() {
		return theoryLength;
	}

	inline double getAverageActivation() {
		return averageActivation;
	}
	inline void setAverageActivation(double pAct) {
		averageActivation=pAct;
	}

	double getAccDiff(){return accDiff;}
	void setAccDiff(double pAccDiff){accDiff=pAccDiff;}

	
	virtual int getNumClassifiers() = 0;
	virtual int getClass(int classifier) = 0;
	virtual int classify(instance * i) = 0;
	virtual void pruneClassifiers(int *classifiersToPrune,
				      int numPruned) = 0;
	virtual void dumpPhenotype(char *string) = 0;
	virtual void dumpGenotype(char *string) {}
	virtual double computateTheoryLength(agentPerformanceTraining *ap) = 0;
	// Second parent of CX & the two sons
	virtual void crossover(classifier *, classifier *, classifier *) = 0;

	// For simplicity reasons...
	virtual void crossover_smart(classifier **,int num){};
	virtual int hasSmartCrossover(){return 0;}

	virtual void doLocalSearch(){};
	virtual int hasLocalSearch(){return 0;}


	virtual void mutation() = 0;
	virtual int numSpecialStages() = 0;
	virtual void doSpecialStage(int stage) = 0;

	virtual double getHierarchicalSelectionLength() = 0;

	virtual int getNiche() {return 0;}
	virtual int getNumNiches() {return 1;}

	virtual void initEvaluation(){}
	virtual void addOutcome(int instance,int outcome) {}
	virtual void endEvaluation() {}
	virtual void postProcessIndividual() {}
	virtual int classifyClearRules(instance * i){return -1;}

	inline int compareToIndividual(classifier * i2, int maxmin) {
		double l1=getHierarchicalSelectionLength();
		double l2=i2->getHierarchicalSelectionLength();

		if (tHierar->activated) {
			if (fabs(fitness-i2->fitness) <= tHierar->threshold) {
				if(l1<l2) return +69;
				if(l1>l2) return -69;
			}
		}

		if (maxmin == MAXIMIZE) {
			if (fitness>i2->fitness)
				return +69;
			if (fitness<i2->fitness)
				return -69;
			if(!rnd<0.5) return +69;
			return -69;
			//return 0;
		}
		if (fitness<i2->fitness)
			return +69;
		if (fitness>i2->fitness)
			return -69;
		if(!rnd<0.5) return +69;
		return -69;
		//return 0;
	}

	virtual void classifyTraining(instanceSet *is,agentPerformanceTraining *ap){};
	virtual void trainAndClean(instanceSet *is,agentPerformanceTraining *ap){};
	virtual int canDoTrainAndClean() {return 0;}
	virtual int canDoSmartClassAsignment() {return 0;}
};

#endif
