#include <math.h>
#include <stdio.h>
#include <string.h>

#include "classifier.h"
#include "attributesInfo.h"
#include "timerPruning.h"
#include "timerGlobals.h"
#include "agentPerformance.h"
#include "agentPerformanceTraining.h"
#include "instanceSetFile.h"
#include "instanceSetVirtual.h"
#include "JVector.h"
#include "messageBuffer.h"
#include "factory.h"

extern attributesInfo ai;
extern timerPruning *tPruning;
extern instanceSet *is;
extern timerGlobals *tGlobals;
extern int lastIteration;
extern messageBuffer mb;
extern int nodeRank;

void classifierStats(classifier & ind, char instancesFile[],
		     char *typeOfFile)
{
	unsigned long long i;
	agentPerformance ap(ind.getNumClassifiers(), ai.getNumClasses());
	instanceSetFile *is = new instanceSetFile(instancesFile, TEST);
	unsigned long long numInstances = is->getNumInstancesOfIteration();

	for (i = 0; i < numInstances; i++) {
		instance *ins = is->getInstance(i);
		int realClass = ins->instanceClass;
		int predictedClass = -1;
		int whichClassifier=ind.classify(ins);
		if (whichClassifier != -1) {
			predictedClass = ind.getClass(whichClassifier);
		}
		ap.addPrediction(realClass, predictedClass,
				 whichClassifier);
	}

	ind.setAccuracy(ap.getAccuracy());
	ap.dumpStats(typeOfFile);
	delete is;
}

void classifierStatsVirtual(classifier & ind, instanceSetVirtual *is)
{
	unsigned long long i;
	agentPerformance ap(ind.getNumClassifiers(), ai.getNumClasses());
	unsigned long long numInstances = is->getNumInstances();

	for (i = 0; i < numInstances; i++) {
		instance *ins = is->getInstanceTest(i);
		int realClass = ins->instanceClass;
		int predictedClass = -1;
		int whichClassifier=ind.classify(ins);
		if (whichClassifier != -1) {
			predictedClass = ind.getClass(whichClassifier);
		}
		ap.addPrediction(realClass, predictedClass,
				 whichClassifier);
	}

	ind.setAccuracy(ap.getAccuracy());
	ap.dumpStats("Train");
	delete is;
}




double classifierTest(classifier & ind, instanceSet *is)
{
	unsigned long long i;
	agentPerformance ap(ind.getNumClassifiers(), ai.getNumClasses());
	instance **instances= is->getAllInstances();
	unsigned long long numInstances = is->getNumInstances();

	for (i = 0; i < numInstances; i++) {
		int predictedClass = -1;
		int whichClassifier=ind.classify(instances[i]);
		if (whichClassifier != -1) {
			predictedClass = ind.getClass(whichClassifier);
		}
		ap.addPrediction(instances[i]->instanceClass, predictedClass,
				 whichClassifier);
	}

	ap.dumpStatsBrief();
	return ap.getAccuracy();
}

double classifierTestVirtual(classifier & ind, instanceSet *is)
{
	unsigned long long i;
	agentPerformance ap(ind.getNumClassifiers(), ai.getNumClasses());
	unsigned long long numInstances = is->getNumInstances();

	for (i = 0; i < numInstances; i++) {
		instance *ins=is->getInstanceTest(i);
		int predictedClass = -1;
		int whichClassifier=ind.classify(ins);
		if (whichClassifier != -1) {
			predictedClass = ind.getClass(whichClassifier);
		}
		ap.addPrediction(ins->instanceClass, predictedClass,
				 whichClassifier);
	}

	ap.dumpStatsBrief();
	return ap.getAccuracy();
}


double classifierFitness(classifier & ind)
{
	instance *ins;
	unsigned long long numInstances = is->getNumInstancesOfIteration();
	agentPerformanceTraining ap(ind.getNumClassifiers(), ai.getNumClasses());

	if(tGlobals->smartClassAssignment && ind.canDoSmartClassAsignment()) {
		ind.classifyTraining(is,&ap);
	//} else if(tGlobals->doTrainAndClean 
	//	&& ind.canDoSmartClassAsignment()) {
	//	ind.trainAndClean(is,&ap);
	} else {
		unsigned long long i;
		for (i = 0; i < numInstances; i++) {
			instance *ins=is->getInstance(i);
			int predictedClass = -1;
			int whichClassifier = ind.classify(ins);
			if (whichClassifier != -1) {
				predictedClass = ind.getClass(whichClassifier);
			}
	
			ap.addPrediction(ins->instanceClass, predictedClass
				, whichClassifier);
		}
	}

	ind.setAliveClassifiers(ap.getAliveClassifiers());
	ind.setAccuracy(ap.getAccuracy());
	double fitness=ap.getFitness(ind);
	tPruning->pruneClassifiers(ind, &ap);
	return fitness;
}

