#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>

#include "timerGlobals.h"
#include "timerCrossover.h"
#include "timerPruning.h"
#include "timerADI.h"
#include "timerHierar.h"
#include "timerMutation.h"
#include "classifier_adaptive.h"
#include "adaptiveRule.h"
#include "random.h"
#include "instanceSet.h"
#include "sampling.h"
#include "ruleOrderAgent.h"

extern timerGlobals *tGlobals;
extern timerCrossover *tCross;
extern timerPruning *tPruning;
extern timerHierar *tHierar;
extern timerMutation *tMut;
extern timerADI *tADI;
extern Random rnd;
extern instanceSet *is;
extern int lastIteration;

double classifier_adaptive::computateTheoryLength(agentPerformanceTraining * ap)
{
	int i;

	theoryLength = 0;
	int *ptr = chromosome;
	for (i = 0; i < numRules; i++) {
		if (ap->getActivationsOfClassifier(i))
			theoryLength += adaptiveRuleTheoryLength(ptr);
		ptr += tADI->ruleSize;
	}

	if (tGlobals->defaultClassPolicy != DISABLED) {
		//if(ap->getActivationsOfClassifier(numRules))
		theoryLength += tADI->defaultRuleTheoryLength;
	}

	return theoryLength;
}


classifier_adaptive::classifier_adaptive()
{
	int i;

	numRules =
	    rnd(tGlobals->minClassifiersInit,
		tGlobals->maxClassifiersInit);

	chromosome = new int[numRules * tADI->ruleSize];
	if (chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	length = 0;
	int *ptr = chromosome;
	for (i = 0; i < numRules; i++) {
		adaptiveRuleConstructor(ptr);
		length += ptr[0];
		ptr += tADI->ruleSize;
	}

	if (tGlobals->defaultClassPolicy == AUTO) {
		defaultClass = rnd(0, ai.getNumClasses() - 1);
	} else {
		defaultClass = tGlobals->defaultClass;
	}

	initialize_chromosome();
}

classifier_adaptive::
classifier_adaptive(const classifier_adaptive & orig, int son)
{
	*this = orig;

	if (!son) {
		chromosome = new int[numRules * tADI->ruleSize];
		if (chromosome == NULL) {
			perror("out of memory");
			exit(1);
		}

		bcopy(orig.chromosome, chromosome,
		      sizeof(int) * numRules * tADI->ruleSize);
	} else {
		chromosome = NULL;
	}
}

classifier_adaptive::~classifier_adaptive()
{
	delete chromosome;
}

void classifier_adaptive::initialize_chromosome()
{
	int i;
	int *ptr = chromosome;

	for (i = 0; i < numRules; i++) {
		adaptiveRuleInitialize(ptr, defaultClass);
		ptr += tADI->ruleSize;
	}
}

void classifier_adaptive::crossover(classifier * in,
				    classifier * out1, classifier * out2)
{
	switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this, (classifier_adaptive *) in,
			      (classifier_adaptive *) out1,
			      (classifier_adaptive *) out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_adaptive *) in,
			      (classifier_adaptive *) out1,
			      (classifier_adaptive *) out2);
	}
}

void classifier_adaptive::mutation()
{
	tMut->mutationCounter++;
	int rule = rnd(0, numRules - 1);
	int *ptr = &chromosome[rule * tADI->ruleSize];
	adaptiveRuleMutation(ptr, defaultClass);
	modif = 1;
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

// Creuament amb 2 punts de cutPoint
void classifier_adaptive::crossover_2px(classifier_adaptive * in1,
		classifier_adaptive * in2,
		classifier_adaptive * out1,
		classifier_adaptive * out2)
{
	int i1ptr1, i1ptr2, i2ptr1, i2ptr2;
	int cutPoint1, cutPoint2;
	int sizeSon1, sizeSon2;
	int i;

	out1->modif = out2->modif = 1;

	i1ptr1 = rnd(0, in1->numRules - 1);
	i1ptr2 = rnd(0, in1->numRules - 1);
	if (i1ptr1 > i1ptr2)
		swap(i1ptr1, i1ptr2);
	i2ptr1 = rnd(0, in2->numRules - 1);
	i2ptr2 = rnd(0, in2->numRules - 1);
	if (i2ptr1 > i2ptr2)
		swap(i2ptr1, i2ptr2);

	cutPoint1 = rnd(0, tGlobals->numAttributes);
	cutPoint2 = rnd(0, tGlobals->numAttributes);

	if (i1ptr1 == i1ptr2 || i2ptr1 == i2ptr2) {
		if (cutPoint1 > cutPoint2)
			swap(cutPoint1, cutPoint2);
	}

	sizeSon1 = (i1ptr1) + (i2ptr2 - i2ptr1) + (in1->numRules - i1ptr2);
	sizeSon2 = (i2ptr1) + (i1ptr2 - i1ptr1) + (in2->numRules - i2ptr2);


	out1->numRules = sizeSon1;
	out2->numRules = sizeSon2;

	out1->chromosome = new int[out1->numRules * tADI->ruleSize];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	out2->chromosome = new int[out2->numRules * tADI->ruleSize];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	bcopy(in1->chromosome, out1->chromosome,
			sizeof(int) * i1ptr1 * tADI->ruleSize);
	bcopy(in2->chromosome, out2->chromosome,
			sizeof(int) * i2ptr1 * tADI->ruleSize);

	adaptiveRuleCrossover(&in1->chromosome[i1ptr1 * tADI->ruleSize]
			, &in2->chromosome[i2ptr1 * tADI->ruleSize]
			, &out1->chromosome[i1ptr1 * tADI->ruleSize]
			, &out2->chromosome[i2ptr1 * tADI->ruleSize]
			, cutPoint1);

	if (i1ptr1 != i1ptr2)
		bcopy(&in1->chromosome[(i1ptr1 + 1) * tADI->ruleSize],
				&out2->chromosome[(i2ptr1 + 1) * tADI->ruleSize],
				sizeof(int) * (i1ptr2 - i1ptr1 -
					1) * tADI->ruleSize);
	if (i2ptr1 != i2ptr2)
		bcopy(&in2->chromosome[(i2ptr1 + 1) * tADI->ruleSize],
				&out1->chromosome[(i1ptr1 + 1) * tADI->ruleSize],
				sizeof(int) * (i2ptr2 - i2ptr1 -
					1) * tADI->ruleSize);

	int *ptr1, *ptr2;
	if (i1ptr1 == i1ptr2) {
		ptr2 = new int[tADI->ruleSize];
		bcopy(&out2->chromosome[i2ptr1 * tADI->ruleSize], ptr2,
				sizeof(int) * tADI->ruleSize);
	} else {
		ptr2 = &in1->chromosome[i1ptr2 * tADI->ruleSize];
	}

	if (i2ptr1 == i2ptr2) {
		ptr1 = new int[tADI->ruleSize];
		bcopy(&out1->chromosome[i1ptr1 * tADI->ruleSize], ptr1,
				sizeof(int) * tADI->ruleSize);
	} else {
		ptr1 = &in2->chromosome[i2ptr2 * tADI->ruleSize];
	}

	adaptiveRuleCrossover(ptr2, ptr1,
			&out2->
			chromosome[(i2ptr1 + i1ptr2 -
				i1ptr1) * tADI->ruleSize]
			,
			&out1->
			chromosome[(i1ptr1 + i2ptr2 -
				i2ptr1) * tADI->ruleSize],
			cutPoint2);

	if (i1ptr1 == i1ptr2)
		delete ptr2;
	if (i2ptr1 == i2ptr2)
		delete ptr1;

	bcopy(&in1->chromosome[(i1ptr2 + 1) * tADI->ruleSize],
			&out1->chromosome[(i1ptr1 + i2ptr2 - i2ptr1 + 1) *
			tADI->ruleSize],
			sizeof(int) * (in1->numRules - i1ptr2 - 1) * tADI->ruleSize);
	bcopy(&in2->chromosome[(i2ptr2 + 1) * tADI->ruleSize],
			&out2->chromosome[(i2ptr1 + i1ptr2 - i1ptr1 + 1) *
			tADI->ruleSize],
			sizeof(int) * (in2->numRules - i2ptr2 - 1) * tADI->ruleSize);

	out1->length = 0;
	int *ptr = out1->chromosome;
	for (i = 0; i < out1->numRules; i++) {
		out1->length += ptr[0];
		ptr += tADI->ruleSize;
	}

	out2->length = 0;
	ptr = out2->chromosome;
	for (i = 0; i < out2->numRules; i++) {
		out2->length += ptr[0];
		ptr += tADI->ruleSize;
	}
}



void classifier_adaptive::dumpPhenotype(char *string)
{
	int *ptr = chromosome;
	char temp[100000];
	int i, j;

	strcpy(string, "");
	for (i = 0; i < numRules; i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		adaptiveRuleDumpPhenotype(ptr, temp, defaultClass);
		strcat(string, temp);
		ptr += tADI->ruleSize;
	}
	if (tGlobals->defaultClassPolicy != DISABLED) {
		sprintf(temp, "%d:Default rule -> %s\n", i,
				ai.getNominalValue(tGlobals->numAttributesMC,
					defaultClass)->cstr());
		strcat(string, temp);
	}
	strcat(string, "\n");
}

void classifier_adaptive::crossover_1px(classifier_adaptive *
		in1,
		classifier_adaptive *
		in2,
		classifier_adaptive *
		out1, classifier_adaptive * out2)
{
	int i;

	out1->modif = out2->modif = 1;

	int sizeSon1, sizeSon2;
	int ruleP1, ruleP2;
	int cutPoint;

	ruleP1 = rnd(0, in1->numRules - 1);
	ruleP2 = rnd(0, in2->numRules - 1);
	cutPoint = rnd(0, tGlobals->numAttributes);
	sizeSon1 = ruleP1 + in2->numRules - ruleP2;
	sizeSon2 = ruleP2 + in1->numRules - ruleP1;

	out1->numRules = sizeSon1;
	out2->numRules = sizeSon2;

	out1->chromosome = new int[out1->numRules * tADI->ruleSize];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	out2->chromosome = new int[out2->numRules * tADI->ruleSize];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	// Copiem les regles anteriors al punt de crossover
	bcopy(in1->chromosome, out1->chromosome,
			sizeof(int) * ruleP1 * tADI->ruleSize);
	bcopy(in2->chromosome, out2->chromosome,
			sizeof(int) * ruleP2 * tADI->ruleSize);

	// Punt de cutPoint
	adaptiveRuleCrossover(&in1->chromosome[ruleP1 * tADI->ruleSize]
			, &in2->chromosome[ruleP2 * tADI->ruleSize]
			, &out1->chromosome[ruleP1 * tADI->ruleSize],
			&out2->chromosome[ruleP2 * tADI->ruleSize]
			, cutPoint);

	// Copiem les regles de desprs del punt
	bcopy(&chromosome[(ruleP1 + 1) * tADI->ruleSize],
			&out2->chromosome[(ruleP2 + 1) * tADI->ruleSize],
			sizeof(int) * (numRules - ruleP1 - 1) * tADI->ruleSize);
	bcopy(&in2->chromosome[(ruleP2 + 1) * tADI->ruleSize],
			&out1->chromosome[(ruleP1 + 1) * tADI->ruleSize],
			sizeof(int) * (in2->numRules - ruleP2 - 1) * tADI->ruleSize);


	out1->length = 0;
	int *ptr = out1->chromosome;
	for (i = 0; i < out1->numRules; i++) {
		out1->length += ptr[0];
		ptr += tADI->ruleSize;
	}

	out2->length = 0;
	ptr = out2->chromosome;
	for (i = 0; i < out2->numRules; i++) {
		out2->length += ptr[0];
		ptr += tADI->ruleSize;
	}
}

void classifier_adaptive::pruneClassifiers(int *classifiersToPrune,
		int numPruned)
{
	int *newChromosome;
	int i;

	if (numRules == 1 || numPruned == 0)
		return;

	newChromosome = new int[(numRules - numPruned) * tADI->ruleSize];

	int countPruned = 0;
	int *ptrOrig = chromosome;
	int *ptrDest = newChromosome;

	for (i = 0; i < numRules; i++) {
		if (countPruned < numPruned) {
			if (i != classifiersToPrune[countPruned]) {
				bcopy(ptrOrig, ptrDest,
						tADI->ruleSize * sizeof(int));
				ptrDest += tADI->ruleSize;
			} else {
				length -= ptrOrig[0];
				countPruned++;
			}
		} else {
			bcopy(ptrOrig, ptrDest,
					tADI->ruleSize * sizeof(int));
			ptrDest += tADI->ruleSize;
		}
		ptrOrig += tADI->ruleSize;
	}

	delete chromosome;
	chromosome = newChromosome;
	numRules -= numPruned;
}

int classifier_adaptive::classify(instance * ins)
{
	int i;
	int *ptr = chromosome;

	for (i = 0; i < numRules; i++) {
		if (adaptiveRuleMatches(ptr, ins)) {
			return i;
		}
		ptr += tADI->ruleSize;
	}

	if (tGlobals->defaultClassPolicy == DISABLED)
		return -1;
	return numRules;
}

int classifier_adaptive::countDiscretizersForAttribute(int attribute,
		int *counters)
{
	int i;
	int count = 0;
	int *ptr = chromosome;

	for (i = 0; i < numRules; i++) {
		count +=
			adaptiveRuleCountDiscretizers(ptr, attribute,
					counters);
		ptr += tADI->ruleSize;
	}
	return count;
}

int classifier_adaptive::countDiscretizers(int *counters)
{
	int i;
	int count = 0;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		count += countDiscretizersForAttribute(i, counters);
	}
	return count;
}


void classifier_adaptive::doSpecialStage(int stage)
{
	if (stage != 0 && stage != 1 && stage != 2)
		return;

	if (tADI->stageOrder[stage] == SPLIT) {
		doSplit();
	} else if (tADI->stageOrder[stage] == MERGE) {
		doMerge();
	} else if (tADI->stageOrder[stage] == REINITIALIZE) {
		doReinitialize();
	}
}

void classifier_adaptive::doSplit()
{
	int i;
	int *ptr = chromosome;

	for (i = 0; i < numRules; i++, ptr += tADI->ruleSize) {
		length -= ptr[0];
		int res = adaptiveRuleDoSplit(ptr);
		if (res)
			modif = 1;
		length += ptr[0];
	}
}

void classifier_adaptive::doMerge()
{
	int i;
	int *ptr = chromosome;

	for (i = 0; i < numRules; i++, ptr += tADI->ruleSize) {
		length -= ptr[0];
		int res = adaptiveRuleDoMerge(ptr);
		if (res)
			modif = 1;
		length += ptr[0];
	}
}

void classifier_adaptive::doReinitialize()
{
	int i;
	int *ptr = chromosome;

	for (i = 0; i < numRules; i++, ptr += tADI->ruleSize) {
		length -= ptr[0];
		int res = adaptiveRuleDoReinitialize(ptr);
		if (res)
			modif = 1;
		length += ptr[0];
	}
}


void classifier_adaptive::postProcessIndividual()
{
	//reorderRules();
}

void classifier_adaptive::findOrder(JVector<matchProfileAgent *>&profiles
	,JVector<int *>&ptrs,JVector<int> &ruleOrder) 
{
	unsigned long long numInst=is->getNumInstancesOfIteration();
	int i,j;

	int totalRules=profiles.size();
	ruleOrderAgent *best=NULL;
	for(j=0;j<tCross->repetitionsRuleOrdering;j++) {
		Sampling samp(totalRules);
		ruleOrderAgent *roa = new ruleOrderAgent(defaultClass,numInst,totalRules);

		for(i=0;i<totalRules;i++) {
			int rule=samp.getSample();
			roa->insertRule(rule,profiles[rule]);
		}

		roa->refineRuleSet();

		if(best==NULL) {
			best=roa;
		} else {
			double accOld=best->getAccuracy();
			double accNew=roa->getAccuracy();
			if(accNew>accOld) {
				delete best;
				best=roa;
			} else if(accNew==accOld 
				&& roa->getNumRules()<best->getNumRules()) {
				delete best;
				best=roa;
			} else {
				delete roa;
			}
		}
	}

	best->copyOrder(ruleOrder);
	if(ruleOrder.size()==0) {
		int pos=rnd(0,totalRules-1);
		ruleOrder.addElement(pos);
	}
	delete best;
}


void classifier_adaptive::crossover_smart(classifier **parents, int num) 
{
	int i, j, k;

	modif = 1;
	if(chromosome) delete chromosome;

	JVector<int *>ptrs;
	JVector<matchProfileAgent *>profiles;
	int ruleCount=0;

	for(k=0;k<num;k++) {
		classifier_adaptive *parent=(classifier_adaptive *)parents[k];
		int *ptr=parent->chromosome;
		for(i=0;i<parent->numRules;ptr+=tADI->ruleSize,i++) {
			matchProfileAgent *mpa = evaluateRuleCX(ptr);
			profiles.addElement(mpa);
			ptrs.addElement(ptr);
			ruleCount++;
		}
	}

	JVector<int> ruleOrder;
	findOrder(profiles,ptrs,ruleOrder);

	numRules = ruleOrder.size(); 
	chromosome = new int[numRules * tADI->ruleSize];
	length = 0;
	for (i = 0; i < numRules; i++) {
		bcopy(ptrs[ruleOrder[i]], &chromosome[i * tADI->ruleSize],
		      sizeof(int) * tADI->ruleSize);
		length += ptrs[ruleOrder[i]][0];
	}

	for(i=0;i<profiles.size();i++) {
		delete profiles[i];
	}
	for(i=ruleCount;i<ptrs.size();i++) {
		delete ptrs[i];
	}
}

void classifier_adaptive::classifyTraining(instanceSet *is,agentPerformanceTraining *ap)
{
	unsigned long long numInstances = is->getNumInstancesOfIteration();
	int instanceMatched[numInstances];
	int instanceMatchedTemp[numInstances];
	int numClasses=ai.getNumClasses();
	double classThresholds[numClasses];
	int classCounts[numClasses];
	int defC=defaultClass;
	if(tGlobals->defaultClassPolicy==DISABLED) defC=-1;
	unsigned long long i,j,k;

	for(i=0;i<numInstances;i++) {
		instanceMatched[i]=0;
	}
	for(j=0;j<numClasses;j++) classThresholds[j]=numInstances*0.02;

	/*for(j=0;j<numClasses;j++) classThresholds[j]=0;
	for(i=0;i<numInstances;i++) {
		classThresholds[is->getInstance(i)->getClass()]++;
	}
        for(j=0;j<numClasses;j++) {
                double old=classThresholds[j];
                classThresholds[j]*=0.05;
                if(classThresholds[j]<5) classThresholds[j]=5;
                if(classThresholds[j]>old*0.20) {
                        classThresholds[j]=old*0.20;
                }
        }*/

	unsigned long long instancesLeft=numInstances;

	int *ptr=chromosome;
	for(i=0;i<numRules;i++) {
		for(j=0;j<numClasses;j++) classCounts[j]=0;
		unsigned long long numMatches=0;

		for(j=0;j<numInstances;j++) {
			if(!instanceMatched[j]) {
				instance *ins=is->getInstance(j);
				if (adaptiveRuleMatches(ptr, ins)) {
					instanceMatchedTemp[numMatches++]=j;
					classCounts[ins->getClass()]++;
				}
			}
		}

		if(numMatches) {
			unsigned long long max=0;
			int posMax=-1;
			for(j=0;j<numClasses;j++) {
				if(j==defC) continue;
				if(posMax==-1 || classCounts[j]>max) {
					max=classCounts[j];
					posMax=j;
				}
			}

			if(max >= classThresholds[posMax]) {
				ptr[1]=posMax;

				for(j=0;j<numClasses;j++) {
					for(k=0;k<classCounts[j];k++) {
						ap->addPrediction(j,posMax,i);
					}
				}

				for(j=0;j<numMatches;j++) {
					instanceMatched[instanceMatchedTemp[j]]=1;
				}

				instancesLeft-=numMatches;
				if(instancesLeft==0) {
					break;
				}
			} else {
				if(lastIteration) 
					adaptiveRuleDisable(ptr);
			}
		}
		ptr+=tADI->ruleSize;
	}

	if(instancesLeft) {
		int predRule=-1;
		if(defC!=-1) predRule=numRules;

		for(j=0;j<numInstances;j++) {
			if(!instanceMatched[j]) {
				instance *ins=is->getInstance(j);
				ap->addPrediction(ins->getClass()
					,defC,predRule);
			}
		}
	}
}

