#ifndef _CLASSIFICADOR_ADAPTATIU_

#define _CLASSIFICADOR_ADAPTATIU_

#include "classifier.h"
#include "timerADI.h"
#include "timerHierar.h"
#include "timerGlobals.h"
#include "agentPerformanceTraining.h"
#include "matchProfileAgent.h"

extern timerADI *tADI;
extern timerGlobals *tGlobals;
extern timerHierar *tHierar;


class classifier_adaptive: public classifier   {
	int *chromosome;
	int numRules;

	void crossover_1px(classifier_adaptive *in1,classifier_adaptive *in2
		,classifier_adaptive *out1,classifier_adaptive *out2);
	void crossover_2px(classifier_adaptive *in1,classifier_adaptive *in2
		,classifier_adaptive *out1,classifier_adaptive *out2);
	void findOrder(JVector<matchProfileAgent *>&profiles,JVector<int *>&ptrs,JVector<int> &ruleOrder);

	void initialize_chromosome(void);

public:
	classifier_adaptive();
	~classifier_adaptive();
	classifier_adaptive(const classifier_adaptive &orig,int son=0);

	inline int getClass(int rule)
	{
		if(rule==numRules) {
			return defaultClass;
		}

		return chromosome[rule * tADI->ruleSize + 1];
	}

	inline double getHierarchicalSelectionLength()
	{
		if(tHierar->useMDL) return getTheoryLength();
		return length*(double)aliveClassifiers/(double)getNumClassifiers();
	}

	inline int getNiche() {
		if(tGlobals->defaultClassPolicy!=AUTO) return 0;
		return defaultClass;
	}
	inline int getNumNiches() {
		if(tGlobals->defaultClassPolicy!=AUTO) return 1;
		return ai.getNumClasses();
	}

		
	int getNumClassifiers() {
		int nC=numRules;
		if(tGlobals->defaultClassPolicy!=DISABLED) nC++;
		return nC;
	}
	void pruneClassifiers(int *classifiersToPrune,int numPruned);
	void dumpPhenotype(int classifier,int attribute,char *string); 
	int classify(instance *i);
	double computateTheoryLength(agentPerformanceTraining *ap);
	void crossover(classifier *,classifier *,classifier *);
	void crossover_smart(classifier **,int num);
	inline int hasSmartCrossover() {return 1;}
	void mutation();
	
	void dumpPhenotype(char *string);
	
	inline int numSpecialStages(){return 3;}
	void doSpecialStage(int stage);
	void doSplit();
	void doMerge();
	void doReinitialize();
	int countDiscretizers(int *counters);
	int countDiscretizersForAttribute(int attribute,int *counters);
	inline double getIntervalsPerAttribute() {
		double res=length-numRules;
		double div=numRules;
		res/=div;
		return res/(double)tGlobals->numAttributesMC;
	}
	void postProcessIndividual();

	int canDoSmartClassAsignment() {return 1;}
	void classifyTraining(instanceSet *is,agentPerformanceTraining *ap);
};

#endif

