#include "classifier_gabil.h"
#include "random.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "attributesInfo.h"
#include "timerSymbolicKR.h"
#include "timerMutation.h"
#include "timerHierar.h"
#include "timerCrossover.h"
#include "timerMDL.h"
#include "timerGlobals.h"
#include "instanceSet.h"
#include "ruleOrderAgent.h"
#include "sampling.h"
#include "messageBuffer.h"

extern messageBuffer mb;
extern attributesInfo ai;
extern timerSymbolicKR *tSymbolic;
extern timerHierar *tHierar;
extern timerGlobals *tGlobals;
extern timerCrossover *tCross;
extern timerMutation *tMut;
extern timerMDL *tMDL;
extern Random rnd;
extern instanceSet *is;
extern int lastIteration;

double classifier_gabil::computateTheoryLength(agentPerformanceTraining *ap)
{
	int i, j, k;
	unsigned char *ptr = chromosome;
	theoryLength = 0;

	for (i = 0; i < numRules; i++) {
		if(ap->getActivationsOfClassifier(i))
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			int numFalse = 0;
			int basePoint = tSymbolic->offsetAttribute[j];
			int numValues = tSymbolic->sizeAttribute[j];
			for (k = 0; k < numValues; k++) {
				if (!ptr[basePoint + k]) numFalse++;
			}
			theoryLength += numValues + numFalse; 
		}
		ptr+=tSymbolic->ruleSize;
	}
	if(tGlobals->defaultClassPolicy!=DISABLED)
		//if(ap->getActivationsOfClassifier(numRules))
			theoryLength+=tSymbolic->defaultRuleTheoryLength;

	return theoryLength;
}

classifier_gabil::classifier_gabil()
{
	numRules =
	    rnd(tGlobals->minClassifiersInit, tGlobals->maxClassifiersInit);
	length = tSymbolic->ruleSize * numRules;
	chromosome = new unsigned char[length];

	if(tGlobals->defaultClassPolicy==AUTO) {
		defaultClass=rnd(0,ai.getNumClasses()-1);
	} else  {
		defaultClass=tGlobals->defaultClass;
	}

	classifier_gabil::initializeChromosome();
}

classifier_gabil::
classifier_gabil(const classifier_gabil & orig, int son)
{
	*this = orig;

	if (!son) {
		chromosome = new unsigned char[length];
		bcopy(orig.chromosome, chromosome,
		      length * sizeof(unsigned char));

	} else {
		chromosome = NULL;
	}
}

classifier_gabil::~classifier_gabil()
{
	delete chromosome;
}

void classifier_gabil::gabilRuleCover(unsigned char *rule,instance *ins,double prob)
{
	int j, k;

	unsigned char *ptr=rule;
	for (j = 0; j < tGlobals->numAttributesMC; j++) {
		int size=tSymbolic->sizeAttribute[j];
		double newProb;
		int value;
		if(ins) {
			value=ins->nominalValues[j];
			newProb=(size*prob-1.0)/(size-1.0);
		} else {
			value = -1;
			newProb=prob;
		}

		for (k = 0; k < size; k++){
			if(k!=value) {
				if (!rnd < prob) ptr[k]=1;
				else ptr[k]=0;
			} else {
				ptr[k]=1;
			}
		}
		ptr+=tSymbolic->sizeAttribute[j];
	}

	if(ins) {
		ptr[0]=ins->instanceClass;
	} else {
		int cl;
		do {
			cl=rnd(0,ai.getNumClasses()-1);
		} while(tGlobals->defaultClassPolicy!=DISABLED 
			&& cl==defaultClass);
		ptr[0]=cl;
	}
}

void classifier_gabil::initializeChromosome()
{
	int i;

	unsigned char *ptr=chromosome;
	for(i=0;i<numRules;i++) {
		instance *ins=NULL;
		if(tGlobals->smartInit) {
			if(tGlobals->defaultClassPolicy!=DISABLED) {
				ins=is->getInstanceInit(defaultClass);
			} else {
				ins=is->getInstanceInit(ai.getNumClasses());
			}
		}

		gabilRuleCover(ptr,ins,tGlobals->probOne);
		ptr+=tSymbolic->ruleSize;
	}
}

void classifier_gabil::setGene(short int rule, short int gene,
				short int value, unsigned char nfo)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules) ||
	    (value < 0) || (value >= tSymbolic->sizeAttribute[gene])) {
		fprintf(stderr,"Invalid access at setGene %d %d %d", rule,
		       gene, value);
		exit(1);
	}
#endif
	chromosome[rule * tSymbolic->ruleSize +
		  tSymbolic->offsetAttribute[gene] + value] = nfo;
	modif = 1;
}

unsigned char classifier_gabil::getGene(short int rule,
					 short int gene, short int value)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules) ||
	    (value < 0) || (value >= tSymbolic->sizeAttribute[gene])) {
		fprintf(stderr,"Invalid access at getGene %d %d %d", rule,
		       gene, value);
		exit(1);
	}
#endif
	return chromosome[rule * tSymbolic->ruleSize +
			 tSymbolic->offsetAttribute[gene] + value];
}

// Els pares sn l'individu que executa la funci i in, i els sons, out1 
// i out2
void classifier_gabil::crossover(classifier * in,
				   classifier * out1, classifier * out2)
{

	switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this, (classifier_gabil *) in,
			      (classifier_gabil *) out1,
			      (classifier_gabil *) out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_gabil *) in,
			      (classifier_gabil *) out1,
			      (classifier_gabil *) out2);
	}

}

// gene indica un gen dins el chromosome sencer, cal separar rule, attribute
// i value
void classifier_gabil::mutation()
{
	int i;
	short int rule, attribute, value;

	// Modificarem el chromosome
	tMut->mutationCounter++;
	modif = 1;

	rule = rnd(0,numRules-1);
	if(!tGlobals->smartClassAssignment && tGlobals->numClasses>1 
			&& !rnd<0.10) {
		attribute = tGlobals->numAttributesMC;
		value=0;
	} else {
		attribute=rnd(0,tGlobals->numAttributesMC-1);
		value=rnd(0,tSymbolic->sizeAttribute[attribute]-1);
	}

	


	if (attribute != tGlobals->numAttributesMC ) {
		if (getGene(rule, attribute, value) == 0)
			setGene(rule, attribute, value, 1);
		else
			setGene(rule, attribute, value, 0);
	} else {
		int oldValue=getGene(rule,attribute,value);
		int newValue;
		do {
			newValue = rnd(0, ai.getNumClasses()-1);
		} while(newValue==oldValue || tGlobals->defaultClassPolicy!=DISABLED && newValue==defaultClass);
		setGene(rule, attribute, value, newValue);
	}
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

void classifier_gabil::crossover_2px(classifier_gabil * in1,
				       classifier_gabil * in2,
				       classifier_gabil * out1,
				       classifier_gabil * out2)
{
	int i1rule1, i1rule2, i2rule1, i2rule2;
	int cutPoint1, cutPoint2, sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	i1rule1 = rnd(0, in1->numRules - 1);
	i1rule2 = rnd(0, in1->numRules - 1);
	if (i1rule1 > i1rule2)
		swap(i1rule1, i1rule2);

	i2rule1 = rnd(0, in2->numRules - 1);
	i2rule2 = rnd(0, in2->numRules - 1);
	if (i2rule1 > i2rule2)
		swap(i2rule1, i2rule2);

	sizeOfSon1 =
	    i1rule1 + i2rule2 - i2rule1 + in1->numRules - i1rule2;
	sizeOfSon2 =
	    i2rule1 + i1rule2 - i1rule1 + in2->numRules - i2rule2;

	cutPoint1 = rnd(0, tSymbolic->ruleSize - 1);
	cutPoint2 = rnd(0, tSymbolic->ruleSize - 1);

	if (i1rule1 == i1rule2 || i2rule1 == i2rule2) {
		if (cutPoint1 > cutPoint2)
			swap(cutPoint1, cutPoint2);
	}

	out1->length = sizeOfSon1 * tSymbolic->ruleSize;
	out2->length = sizeOfSon2 * tSymbolic->ruleSize;
	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;

	out1->chromosome = new unsigned char[sizeOfSon1 * tSymbolic->ruleSize];
	out2->chromosome = new unsigned char[sizeOfSon2 * tSymbolic->ruleSize];

	bcopy(chromosome, out1->chromosome, i1rule1 * tSymbolic->ruleSize);
	bcopy(in2->chromosome, out2->chromosome,
	      i2rule1 * tSymbolic->ruleSize);

	bcopy(&in1->chromosome[i1rule1 * tSymbolic->ruleSize],
	      &out1->chromosome[i1rule1 * tSymbolic->ruleSize], cutPoint1);
	bcopy(&in2->chromosome[i2rule1 * tSymbolic->ruleSize],
	      &out2->chromosome[i2rule1 * tSymbolic->ruleSize], cutPoint1);

	bcopy(&in2->chromosome[i2rule1 * tSymbolic->ruleSize + cutPoint1],
	      &out1->chromosome[i1rule1 * tSymbolic->ruleSize + cutPoint1],
	      tSymbolic->ruleSize - cutPoint1);
	bcopy(&in1->chromosome[i1rule1 * tSymbolic->ruleSize + cutPoint1],
	      &out2->chromosome[i2rule1 * tSymbolic->ruleSize + cutPoint1],
	      tSymbolic->ruleSize - cutPoint1);

	if (i2rule1 != i2rule2)
		bcopy(&in2->
		      chromosome[(i2rule1 + 1) * tSymbolic->ruleSize],
		      &out1->chromosome[(i1rule1 + 1) *
				       tSymbolic->ruleSize],
		      (i2rule2 - (i2rule1 + 1)) * tSymbolic->ruleSize);
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[(i1rule1 + 1) * tSymbolic->ruleSize],
		      &out2->chromosome[(i2rule1 + 1) *
				       tSymbolic->ruleSize],
		      (i1rule2 - (i1rule1 + 1)) * tSymbolic->ruleSize);


	if (i2rule1 != i2rule2)
		bcopy(&in2->chromosome[i2rule2 * tSymbolic->ruleSize],
		      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
				       tSymbolic->ruleSize], cutPoint2);
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[i1rule2 * tSymbolic->ruleSize],
		      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
				       tSymbolic->ruleSize], cutPoint2);
	
	bcopy(&in1->chromosome[i1rule2 * tSymbolic->ruleSize + cutPoint2],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
			       tSymbolic->ruleSize + cutPoint2],
	      tSymbolic->ruleSize - cutPoint2);
	bcopy(&in2->chromosome[i2rule2 * tSymbolic->ruleSize + cutPoint2],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
			       tSymbolic->ruleSize + cutPoint2],
	      tSymbolic->ruleSize - cutPoint2);

	bcopy(&in1->chromosome[(i1rule2 + 1) * tSymbolic->ruleSize],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1 + 1) *
			       tSymbolic->ruleSize],
	      (in1->numRules - (i1rule2 + 1)) * tSymbolic->ruleSize);
	bcopy(&in2->chromosome[(i2rule2 + 1) * tSymbolic->ruleSize],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1 + 1) *
			       tSymbolic->ruleSize],
	      (in2->numRules - (i2rule2 + 1)) * tSymbolic->ruleSize);

}

void classifier_gabil::dumpPhenotype(char *string)
{
	unsigned char *ptr = chromosome;
	char temp[10000];
	char temp2[1000];
	int i, j, k;

	strcpy(string, "");
	for (i = 0; i < numRules; i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		unsigned char *ptr2 = ptr;
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			sprintf(temp,"Att %s is "
				,ai.getAttributeName(j)->cstr());
			int irr=1;
			for (k = 0; k < tSymbolic->sizeAttribute[j]; k++) {
				if(ptr2[k]) {
					sprintf(temp2,"%s,",ai.getNominalValue(j,k)->cstr());
					strcat(temp,temp2);
				} else {
					irr=0;
				}
			}

			if(!irr) {
				if(temp[strlen(temp) - 1] == ',')
					temp[strlen(temp) - 1] = 0;
				strcat(string, temp);
				strcat(string, "|");
			}
			ptr2 += tSymbolic->sizeAttribute[j];
		}
		int cl=(int)*ptr2;
		sprintf(temp, "%s\n",
			ai.getNominalValue(tGlobals->numAttributesMC
				, cl)->cstr());
		strcat(string, temp);
		ptr += tSymbolic->ruleSize;
	}
	if(tGlobals->defaultClassPolicy!=DISABLED) {
		sprintf(temp, "%d:Default rule -> %s\n", i
			,ai.getNominalValue(tGlobals->numAttributesMC
				,defaultClass)->cstr());
		strcat(string, temp);
	}
	strcat(string, "\n");
}

void classifier_gabil::dumpPhenotypeRule(unsigned char *ptr)
{
	int j,k;
	char temp[10000];

	for (j = 0; j < tGlobals->numAttributesMC; j++) {
		for (k = 0; k < tSymbolic->sizeAttribute[j]; k++) {
			printf("%d",ptr[k]);
		}
		printf("|");
		ptr += tSymbolic->sizeAttribute[j];
	}
	int cl=(int)*ptr;
	printf("%s\n",
		ai.getNominalValue(tGlobals->numAttributesMC
			, cl)->cstr());
}

void classifier_gabil::informedCrossover(unsigned char *father1
	,unsigned char *father2,unsigned char *son1,unsigned char *son2) {
	int i,j;

	
	for(i=0;i<tCross->numBB;i++) {
		// Let's choose parent...
		if(!rnd<0.5) {
			for(j=0;j<tCross->sizeBBs[i];j++) {
				int att=tCross->defBBs[i][j];
				bcopy(&father1[tSymbolic->offsetAttribute[att]]
					,&son1[tSymbolic->offsetAttribute[att]]
					,sizeof(unsigned char)
						*tSymbolic->sizeAttribute[att]);
				bcopy(&father2[tSymbolic->offsetAttribute[att]]
					,&son2[tSymbolic->offsetAttribute[att]]
					,sizeof(unsigned char)
						*tSymbolic->sizeAttribute[att]);
			}
		} else {
			for(j=0;j<tCross->sizeBBs[i];j++) {
				int att=tCross->defBBs[i][j];
				bcopy(&father1[tSymbolic->offsetAttribute[att]]
					,&son2[tSymbolic->offsetAttribute[att]]
					,sizeof(unsigned char)
						*tSymbolic->sizeAttribute[att]);
				bcopy(&father2[tSymbolic->offsetAttribute[att]]
					,&son1[tSymbolic->offsetAttribute[att]]
					,sizeof(unsigned char)
						*tSymbolic->sizeAttribute[att]);
			}
		}
	}
}

void classifier_gabil::crossover_1px(classifier_gabil * in1,
				       classifier_gabil * in2,
				       classifier_gabil * out1,
				       classifier_gabil * out2)
{
	out1->modif = out2->modif = 1;

	unsigned int attribute, rulep1, rulep2;
	int sizeOfSon1, sizeOfSon2;

	rulep1 = rnd(0, in1->numRules - 1);
	rulep2 = rnd(0, in2->numRules - 1);
	sizeOfSon1 = rulep1 + in2->numRules - rulep2;
	sizeOfSon2 = rulep2 + in1->numRules - rulep1;


	out1->length = sizeOfSon1 * tSymbolic->ruleSize;
	out2->length = sizeOfSon2 * tSymbolic->ruleSize;
	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;
	
	out1->chromosome = new unsigned char[sizeOfSon1 * tSymbolic->ruleSize];
	out2->chromosome = new unsigned char[sizeOfSon2 * tSymbolic->ruleSize];

	//Before cut point	
	bcopy(in1->chromosome, out1->chromosome,
	      rulep1 * tSymbolic->ruleSize);

	bcopy(in2->chromosome, out2->chromosome,
	      rulep2 * tSymbolic->ruleSize);

	//Rule in cut point
	if(tCross->informedCrossover && !rnd<0.25) {
		informedCrossover(&in1->chromosome[rulep1 * tSymbolic->ruleSize]
			,&in2->chromosome[rulep2 * tSymbolic->ruleSize]
			,&out1->chromosome[rulep1 * tSymbolic->ruleSize]
			,&out2->chromosome[rulep2 * tSymbolic->ruleSize]);
	} else {
		int cutPoint = rnd(0, tSymbolic->ruleSize - 1);
		bcopy(&in1->chromosome[rulep1 * tSymbolic->ruleSize],
		      &out1->chromosome[rulep1 * tSymbolic->ruleSize], cutPoint);
		bcopy(&in2->chromosome[rulep2 * tSymbolic->ruleSize],
		      &out2->chromosome[rulep2 * tSymbolic->ruleSize], cutPoint);
		
		bcopy(&in1->chromosome[rulep1 * tSymbolic->ruleSize + cutPoint],
		      &out2->chromosome[rulep2 * tSymbolic->ruleSize + cutPoint],
		      tSymbolic->ruleSize - cutPoint);
		bcopy(&in2->chromosome[rulep2 * tSymbolic->ruleSize + cutPoint],
		      &out1->chromosome[rulep1 * tSymbolic->ruleSize + cutPoint],
		      tSymbolic->ruleSize - cutPoint);
	}
	
	//After cut point
	bcopy(&in2->chromosome[(rulep2 + 1) * tSymbolic->ruleSize],
	      &out1->chromosome[(rulep1 + 1) * tSymbolic->ruleSize],
	      (in2->numRules - (rulep2 + 1)) * tSymbolic->ruleSize);

	bcopy(&in1->chromosome[(rulep1 + 1) * tSymbolic->ruleSize],
	      &out2->chromosome[(rulep2 + 1) * tSymbolic->ruleSize],
	      (in1->numRules - (rulep1 + 1)) * tSymbolic->ruleSize);

}

void classifier_gabil::pruneClassifiers(int *classifiersToPrune, int numPruned)
{
	unsigned char *newChromosome;
	int i;

	if (numRules == 1 || numPruned==0) return;

	if(classifiersToPrune[numPruned-1]==numRules) numPruned--;

	newChromosome = new unsigned char[(numRules - numPruned) *
			      tSymbolic->ruleSize];
	if (newChromosome == NULL) {
		fprintf(stderr,"Out of memory\n");
		exit(1);
	}

	int countPruned = 0;
	unsigned char *ptrOrig = chromosome;
	unsigned char *ptrDest = newChromosome;
	for (i = 0; i < numRules; i++) {
		if (countPruned < numPruned) {
			if (i != classifiersToPrune[countPruned]) {
				bcopy(ptrOrig, ptrDest, tSymbolic->ruleSize);
				ptrDest += tSymbolic->ruleSize;
			} else {
				countPruned++;
			}
		} else {
			bcopy(ptrOrig, ptrDest, tSymbolic->ruleSize);
			ptrDest += tSymbolic->ruleSize;
		}
		ptrOrig += tSymbolic->ruleSize;
	}

	delete chromosome;
	chromosome = newChromosome;

	numRules -= numPruned;
	length = numRules * tSymbolic->ruleSize;
}


int classifier_gabil::classify(instance *ins)
{
	unsigned char *ptr = chromosome;
	int i, j, match;

	if (tGlobals->ignoreMissingValues && ins->hasMissingValues()) {
		for (i = 0; i < numRules; i++) {
			if(ruleMatchesMissing(ptr,ins)) return i;
			ptr += tSymbolic->ruleSize;
		}
	} else {
		for (i = 0; i < numRules; i++) {
			if(ruleMatches(ptr,ins)) return i;
			ptr += tSymbolic->ruleSize;
		}

	}

	if(tGlobals->defaultClassPolicy==DISABLED) return -1;
	return numRules;
}


/*void classifier_gabil::classifyTraining(instanceSet *is,agentPerformanceTraining *ap)
{
	int numInstances = is->getNumInstancesOfIteration();	
	instance *instances[numInstances];
	int classes[numInstances];
	int instanceMatched[numInstances];
	int instanceMatchedTemp[numInstances];
	int numClasses=ai.getNumClasses();
	double classThresholds[numClasses];
	int classCounts[numClasses];
	int defC=defaultClass;
	if(tGlobals->defaultClassPolicy==DISABLED) defC=-1;
	int i,j,k;

	for(i=0;i<numInstances;i++) {
		instances[i]=is->getInstance(i);
		classes[i]=instances[i]->instanceClass;
		instanceMatched[i]=0;
	}
	//for(j=0;j<numClasses;j++) classThresholds[j]=numInstances*0.02;

	for(j=0;j<numClasses;j++) classThresholds[j]=0;
	for(i=0;i<numInstances;i++) {
		classThresholds[classes[i]]++;
	}
	for(j=0;j<numClasses;j++) {
		double old=classThresholds[j];
		classThresholds[j]=5;
		if(classThresholds[j]>old*0.20) {
			classThresholds[j]=old*0.20;
		}
	}

	int instancesLeft=numInstances;

	unsigned char *ptr=chromosome;	
	for(i=0;i<numRules;i++) {
		for(j=0;j<numClasses;j++) classCounts[j]=0;
	
		int numMatches=0;	
		for(j=0;j<numInstances;j++) {
			if(!instanceMatched[j]) {
				if(ruleMatches(ptr,instances[j])) {
					instanceMatchedTemp[numMatches++]=j;
					classCounts[classes[j]]++;
				}
			}
		}

		if(numMatches) {
			int max=0;
			int posMax=-1;
			for(j=0;j<numClasses;j++) {
				if(j==defC) continue;
				if(posMax==-1 || classCounts[j]>max) {
					max=classCounts[j];
					posMax=j;
				}
			}

			if(max >= classThresholds[posMax]) {
				ptr[tSymbolic->ruleSize - 1]=posMax;
	
				for(j=0;j<numClasses;j++) {
					for(k=0;k<classCounts[j];k++) {
						ap->addPrediction(j,posMax,i);
					}
				}

				for(j=0;j<numMatches;j++) {
					instanceMatched[instanceMatchedTemp[j]]=1;
				}

				instancesLeft-=numMatches;
				if(instancesLeft==0) {
					break;
				}
			} else {
				if(lastIteration) {
					int att=rnd(0,tGlobals->numAttributesMC-1);
					int index = tSymbolic->offsetAttribute[att];
					int numValues = tSymbolic->sizeAttribute[att];
					for(j=0;j<numValues;j++) {
						ptr[index++]=0;
					}
				}
			}
		}
		ptr+=tSymbolic->ruleSize;
	}

	if(instancesLeft) {
		int predRule=-1;
		if(defC!=-1) predRule=numRules;

		for(j=0;j<numInstances;j++) {
			if(!instanceMatched[j]) {
				ap->addPrediction(classes[j],defC,predRule);
			}
		}
	}
}*/

int classifier_gabil::ruleCleaning(unsigned char *ptr,unsigned long long *matched,unsigned long long numMatches)
{
        unsigned long long i,j,k;
        instance **instances=is->getInstancesOfIteration();

        unsigned long long *listOK = new unsigned long long[numMatches];
        unsigned long long *listKO = new unsigned long long[numMatches];
        unsigned long long numOK=0,numKO=0;

        int cl=ptr[tSymbolic->ruleSize-1];
        for(i=0;i<numMatches;i++) {
                if(instances[matched[i]]->instanceClass==cl) {
                        listOK[numOK++]=matched[i];
                } else {
                        listKO[numKO++]=matched[i];
                }
        }

        cleanTarget ct=cleanTargetOfRule(listOK,numOK,listKO,numKO);
        int changed=0;
        if(ct.maxNeg>0) {
                modif = 1;
                changed=1;
		ptr[tSymbolic->offsetAttribute[ct.maxAtt]+ct.maxValue]=0;
        }

        delete listOK;
        delete listKO;
        return changed;
}

unsigned long long classifier_gabil::computeMatches(unsigned char *ptr,unsigned long long *candidates,unsigned long long numCandidates,unsigned long long *matched)
{
        unsigned long long j;
        unsigned long long numMatches=0;
        instance **instances=is->getInstancesOfIteration();

        for(j=0;j<numCandidates;j++) {
                unsigned long long inst=candidates[j];
                if(ruleMatches(ptr,instances[inst])) {
                        matched[numMatches++]=inst;
                }
        }

        return numMatches;
}

unsigned long long classifier_gabil::updateMatches(unsigned long long *candidates,unsigned long long numCand,unsigned long long *matched,unsigned long long numMatched)
{
        unsigned long long numLeft=0;
        unsigned long long i;
        unsigned long long *candidatesLeft = new unsigned long long[numCand-numMatched];
        unsigned long long indexMatched=0;
        unsigned long long currMatched=matched[0];

        for(i=0;i<numCand;i++) {
                if(candidates[i]==currMatched) {
                        indexMatched++;
                        currMatched=matched[indexMatched];
                } else {
                        candidatesLeft[numLeft++]=candidates[i];
                }
        }

        for(i=0;i<numLeft;i++) {
                candidates[i]=candidatesLeft[i];
        }

        delete candidatesLeft;

        return numLeft;
}

unsigned char *classifier_gabil::ruleSplitting(unsigned char *ptr,unsigned long long *matched,unsigned long long numInst)
{
        unsigned long long i,j,k,l,m,n,o,p,q;
        instance **instances=is->getInstancesOfIteration();

        unsigned long long *listOK = new unsigned long long[numInst];
        unsigned long long *listKO = new unsigned long long[numInst];
        unsigned long long numOK=0,numKO=0;

        int cl=ptr[tSymbolic->ruleSize-1];
        for(i=0;i<numInst;i++) {
                if(instances[matched[i]]->instanceClass==cl) {
                        listOK[numOK++]=matched[i];
                } else {
                        listKO[numKO++]=matched[i];
                }
        }

        if(numKO==0) return NULL;
        activationsAtt *act=computeActivationStats(listOK,numOK,listKO,numKO);

        unsigned long long maxNeg=0;
        int splittedAtt;
        int splittedValue;
        int valueOrRest;
        JVector<int>conflicting;
        int negAtt;
        int negValue;

        for(i=0;i<tGlobals->numAttributesMC;i++) {
                int numConflictingValues=0;
                int targetConflicting[act[i].numValues];
                for(j=0;j<act[i].numValues;j++) {
                        if(act[i].actValues[j][0]>0 && act[i].actValues[j][1]>0) {
                                targetConflicting[numConflictingValues++]=j;
                        }
                }

                if(numConflictingValues>1) {
                        attInstances att=computeInstOfAtt(i,listOK,numOK,listKO,numKO);
                        for(l=0;l<numConflictingValues;l++) {
                                int target=targetConflicting[l];

                                unsigned long long numOK1=att.numInst[target][0];
                                unsigned long long numKO1=att.numInst[target][1];
                                unsigned long long numOK2=numOK-numOK1;
                                unsigned long long numKO2=numKO-numKO1;
                                unsigned long long *listOK2 = new unsigned long long[numOK2];
                                unsigned long long *listKO2 = new unsigned long long[numKO2];
                                unsigned long long countOK=0;
                                unsigned long long countKO=0;
                                for(m=0;m<act[i].numValues;m++) {
                                        if(m!=target) {
                                                for(n=0;n<att.numInst[m][0];n++) {
                                                        listOK2[countOK++]=att.valueInstances[m][0][n];
                                                }
                                                for(n=0;n<att.numInst[m][1];n++) {
                                                        listKO2[countKO++]=att.valueInstances[m][1][n];
                                                }
                                        }
                                }

                                cleanTarget ct=cleanTargetOfRule(att.valueInstances[target][0]
                                        ,numOK1,att.valueInstances[target][1],numKO1);
                                if(ct.maxNeg>maxNeg) {
                                        maxNeg=ct.maxNeg;
                                        splittedAtt=i;
                                        splittedValue=target;
                                        valueOrRest=1;
                                        conflicting.removeAllElements();
                                        for(q=0;q<numConflictingValues;q++) {
                                                if(targetConflicting[q]!=target) {
                                                        conflicting.addElement(targetConflicting[q]);
                                                }
                                        }
                                        negAtt=ct.maxAtt;
                                        negValue=ct.maxValue;
                                }

                                ct=cleanTargetOfRule(listOK2,numOK2,listKO2,numKO2);
                                if(ct.maxNeg>maxNeg) {
                                        maxNeg=ct.maxNeg;
                                        splittedAtt=i;
                                        splittedValue=target;
                                        valueOrRest=0;
                                        conflicting.removeAllElements();
                                        for(q=0;q<numConflictingValues;q++) {
                                                if(targetConflicting[q]!=target) {
                                                        conflicting.addElement(targetConflicting[q]);
                                                }
                                        }
                                        negAtt=ct.maxAtt;
                                        negValue=ct.maxValue;
                                }

                                delete listOK2;
                                delete listKO2;
                        }
                        removeInstOfAtt(att);
                }
        }

        unsigned char *newRule=NULL;
        if(maxNeg>0) {
                newRule = new unsigned char[tSymbolic->ruleSize];
                bcopy(ptr,newRule,sizeof(unsigned char)*tSymbolic->ruleSize);

                for(i=0;i<conflicting.size();i++) {
			ptr[tSymbolic->offsetAttribute[splittedAtt]+conflicting[i]]=0;
                }
		newRule[tSymbolic->offsetAttribute[splittedAtt]+splittedValue]=0;

                if(valueOrRest) {
			ptr[tSymbolic->offsetAttribute[negAtt]+negValue]=0;
                } else {
			newRule[tSymbolic->offsetAttribute[negAtt]+negValue]=0;
                }
                modif = 1;
        }

        removeActivationStats(act);
        delete listOK;
        delete listKO;
        return newRule;
}

attInstances classifier_gabil::computeInstOfAtt(int attribute,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO)
{
        unsigned long long numInst=numOK+numKO;
        instance **instances=is->getInstancesOfIteration();
        unsigned long long i,j,k;
        int numValues=tSymbolic->sizeAttribute[attribute];

        attInstances att;
        att.numValues=numValues;
        att.valueInstances = new unsigned long long **[numValues];
        att.numInst = new unsigned long long *[numValues];
        for(i=0;i<numValues;i++) {
                att.numInst[i]= new unsigned long long[2];
                att.numInst[i][0]=0;
                att.numInst[i][1]=0;
                att.valueInstances[i]= new unsigned long long *[2];
                att.valueInstances[i][0] = new unsigned long long[numInst];
                att.valueInstances[i][1] = new unsigned long long[numInst];
        }

        for(i=0;i<numOK;i++) {
                unsigned long long inst=instOK[i];
                int value=instances[inst]->nominalValues[attribute];
                att.valueInstances[value][0][att.numInst[value][0]++]=inst;
        }
        for(i=0;i<numKO;i++) {
                unsigned long long inst=instKO[i];
                int value=instances[inst]->nominalValues[attribute];
                att.valueInstances[value][1][att.numInst[value][1]++]=inst;
        }

        return att;
}

void classifier_gabil::removeInstOfAtt(attInstances att)
{
        unsigned long long o,p;

        for(o=0;o<att.numValues;o++) {
                for(p=0;p<2;p++) {
                        delete att.valueInstances[o][p];
                }
                delete att.valueInstances[o];
                delete att.numInst[o];
        }
        delete att.valueInstances;
        delete att.numInst;
}

cleanTarget classifier_gabil::cleanTargetOfRule(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO)
{
        instance **instances=is->getInstancesOfIteration();
        unsigned long long i,j,k;

        int numAttributes=tGlobals->numAttributesMC;
        unsigned long long *countKO[numAttributes];
        unsigned long long *countOK[numAttributes];
        int candidateAtt[numAttributes];
        int countCandidateValues[numAttributes];
        cleanTarget ct;
        ct.maxValue=-1;
        ct.maxAtt=-1;
        ct.maxNeg=0;

        int numCandidateAtts=numAttributes;
        for(i=0;i<numAttributes;i++) {
                candidateAtt[i]=i;
		int numValues=tSymbolic->sizeAttribute[i];
		countOK[i]= new unsigned long long[numValues];
		countKO[i]= new unsigned long long[numValues];
                countCandidateValues[i]=numValues;
                for(j=0;j<numValues;j++) {
                        countOK[i][j]=countKO[i][j]=0;
                }
        }

        for(i=0;i<numOK && numCandidateAtts>0;i++) {
                instance *ins=instances[instOK[i]];
                int indexAtt=0;
                while(indexAtt<numCandidateAtts) {
                        register int att=candidateAtt[indexAtt];
                        register int value=ins->nominalValues[att];
                        if(countOK[att][value]==0) {
                                countCandidateValues[att]--;
                                if(countCandidateValues[att]==0) {
                                        candidateAtt[indexAtt]=candidateAtt[numCandidateAtts-1];
                                        numCandidateAtts--;
                                } else {
                                        indexAtt++;
                                }
                        } else {
                                indexAtt++;
                        }
                        countOK[att][value]++;
                }
        }

        if(numCandidateAtts==0) {
	        for(i=0;i<numAttributes;i++) {
			delete countOK[i];
			delete countKO[i];
		}
		return ct;
	}

        for(i=0;i<numKO;i++) {
                instance *ins=instances[instKO[i]];
                for(j=0;j<numCandidateAtts;j++) {
                        register int att=candidateAtt[j];
                        register int value=ins->nominalValues[att];
                        countKO[att][value]++;
                }
        }

        for(j=0;j<numCandidateAtts;j++) {
                register int att=candidateAtt[j];
		int numValues=tSymbolic->sizeAttribute[att];
                for(i=0;i<numValues;i++) {
                        if(countOK[att][i]==0 && countKO[att][i]>ct.maxNeg) {
                                ct.maxAtt=att;
                                ct.maxValue=i;
                                ct.maxNeg=countKO[att][i];
                        }
                }
        }

        for(i=0;i<numAttributes;i++) {
		delete countOK[i];
		delete countKO[i];
	}

        return ct;
}

activationsAtt *classifier_gabil::computeActivationStats(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO)
{
        unsigned long long i,j;
        instance **instances=is->getInstancesOfIteration();

        activationsAtt *act=new activationsAtt[tGlobals->numAttributesMC];
        for(i=0;i<tGlobals->numAttributesMC;i++) {
                int numValues=tSymbolic->sizeAttribute[i];
                act[i].numValues=numValues;
                act[i].actValues=new unsigned long long *[numValues];
                for(j=0;j<numValues;j++) {
                        act[i].actValues[j]=new unsigned long long[2];
                        act[i].actValues[j][0]=act[i].actValues[j][1]=0;
                }
        }


        for(j=0;j<numOK;j++) {
                instance *ins=instances[instOK[j]];
                for(i=0;i<tGlobals->numAttributesMC;i++) {
                        act[i].actValues[ins->nominalValues[i]][0]++;
                }
        }
        for(j=0;j<numKO;j++) {
                instance *ins=instances[instKO[j]];
                for(i=0;i<tGlobals->numAttributesMC;i++) {
                        act[i].actValues[ins->nominalValues[i]][1]++;
                }
        }

        return act;
}

void classifier_gabil::removeActivationStats(activationsAtt *act)
{
        int i,j;

        for(i=0;i<tGlobals->numAttributesMC;i++) {
                for(j=0;j<act[i].numValues;j++) {
                        delete act[i].actValues[j];
                }
                delete act[i].actValues;
        }
        delete act;
}

void classifier_gabil::generalizeRule(unsigned char *rule,unsigned long long *candidates,unsigned long long numCand,unsigned long long *newInstances,unsigned long long &numNew)
{
        unsigned long long i,j,k;
        int len=tSymbolic->ruleSize-1;
        unsigned long long countPos[len];
        unsigned long long countNeg[len];
        JVector<unsigned long long> *candInst[len];
        instance **instances=is->getInstancesOfIteration();
        int cl=rule[tSymbolic->ruleSize-1];

        for(i=0;i<len;i++) {
                countPos[i]=countNeg[i]=0;
                candInst[i]=NULL;
        }

        for(i=0;i<numCand;i++) {
                unsigned long long posInst=candidates[i];
                instance *ins=instances[posInst];
                int numMatches=0;
                int whichPos;
                for(j=0;j<tGlobals->numAttributesMC;j++) {
                        int value=ins->nominalValues[j];
                        int pos=tSymbolic->offsetAttribute[j]+value;
                        if(rule[pos]==0) {
                                numMatches++;
                                whichPos=pos;
                                if(numMatches>=2) break;
                        }
                }

                if(numMatches==1) {
                        if(ins->instanceClass==cl) {
                                countPos[whichPos]++;
                                if(countNeg[whichPos]==0) {
                                        if(candInst[whichPos]==NULL) {
                                                candInst[whichPos] = new JVector<unsigned long long>((int)(numCand*0.1));
                                        }
                                        candInst[whichPos]->addElement(posInst);
                                }
                        } else {
                                countNeg[whichPos]++;
                                if(candInst[whichPos]) {
                                        delete candInst[whichPos];
                                        candInst[whichPos]=NULL;
                                }
                        }
                }
        }

        unsigned long long max=0;
        int bestPos;
        for(i=0;i<len;i++) {
                if(countNeg[i]==0 && countPos[i]>max) {
                        bestPos=i;
                        max=countPos[i];
                }
        }

        numNew=max;
        if(max>0) {
		rule[bestPos]=1;
                modif = 1;

                for(i=0;i<max;i++) {
                        newInstances[i]=candInst[bestPos]->elementAt(i);
                }
        }

        for(i=0;i<len;i++) {
                if(candInst[i]) delete candInst[i];
        }
}

void classifier_gabil::doLocalSearch()
{
        unsigned long long numInstances = is->getNumInstancesOfIteration();
        unsigned long long *remainingInstances = new unsigned long long[numInstances];
        unsigned long long *matched = new unsigned long long[numInstances];
        unsigned long long i;

        for(i=0;i<numInstances;i++) {
                remainingInstances[i]=i;
        }

        int index=0;
        int lastSplitted=0;
        while(index<numRules) {
                unsigned char *ptr=&chromosome[index*tSymbolic->ruleSize];

                unsigned long long numMatches = computeMatches(ptr,remainingInstances,numInstances,matched);
                if(tGlobals->doRuleCleaning) {
                        if(ruleCleaning(ptr,matched,numMatches)) {
                                numMatches = computeMatches(ptr,remainingInstances,numInstances,matched);
                        }
                }

                if(tGlobals->doRuleSplitting) {
                        if(lastSplitted) {
                                lastSplitted=0;
                        } else {
                                unsigned char *newRule=ruleSplitting(ptr,matched,numMatches);
                                if(newRule!=NULL) {
                                        unsigned char *newChromosome = new unsigned char[tSymbolic->ruleSize*(numRules+1)];
                                        bcopy(chromosome,newChromosome,(index+1)*tSymbolic->ruleSize*sizeof(unsigned char));
                                        bcopy(newRule,&newChromosome[(index+1)*tSymbolic->ruleSize],tSymbolic->ruleSize*sizeof(unsigned char));
                                        bcopy(&chromosome[(index+1)*tSymbolic->ruleSize],&newChromosome[(index+2)*tSymbolic->ruleSize],(numRules-index-1)*tSymbolic->ruleSize*sizeof(unsigned char));
                                        delete chromosome;
                                        chromosome=newChromosome;
                                        numRules++;
                                        length=numRules*tSymbolic->ruleSize;

                                        ptr=&chromosome[index*tSymbolic->ruleSize];
                                        numMatches = computeMatches(ptr,remainingInstances,numInstances,matched);
                                        delete newRule;
                                        lastSplitted=1;
                                }
                        }
                }

                numInstances=updateMatches(remainingInstances,numInstances,matched,numMatches);
                if(numInstances==0) {
                        break;
                }

                if(tGlobals->doRuleGeneralizing) {
                        generalizeRule(ptr,remainingInstances,numInstances,matched,numMatches);
                        if(numMatches>0) {
                                numInstances=updateMatches(remainingInstances,numInstances,matched,numMatches);
                                if(numInstances==0) {
                                        break;
                                }
                        }
                }

                index++;
        }

        delete remainingInstances;
        delete matched;
}

void classifier_gabil::findOrder(JVector<matchProfileAgent *>&profiles
	,JVector<unsigned char *>&ptrs,JVector<int> &ruleOrder) 
{
	unsigned long long numInst=is->getNumInstancesOfIteration();
	unsigned long long i,j;

	ruleOrderAgent *best=NULL;
	for(j=0;j<tCross->repetitionsRuleOrdering;j++) {
		int totalRules=profiles.size();
		Sampling samp(totalRules);
		ruleOrderAgent *roa = new ruleOrderAgent(defaultClass,numInst,totalRules);

		for(i=0;i<totalRules;i++) {
			int rule=samp.getSample();
			roa->insertRule(rule,profiles[rule]);
		}

                if(tGlobals->doRuleCleaning) {
                        int num=roa->getNumRules()-1;
                        for(i=0;i<num;i++) {
                                matchProfileAgent *mpa=roa->generateActualMPA(i);
                                if(cleanRule(ptrs[roa->getRule(i)],mpa)) {
                                        roa->removeMatchesOfRule(i,mpa);
                                }
                                delete mpa;
                        }
                }

                if(tGlobals->doRuleGeneralizing) {
                        unsigned long long numCand;
                        unsigned long long candidates[numInst];
                        for(i=0;i<roa->getNumRules()-1;i++) {
                                int whichRule=roa->getRule(i);
                                unsigned char *ptr=ptrs[whichRule];
                                roa->getPossibleInstancesOfRule(i,candidates,numCand);
                                unsigned long long newInstances[numCand];
                                unsigned long long numNew;
                                generalizeRule(ptr,candidates,numCand,newInstances,numNew);
                                if(numNew>0) {
                                        delete profiles[whichRule];
                                        matchProfileAgent *mpa=evaluateRuleCX(ptr);
                                        profiles[whichRule]=mpa;
                                        roa->addPositiveExamples(i,newInstances,numNew,mpa);
                                }
                        }
                }


                if(tGlobals->doRuleSplitting) {
                        for(i=0;i<roa->getNumRules()-1;i++) {
                                matchProfileAgent *mpa=roa->generateActualMPA(i);
                                splittedRule spl;
                                if(editRule(ptrs[roa->getRule(i)],mpa,spl)) {
                                        roa->removeRule(i);
                                        profiles.addElement(spl.mpa1);
                                        profiles.addElement(spl.mpa2);
                                        ptrs.addElement(spl.ptr1);
                                        ptrs.addElement(spl.ptr2);
                                        roa->insertRule(profiles.size()-2,profiles[profiles.size()-2]);
                                        roa->insertRule(profiles.size()-1,profiles[profiles.size()-1]);
                                }
                                delete mpa;
                        }
                }


		roa->refineRuleSet();

		if(best==NULL) {
			best=roa;
		} else {
			double accOld=best->getAccuracy();
			double accNew=roa->getAccuracy();
			if(accNew>accOld) {
				delete best;
				best=roa;
			} else if(accNew==accOld 
				&& roa->getNumRules()<best->getNumRules()) {
				delete best;
				best=roa;
			} else {
				delete roa;
			}
		}
	}

	best->copyOrder(ruleOrder);
	if(ruleOrder.size()==0) {
		int pos=rnd(0,profiles.size()-1);
		ruleOrder.addElement(pos);
	}
	delete best;
}

matchProfileAgent *classifier_gabil::evaluateRuleCX(unsigned char *rule)
{
	unsigned long long i;
	int cl = rule[tSymbolic->ruleSize-1];
	unsigned long long max = is->getNumInstancesOfIteration();
	instance **instances=is->getInstancesOfIteration();

	matchProfileAgent *mpa=new matchProfileAgent(max,cl);

	for (i = 0; i < max; i++) {
		if (ruleMatches(rule, instances[i])) {
			if (instances[i]->instanceClass == cl) {
				mpa->addOK(i);
			} else {
				mpa->addKO(i);
			}
		}
	}
	mpa->generateProfiles();

	return mpa;
}

void classifier_gabil::instOfAttAndValue(int attribute,int value,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO,unsigned long long *list)
{

	instance **instances=is->getInstancesOfIteration();
	unsigned long long i;

	unsigned long long num=0;

	for(i=0;i<numOK;i++) {
		int instValue=instances[instOK[i]]->nominalValues[attribute];
		if(instValue==value) {
			list[num++]=instOK[i];
		}
	}
	for(i=0;i<numKO;i++) {
		int instValue=instances[instKO[i]]->nominalValues[attribute];
		if(instValue==value) {
			list[num++]=instKO[i];
		}
	}
}

int classifier_gabil::editRule(unsigned char *ptr,matchProfileAgent *mpa, splittedRule &spl)
{
	unsigned long long i,j,k,l,m,n,o,p,q;

	if(mpa->numKO==0) return 0;
	activationsAtt *act=computeActivationStats(mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO);

	unsigned long long maxNeg=0;
	int splittedAtt;
	int splittedValue;
	int valueOrRest;
	JVector<int>conflicting;
	int negAtt;
	int negValue;

	for(i=0;i<tGlobals->numAttributesMC;i++) {
		int numConflictingValues=0;	
		int targetConflicting[act[i].numValues];
		for(j=0;j<act[i].numValues;j++) {
			if(act[i].actValues[j][0]>0 && act[i].actValues[j][1]>0) {
				targetConflicting[numConflictingValues++]=j;
			}
		}

		if(numConflictingValues>1) {
			attInstances att=computeInstOfAtt(i,mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO);
			for(l=0;l<numConflictingValues;l++) {
				int target=targetConflicting[l];

				unsigned long long numOK1=att.numInst[target][0];
				unsigned long long numKO1=att.numInst[target][1];
				unsigned long long numOK2=mpa->numOK-numOK1;
				unsigned long long numKO2=mpa->numKO-numKO1;
				unsigned long long *listOK2 = new unsigned long long[numOK2];
				unsigned long long *listKO2 = new unsigned long long[numKO2];
				unsigned long long countOK=0;
				unsigned long long countKO=0;
				for(m=0;m<act[i].numValues;m++) {
					if(m!=target) {
						for(n=0;n<att.numInst[m][0];n++) {
							listOK2[countOK++]=att.valueInstances[m][0][n];
						}
						for(n=0;n<att.numInst[m][1];n++) {
							listKO2[countKO++]=att.valueInstances[m][1][n];
						}
					}
				}

                                cleanTarget ct=cleanTargetOfRule(att.valueInstances[target][0]
                                        ,numOK1,att.valueInstances[target][1],numKO1);
                                if(ct.maxNeg>maxNeg) {
                                        maxNeg=ct.maxNeg;
                                        splittedAtt=i;
                                        splittedValue=target;
                                        valueOrRest=1;
                                        conflicting.removeAllElements();
                                        for(q=0;q<numConflictingValues;q++) {
                                                if(targetConflicting[q]!=target) {
                                                        conflicting.addElement(targetConflicting[q]);
                                                }
                                        }
                                        negAtt=ct.maxAtt;
                                        negValue=ct.maxValue;
                                }

                                ct=cleanTargetOfRule(listOK2,numOK2,listKO2,numKO2);
                                if(ct.maxNeg>maxNeg) {
                                        maxNeg=ct.maxNeg;
                                        splittedAtt=i;
                                        splittedValue=target;
                                        valueOrRest=0;
                                        conflicting.removeAllElements();
                                        for(q=0;q<numConflictingValues;q++) {
                                                if(targetConflicting[q]!=target) {
                                                        conflicting.addElement(targetConflicting[q]);
                                                }
                                        }
                                        negAtt=ct.maxAtt;
                                        negValue=ct.maxValue;
                                }



				delete listOK2;
				delete listKO2;
			}
			removeInstOfAtt(att);
		}
	}

	int splitted=0;
	if(maxNeg>0) {
		splitted=1;
		unsigned char *ptr1= new unsigned char[tSymbolic->ruleSize];
		bcopy(ptr,ptr1,sizeof(unsigned char)*tSymbolic->ruleSize);
		unsigned char *ptr2= new unsigned char[tSymbolic->ruleSize];
		bcopy(ptr,ptr2,sizeof(unsigned char)*tSymbolic->ruleSize);

		for(i=0;i<conflicting.size();i++) {
			ptr1[tSymbolic->offsetAttribute[splittedAtt]+conflicting[i]]=0;
		}
		ptr2[tSymbolic->offsetAttribute[splittedAtt]+splittedValue]=0;

		if(valueOrRest) {
			ptr1[tSymbolic->offsetAttribute[negAtt]+negValue]=0;
		} else {
			ptr2[tSymbolic->offsetAttribute[negAtt]+negValue]=0;
		}
		spl.ptr1=ptr1;
		spl.ptr2=ptr2;
		spl.mpa1=evaluateRuleCX(ptr1);
		spl.mpa2=evaluateRuleCX(ptr2);
	}

	removeActivationStats(act);
	return splitted;
}

int classifier_gabil::cleanRule(unsigned char *ptr,matchProfileAgent *mpa)
{
        unsigned long long i,j;

        if(mpa->numKO==0) return 0;

        cleanTarget ct=cleanTargetOfRule(mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO);
        int changed=0;
        if(ct.maxNeg>0) {
                changed=1;

		ptr[tSymbolic->offsetAttribute[ct.maxAtt]+ct.maxValue]=0;
                unsigned long long removedExamples[ct.maxNeg];
                instOfAttAndValue(ct.maxAtt,ct.maxValue,mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO,removedExamples);
                mpa->removeMatched(removedExamples,ct.maxNeg);
        }

        return changed;
}


void classifier_gabil::crossover_smart(classifier **parents, int num) 
{
	int i, j, k;

	modif = 1;
	if(chromosome) delete chromosome;

	JVector<unsigned char *>ptrs;
	JVector<matchProfileAgent *>profiles;
	int ruleCount=0;

	for(k=0;k<num;k++) {
		classifier_gabil *parent=(classifier_gabil *)parents[k];
		unsigned char *ptr=parent->chromosome;
		for(i=0;i<parent->numRules;ptr+=tSymbolic->ruleSize,i++) {
			unsigned char *newRule= new unsigned char[tSymbolic->ruleSize];
			bcopy(ptr,newRule,sizeof(unsigned char)*tSymbolic->ruleSize);
			matchProfileAgent *mpa = evaluateRuleCX(newRule);
			ptrs.addElement(newRule);
			profiles.addElement(mpa);
		}
	}

	JVector<int> ruleOrder;
	findOrder(profiles,ptrs,ruleOrder);

	numRules = ruleOrder.size(); 
	length = numRules * tSymbolic->ruleSize;
	chromosome = new unsigned char[length];
	for (i = 0; i < numRules; i++) {
		bcopy(ptrs[ruleOrder[i]], &chromosome[i * tSymbolic->ruleSize],
		      sizeof(unsigned char) * tSymbolic->ruleSize);
	}

	for(i=0;i<profiles.size();i++) {
		delete profiles[i];
	}
	for(i=0;i<ptrs.size();i++) {
		delete ptrs[i];
	}
}


