#ifndef _CLASSIFIER_GABIL_

#define _CLASSIFIER_GABIL_

#include "classifier.h"
#include "agentPerformanceTraining.h"
#include "timerGlobals.h"
#include "timerEvolutionStats.h"
#include "timerSymbolicKR.h"
#include "matchProfileAgent.h"


extern timerGlobals *tGlobals;
extern timerEvolutionStats *tEvolStats;
extern timerSymbolicKR *tSymbolic;

typedef struct {
        unsigned char *ptr1;
        matchProfileAgent *mpa1;
        unsigned char *ptr2;
        matchProfileAgent *mpa2;
} splittedRule;

typedef struct {
        int numValues;
        unsigned long long ***valueInstances;
        unsigned long long **numInst;
} attInstances;

typedef struct {
	int maxAtt;
	int maxValue;
	unsigned long long maxNeg;
} cleanTarget;



typedef struct {
        int numValues;
        unsigned long long **actValues;
} activationsAtt;




class classifier_gabil: public classifier   {
	unsigned char *chromosome;
	int numRules;

	void dumpPhenotypeRule(unsigned char *ptr);
	void informedCrossover(unsigned char *father1,unsigned char *father2
		,unsigned char *son1,unsigned char *son2);
	void crossover_1px(classifier_gabil *in1
		,classifier_gabil *in2
		,classifier_gabil *out1
		,classifier_gabil *out2);
	void crossover_2px(classifier_gabil *in1,classifier_gabil *in2
		, classifier_gabil *out1
		,classifier_gabil *out2);
	unsigned char getGene(short int rule,short int gene,short int value); 
	void setGene(short int rule,short int gene,short int value
		,unsigned char nfo);
	void gabilRuleCover(unsigned char *rule,instance *ins,double prob);

	void initializeChromosome(void);

        int ruleCleaning(unsigned char *ptr,unsigned long long *instances,unsigned long long numInstances);
        unsigned long long updateMatches(unsigned long long *instanceMatched,unsigned long long totalInstances,unsigned long long *instanceMatchedTemp,unsigned long long numMatched);
        unsigned long long computeMatches(unsigned char *ptr,unsigned long long *instanceMatched,unsigned long long totalInstances,unsigned long long *instanceMatchedTemp);
        unsigned char *ruleSplitting(unsigned char *ptr,unsigned long long *instances,unsigned long long numInstances);
        cleanTarget cleanTargetOfRule(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO);
        void instOfAttAndValue(int attribute,int value,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO,unsigned long long *list);
        attInstances computeInstOfAtt(int attribute,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO);
        void removeInstOfAtt(attInstances att);
        activationsAtt *computeActivationStats(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO);
        void removeActivationStats(activationsAtt *act);
	void findOrder(JVector<matchProfileAgent *>&profiles
		,JVector<unsigned char *>&ptrs,JVector<int> &ruleOrder);
	matchProfileAgent *evaluateRuleCX(unsigned char *rule);
        int cleanRule(unsigned char *ptr,matchProfileAgent *mpa);
        int editRule(unsigned char *ptr,matchProfileAgent *mpa,splittedRule &spl);
        void generalizeRule(unsigned char *rule,unsigned long long *candidates,unsigned long long numCand,unsigned long long *newInst,unsigned long long &numNew);

	
	inline int ruleMatches(unsigned char *rule,instance *ins) 
	{
		int j;

		tEvolStats->numExampleEvaluations++;
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (rule[tSymbolic->offsetAttribute[j]
				   + ins->nominalValues[j]] == 0) {
				return 0;
			}
		}
		return 1;
	}

	inline int ruleMatchesMissing(unsigned char *rule,instance *ins) 
	{
		int j;

		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (!ins->isMissing(j) 
				&& rule[tSymbolic->offsetAttribute[j]
				   + ins->nominalValues[j]] == 0) {
				return 0;
			}
		}
		return 1;
	}

	inline virtual void dumpGenotype(char *string) 
	{
		unsigned char *ptr=chromosome;
		int i,j,k;

		string[0]=0;
		for (i = 0; i < numRules; i++) {
			char tmp[tSymbolic->ruleSize+tGlobals->numAttributes];
			int countPos=0;
			int dead=0;
			for (j=0;j<tGlobals->numAttributesMC && !dead;j++) {
				int attDead=1;
				for(k=tSymbolic->offsetAttribute[j];k<tSymbolic->offsetAttribute[j]+tSymbolic->sizeAttribute[j];k++) {
					if(ptr[k]==1) attDead=0;
					tmp[countPos++]=ptr[k]+'0';
				}
				tmp[countPos++]='|';
				if(attDead==1) dead=1;
			}
	
			if(!dead) {
				tmp[countPos++]=ptr[tSymbolic->ruleSize-1]+'0';
				tmp[countPos]=0;
				strcat(string,tmp);
				strcat(string,"\n");
			}

			ptr+=tSymbolic->ruleSize;
		}
	}



public:
	classifier_gabil();
	classifier_gabil(const classifier_gabil &orig,int son=0);
	~classifier_gabil();

	inline int getClass(int rule) {
		if(rule==numRules) return defaultClass;
		return (int) chromosome[rule * tSymbolic->ruleSize 
			+ tSymbolic->ruleSize - 1];
	}

	inline int getNumClassifiers() {
		int nC=numRules;
		if(tGlobals->defaultClassPolicy!=DISABLED) nC++;
		return nC;
	}

	void pruneClassifiers(int *classifiersToPrune,int numPruned);

	int classify(instance *ins);
	double computateTheoryLength(agentPerformanceTraining *ap);
	
	inline int hasSmartCrossover() {return 1;}
	void crossover_smart(classifier **parents, int num);

	inline double getHierarchicalSelectionLength()
	{
		if(tHierar->useMDL) return getTheoryLength();
		return aliveClassifiers;
	}

	inline int getNiche() {
		if(tGlobals->defaultClassPolicy!=AUTO) return 0;
		return getClass(getNumClassifiers()-1);
	}
	inline int getNumNiches() {
		if(tGlobals->defaultClassPolicy!=AUTO) return 1;
		return ai.getNumClasses();
	}

	void crossover(classifier *,classifier *,classifier *);
	void mutation();
	void dumpPhenotype(char *string);
	inline int numSpecialStages(){return 0;}
	void doSpecialStage(int stage){}

	//void classifyTraining(instanceSet *is,agentPerformanceTraining *ap);
	//int canDoSmartClassAsignment() {return 1;}
	//int canDoTrainAndClean() {return 1;}

        void doLocalSearch();
        int hasLocalSearch(){return 1;}


};

#endif

