#include "classifier_gabil_packed.h"
#include "random.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "attributesInfo.h"
#include "timerSymbolicKR.h"
#include "timerMutation.h"
#include "timerHierar.h"
#include "timerCrossover.h"
#include "timerMDL.h"
#include "timerGlobals.h"
#include "instanceSet.h"
#include "ruleOrderAgent.h"
#include "sampling.h"
#include "messageBuffer.h"

extern messageBuffer mb;
extern attributesInfo ai;
extern timerSymbolicKR *tSymbolic;
extern timerHierar *tHierar;
extern timerGlobals *tGlobals;
extern timerCrossover *tCross;
extern timerMutation *tMut;
extern timerMDL *tMDL;
extern Random rnd;
extern instanceSet *is;
extern int lastIteration;
extern int nodeRank;

double classifier_gabil_packed::computateTheoryLength(agentPerformanceTraining *ap)
{
	int i, j, k, l, countAtt;
	aligned_uint *ptr = chromosome;
	theoryLength = 0;

	for (i = 0; i < numRules; i++) {
		if(ap->getActivationsOfClassifier(i))
		for(j=0,countAtt=0;j<tSymbolic->recodeBlocks;j++) {
			aligned_uint tmp=ptr[j];
			for(k=0;k<ATT_PER_BLOCK && countAtt<tGlobals->numAttributesMC;k++,countAtt++) {
				int numFalse = 0;
				for (l = 0; l < 2; l++) {
					if (!GET_COND(tmp,k,l)) numFalse++;
				}
				theoryLength += 2 + numFalse;
			} 
		}
		ptr+=tSymbolic->ruleSize;
	}
	if(tGlobals->defaultClassPolicy!=DISABLED)
		//if(ap->getActivationsOfClassifier(numRules))
			theoryLength+=tSymbolic->defaultRuleTheoryLength;

	return theoryLength;
}

classifier_gabil_packed::classifier_gabil_packed()
{
	int i;

	numRules =
	    rnd(tGlobals->minClassifiersInit, tGlobals->maxClassifiersInit);
	length = tSymbolic->ruleSize * numRules;
	chromosome = new aligned_uint[length];

	if(tGlobals->defaultClassPolicy==AUTO) {
		defaultClass=rnd(0,ai.getNumClasses()-1);
	} else  {
		defaultClass=tGlobals->defaultClass;
	}

	classifier_gabil_packed::initializeChromosome();
}

classifier_gabil_packed::
classifier_gabil_packed(const classifier_gabil_packed & orig, int son)
{
	*this = orig;

	if (!son) {
		chromosome = new aligned_uint[length];
		bcopy(orig.chromosome, chromosome,
		      length * sizeof(aligned_uint));
	} else {
		chromosome = NULL;
	}
}

classifier_gabil_packed::~classifier_gabil_packed()
{
	delete chromosome;
}

void classifier_gabil_packed::gabilRuleCover(aligned_uint *rule,instance *ins,double prob)
{
	int j, k, l, countAtt;

	aligned_uint *ptr=rule;
	for(j=0,countAtt=0;j<tSymbolic->recodeBlocks;j++) {
		aligned_uint tmp;
		for(k=0;k<ATT_PER_BLOCK;k++,countAtt++) {
			if(countAtt<tGlobals->numAttributesMC) {
				int value=-1;
				if(ins) value=ins->nominalValues[countAtt];
				for(l=0;l<2;l++){
					if(l!=value) {
						if(!rnd<prob) SET_COND(tmp,k,l,1);
						else SET_COND(tmp,k,l,0);
					} else {
						SET_COND(tmp,k,l,1);
					}
				}
			} else {
				SET_ATT(tmp,k,0);
			}
		}
		ptr[j]=tmp;	
	}

	if(ins) {
		ptr[j]=ins->instanceClass;
	} else {
		int cl;
		do {
			cl=rnd(0,ai.getNumClasses()-1);
		} while(tGlobals->defaultClassPolicy!=DISABLED 
			&& cl==defaultClass);
		ptr[j]=cl;
	}
}

void classifier_gabil_packed::initializeChromosome()
{
	int i;

	aligned_uint *ptr=chromosome;
	for(i=0;i<numRules;i++) {
		instance *ins=NULL;
		if(tGlobals->smartInit) {
			if(tGlobals->defaultClassPolicy!=DISABLED) {
				ins=is->getInstanceInit(defaultClass);
			} else {
				ins=is->getInstanceInit(ai.getNumClasses());
			}
		}

		gabilRuleCover(ptr,ins,tGlobals->probOne);
		ptr+=tSymbolic->ruleSize;
	}
}

void classifier_gabil_packed::setGene(short int rule, short int gene,
				short int value, unsigned char nfo)
{
	aligned_uint *ptr=&chromosome[rule*tSymbolic->ruleSize];
	if(gene==tGlobals->numAttributesMC) {
		ptr[tSymbolic->recodeBlocks]=nfo;
	} else {
		int blockNum=gene/ATT_PER_BLOCK;
		int idx=gene%ATT_PER_BLOCK;
		SET_COND(ptr[blockNum],idx,value,nfo);
	}
	modif = 1;
}

unsigned char classifier_gabil_packed::getGene(short int rule,
					 short int gene, short int value)
{
	aligned_uint *ptr=&chromosome[rule*tSymbolic->ruleSize];
	if(gene==tGlobals->numAttributesMC) {
		return (unsigned char)ptr[tSymbolic->recodeBlocks];
	} else {
		int blockNum=gene/ATT_PER_BLOCK;
		int idx=gene%ATT_PER_BLOCK;
		return GET_COND(ptr[blockNum],idx,value);
	}
}

void classifier_gabil_packed::crossover(classifier * in,
				   classifier * out1, classifier * out2)
{

	/*switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this, (classifier_gabil_packed *) in,
			      (classifier_gabil_packed *) out1,
			      (classifier_gabil_packed *) out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_gabil_packed *) in,
			      (classifier_gabil_packed *) out1,
			      (classifier_gabil_packed *) out2);
	}*/
	crossover_1px(this, (classifier_gabil_packed *) in,
	      (classifier_gabil_packed *) out1,
	      (classifier_gabil_packed *) out2);

}

void classifier_gabil_packed::mutation()
{
	int i;
	short int rule, attribute, value;

	tMut->mutationCounter++;
	modif = 1;

	rule = rnd(0,numRules-1);
	if(!tGlobals->smartClassAssignment && tGlobals->numClasses>1 
			&& !rnd<0.10) {
		attribute = tGlobals->numAttributesMC;
		value=0;
	} else {
		attribute=rnd(0,tGlobals->numAttributesMC-1);
		value=rnd(0,1);
	}

	if (attribute != tGlobals->numAttributesMC ) {
		if (getGene(rule, attribute, value) == 0)
			setGene(rule, attribute, value, 1);
		else
			setGene(rule, attribute, value, 0);
	} else {
		int oldValue=getGene(rule,attribute,value);
		int newValue;
		do {
			newValue = rnd(0, ai.getNumClasses()-1);
		} while(newValue==oldValue || tGlobals->defaultClassPolicy!=DISABLED && newValue==defaultClass);
		setGene(rule, attribute, value, newValue);
	}
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

/*void classifier_gabil_packed::crossover_2px(classifier_gabil_packed * in1,
				       classifier_gabil_packed * in2,
				       classifier_gabil_packed * out1,
				       classifier_gabil_packed * out2)
{
	int i1rule1, i1rule2, i2rule1, i2rule2;
	int cutPoint1, cutPoint2, sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	i1rule1 = rnd(0, in1->numRules - 1);
	i1rule2 = rnd(0, in1->numRules - 1);
	if (i1rule1 > i1rule2)
		swap(i1rule1, i1rule2);

	i2rule1 = rnd(0, in2->numRules - 1);
	i2rule2 = rnd(0, in2->numRules - 1);
	if (i2rule1 > i2rule2)
		swap(i2rule1, i2rule2);

	sizeOfSon1 =
	    i1rule1 + i2rule2 - i2rule1 + in1->numRules - i1rule2;
	sizeOfSon2 =
	    i2rule1 + i1rule2 - i1rule1 + in2->numRules - i2rule2;

	cutPoint1 = rnd(0, tSymbolic->ruleSize - 1);
	cutPoint2 = rnd(0, tSymbolic->ruleSize - 1);

	if (i1rule1 == i1rule2 || i2rule1 == i2rule2) {
		if (cutPoint1 > cutPoint2)
			swap(cutPoint1, cutPoint2);
	}

	out1->length = sizeOfSon1 * tSymbolic->ruleSize;
	out2->length = sizeOfSon2 * tSymbolic->ruleSize;
	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;

	out1->chromosome = new unsigned int[sizeOfSon1 * tSymbolic->ruleSize];
	out2->chromosome = new unsigned int[sizeOfSon2 * tSymbolic->ruleSize];

	bcopy(chromosome, out1->chromosome, i1rule1 * tSymbolic->ruleSize);
	bcopy(in2->chromosome, out2->chromosome,
	      i2rule1 * tSymbolic->ruleSize);

	bcopy(&in1->chromosome[i1rule1 * tSymbolic->ruleSize],
	      &out1->chromosome[i1rule1 * tSymbolic->ruleSize], cutPoint1);
	bcopy(&in2->chromosome[i2rule1 * tSymbolic->ruleSize],
	      &out2->chromosome[i2rule1 * tSymbolic->ruleSize], cutPoint1);

	bcopy(&in2->chromosome[i2rule1 * tSymbolic->ruleSize + cutPoint1],
	      &out1->chromosome[i1rule1 * tSymbolic->ruleSize + cutPoint1],
	      tSymbolic->ruleSize - cutPoint1);
	bcopy(&in1->chromosome[i1rule1 * tSymbolic->ruleSize + cutPoint1],
	      &out2->chromosome[i2rule1 * tSymbolic->ruleSize + cutPoint1],
	      tSymbolic->ruleSize - cutPoint1);

	if (i2rule1 != i2rule2)
		bcopy(&in2->
		      chromosome[(i2rule1 + 1) * tSymbolic->ruleSize],
		      &out1->chromosome[(i1rule1 + 1) *
				       tSymbolic->ruleSize],
		      (i2rule2 - (i2rule1 + 1)) * tSymbolic->ruleSize);
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[(i1rule1 + 1) * tSymbolic->ruleSize],
		      &out2->chromosome[(i2rule1 + 1) *
				       tSymbolic->ruleSize],
		      (i1rule2 - (i1rule1 + 1)) * tSymbolic->ruleSize);


	if (i2rule1 != i2rule2)
		bcopy(&in2->chromosome[i2rule2 * tSymbolic->ruleSize],
		      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
				       tSymbolic->ruleSize], cutPoint2);
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[i1rule2 * tSymbolic->ruleSize],
		      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
				       tSymbolic->ruleSize], cutPoint2);
	
	bcopy(&in1->chromosome[i1rule2 * tSymbolic->ruleSize + cutPoint2],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
			       tSymbolic->ruleSize + cutPoint2],
	      tSymbolic->ruleSize - cutPoint2);
	bcopy(&in2->chromosome[i2rule2 * tSymbolic->ruleSize + cutPoint2],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
			       tSymbolic->ruleSize + cutPoint2],
	      tSymbolic->ruleSize - cutPoint2);

	bcopy(&in1->chromosome[(i1rule2 + 1) * tSymbolic->ruleSize],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1 + 1) *
			       tSymbolic->ruleSize],
	      (in1->numRules - (i1rule2 + 1)) * tSymbolic->ruleSize);
	bcopy(&in2->chromosome[(i2rule2 + 1) * tSymbolic->ruleSize],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1 + 1) *
			       tSymbolic->ruleSize],
	      (in2->numRules - (i2rule2 + 1)) * tSymbolic->ruleSize);

}*/

void classifier_gabil_packed::dumpPhenotype(char *string)
{
	aligned_uint *ptr = chromosome;
	char temp[10000];
	char temp2[1000];
	int i, j, k, l;
	int countAtt;

	strcpy(string, "");
	for (i = 0; i < numRules; i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		for(j=0,countAtt=0;j<tSymbolic->recodeBlocks;j++) {
			aligned_uint tmp=ptr[j];
			for(k=0;k<ATT_PER_BLOCK && countAtt<tGlobals->numAttributesMC;k++,countAtt++) {
				sprintf(temp,"Att %s is "
					,ai.getAttributeName(countAtt)->cstr());
				int irr=1;
				for (l = 0; l < 2; l++) {
					if(GET_COND(tmp,k,l)) {
						sprintf(temp2,"%s,",ai.getNominalValue(countAtt,l)->cstr());
						strcat(temp,temp2);
					} else {
						irr=0;
					}
				}
	
				if(!irr) {
					if(temp[strlen(temp) - 1] == ',')
						temp[strlen(temp) - 1] = 0;
					strcat(string, temp);
					strcat(string, "|");
				}
			}
		}

		int cl=ptr[tSymbolic->recodeBlocks];
		sprintf(temp, "%s\n",
			ai.getNominalValue(tGlobals->numAttributesMC
				, cl)->cstr());
		strcat(string, temp);
		ptr += tSymbolic->ruleSize;
	}
	if(tGlobals->defaultClassPolicy!=DISABLED) {
		sprintf(temp, "%d:Default rule -> %s\n", i
			,ai.getNominalValue(tGlobals->numAttributesMC
				,defaultClass)->cstr());
		strcat(string, temp);
	}
	strcat(string, "\n");
}

void classifier_gabil_packed::crossover_1px(classifier_gabil_packed * in1,
				       classifier_gabil_packed * in2,
				       classifier_gabil_packed * out1,
				       classifier_gabil_packed * out2)
{
	out1->modif = out2->modif = 1;

	unsigned int attribute, rulep1, rulep2;
	int sizeOfSon1, sizeOfSon2;

	rulep1 = rnd(0, in1->numRules - 1);
	rulep2 = rnd(0, in2->numRules - 1);
	sizeOfSon1 = rulep1 + in2->numRules - rulep2;
	sizeOfSon2 = rulep2 + in1->numRules - rulep1;

	out1->length = sizeOfSon1 * tSymbolic->ruleSize;
	out2->length = sizeOfSon2 * tSymbolic->ruleSize;
	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;
	
	out1->chromosome = new aligned_uint[sizeOfSon1 * tSymbolic->ruleSize];
	out2->chromosome = new aligned_uint[sizeOfSon2 * tSymbolic->ruleSize];

	//Before cut point	
	bcopy(in1->chromosome, out1->chromosome,
	      rulep1 * tSymbolic->ruleSize*sizeof(aligned_uint));
	bcopy(in2->chromosome, out2->chromosome,
	      rulep2 * tSymbolic->ruleSize*sizeof(aligned_uint));

	int len = tGlobals->numAttributesMC*2;
	int cutPoint = rnd(0, len);
	int blocksBefore=cutPoint/(2*ATT_PER_BLOCK);
	int blockCutPoint=cutPoint%(2*ATT_PER_BLOCK);
	int blocksAfter=tSymbolic->recodeBlocks-blocksBefore-1;

	//Blocks Before cut point
	bcopy(&in1->chromosome[rulep1 * tSymbolic->ruleSize],
	      &out1->chromosome[rulep1 * tSymbolic->ruleSize], 
		blocksBefore*sizeof(aligned_uint));
	bcopy(&in2->chromosome[rulep2 * tSymbolic->ruleSize],
	      &out2->chromosome[rulep2 * tSymbolic->ruleSize], 
		blocksBefore*sizeof(aligned_uint));

	aligned_uint mask1=(aligned_uint)pow(2,blockCutPoint)-1;
	aligned_uint mask2=((aligned_uint)pow(2,sizeof(aligned_uint)*8-blockCutPoint)-1)<<blockCutPoint;
	aligned_uint p1=in1->chromosome[rulep1*tSymbolic->ruleSize+blocksBefore];
	aligned_uint p2=in2->chromosome[rulep2*tSymbolic->ruleSize+blocksBefore];
	aligned_uint s1=(p1&mask1)|(p2&mask2);
	aligned_uint s2=(p2&mask1)|(p1&mask2);

	out1->chromosome[rulep1*tSymbolic->ruleSize+blocksBefore]=s1;
	out2->chromosome[rulep2*tSymbolic->ruleSize+blocksBefore]=s2;
	
	//Blocks after cut point
	bcopy(&in1->chromosome[rulep1 * tSymbolic->ruleSize + blocksBefore+1],
	      &out2->chromosome[rulep2 * tSymbolic->ruleSize + blocksBefore+1],
	      blocksAfter*sizeof(aligned_uint));
	bcopy(&in2->chromosome[rulep2 * tSymbolic->ruleSize + blocksBefore+1],
	      &out1->chromosome[rulep1 * tSymbolic->ruleSize + blocksBefore+1],
	      blocksAfter*sizeof(aligned_uint));

	// The class gene
	out2->chromosome[rulep2 * tSymbolic->ruleSize+tSymbolic->recodeBlocks] =
		in1->chromosome[rulep1 * tSymbolic->ruleSize+tSymbolic->recodeBlocks];
	out1->chromosome[rulep1 * tSymbolic->ruleSize+tSymbolic->recodeBlocks] =
		in2->chromosome[rulep2 * tSymbolic->ruleSize+tSymbolic->recodeBlocks];
	
	//Rules after cut point
	bcopy(&in2->chromosome[(rulep2 + 1) * tSymbolic->ruleSize],
	      &out1->chromosome[(rulep1 + 1) * tSymbolic->ruleSize],
	      (in2->numRules - (rulep2 + 1)) * tSymbolic->ruleSize*sizeof(aligned_uint));

	bcopy(&in1->chromosome[(rulep1 + 1) * tSymbolic->ruleSize],
	      &out2->chromosome[(rulep2 + 1) * tSymbolic->ruleSize],
	      (in1->numRules - (rulep1 + 1)) * tSymbolic->ruleSize*sizeof(aligned_uint));

}

void classifier_gabil_packed::pruneClassifiers(int *classifiersToPrune,
					  int numPruned)
{
	aligned_uint *newChromosome;
	int i;

	if (numRules == 1 || numPruned==0) return;

	if(classifiersToPrune[numPruned-1]==numRules) numPruned--;

	newChromosome = new aligned_uint[(numRules - numPruned) *
			      tSymbolic->ruleSize];

	int countPruned = 0;
	aligned_uint *ptrOrig = chromosome;
	aligned_uint *ptrDest = newChromosome;
	for (i = 0; i < numRules; i++) {
		if (countPruned < numPruned) {
			if (i != classifiersToPrune[countPruned]) {
				bcopy(ptrOrig, ptrDest, tSymbolic->ruleSize*sizeof(aligned_uint));
				ptrDest += tSymbolic->ruleSize;
			} else {
				countPruned++;
			}
		} else {
			bcopy(ptrOrig, ptrDest, tSymbolic->ruleSize*sizeof(aligned_uint));
			ptrDest += tSymbolic->ruleSize;
		}
		ptrOrig += tSymbolic->ruleSize;
	}

	delete chromosome;
	chromosome = newChromosome;

	numRules -= numPruned;
	length = numRules * tSymbolic->ruleSize;
}

int classifier_gabil_packed::classify(instance *ins)
{
	aligned_uint *ptr = chromosome;
	int i, j, match;

	for (i = 0; i < numRules; i++) {
		if(ruleMatches(ptr,ins)) return i;
		ptr += tSymbolic->ruleSize;
	}

	if(tGlobals->defaultClassPolicy==DISABLED) return -1;
	return numRules;
}


int classifier_gabil_packed::ruleCleaning(aligned_uint *ptr,unsigned long long *matched,unsigned long long numMatches)
{
	unsigned long long i,j,k;
	instance **instances=is->getInstancesOfIteration();

	unsigned long long *listOK = new unsigned long long[numMatches];
	unsigned long long *listKO = new unsigned long long[numMatches];
	unsigned long long numOK=0,numKO=0;

	int cl=ptr[tSymbolic->recodeBlocks];
	for(i=0;i<numMatches;i++) {
		if(instances[matched[i]]->instanceClass==cl) {
			listOK[numOK++]=matched[i];
		} else {
			listKO[numKO++]=matched[i];
		}
	}

	cleanTargetPacked ct=cleanTargetOfRule(listOK,numOK,listKO,numKO);
	int changed=0;
	if(ct.maxNeg>0) {
		changed=1;

		int blockNum=ct.maxAtt/ATT_PER_BLOCK;
		int idx=ct.maxAtt%ATT_PER_BLOCK;
		SET_COND(ptr[blockNum],idx,ct.maxValue,0);
	}

	delete listOK;
	delete listKO;
	return changed;
}

unsigned long long classifier_gabil_packed::computeMatches(aligned_uint *ptr,unsigned long long *candidates,unsigned long long numCandidates,unsigned long long *matched)
{
	unsigned long long j;
	unsigned long long numMatches=0;
	instance **instances=is->getInstancesOfIteration();

	for(j=0;j<numCandidates;j++) {
		unsigned long long inst=candidates[j];
		if(ruleMatches(ptr,instances[inst])) {
			matched[numMatches++]=inst;
		}
	}

	return numMatches;
}

unsigned long long classifier_gabil_packed::updateMatches(unsigned long long *candidates,unsigned long long numCand,unsigned long long *matched,unsigned long long numMatched)
{
	unsigned long long numLeft=0;
	unsigned long long i;
	unsigned long long *candidatesLeft = new unsigned long long[numCand-numMatched];
	unsigned long long indexMatched=0;
	unsigned long long currMatched=matched[0];

	for(i=0;i<numCand;i++) {
		if(candidates[i]==currMatched) {
			indexMatched++;
			currMatched=matched[indexMatched];
		} else {
			candidatesLeft[numLeft++]=candidates[i];
		}
	}

	for(i=0;i<numLeft;i++) {
		candidates[i]=candidatesLeft[i];
	}

	delete candidatesLeft;

	return numLeft;
}

aligned_uint *classifier_gabil_packed::ruleSplitting(aligned_uint *ptr,unsigned long long *matched,unsigned long long numInst)
{
	unsigned long long i,j,k,l,m,n,o,p,q;
	instance **instances=is->getInstancesOfIteration();

	unsigned long long *listOK = new unsigned long long[numInst];
	unsigned long long *listKO = new unsigned long long[numInst];
	unsigned long long numOK=0,numKO=0;

	int cl=ptr[tSymbolic->recodeBlocks];
	for(i=0;i<numInst;i++) {
		if(instances[matched[i]]->instanceClass==cl) {
			listOK[numOK++]=matched[i];
		} else {
			listKO[numKO++]=matched[i];
		}
	}

	if(numKO==0) return NULL;
	activationsAttPacked *act=computeActivationStats(listOK,numOK,listKO,numKO);

	unsigned long long maxNeg=0;
	int splittedAtt;
	int splittedValue;
	int valueOrRest;
	JVector<int>conflicting;
	int negAtt;
	int negValue;

	for(i=0;i<tGlobals->numAttributesMC;i++) {
		int numConflictingValues=0;	
		int targetConflicting[act[i].numValues];
		for(j=0;j<act[i].numValues;j++) {
			if(act[i].actValues[j][0]>0 && act[i].actValues[j][1]>0) {
				targetConflicting[numConflictingValues++]=j;
			}
		}

		if(numConflictingValues>1) {
			attInstancesPacked att=computeInstOfAtt(i,listOK,numOK,listKO,numKO);
			for(l=0;l<numConflictingValues;l++) {
				int target=targetConflicting[l];

				unsigned long long numOK1=att.numInst[target][0];
				unsigned long long numKO1=att.numInst[target][1];
				unsigned long long numOK2=numOK-numOK1;
				unsigned long long numKO2=numKO-numKO1;
				unsigned long long *listOK2 = new unsigned long long[numOK2];
				unsigned long long *listKO2 = new unsigned long long[numKO2];
				unsigned long long countOK=0;
				unsigned long long countKO=0;
				for(m=0;m<act[i].numValues;m++) {
					if(m!=target) {
						for(n=0;n<att.numInst[m][0];n++) {
							listOK2[countOK++]=att.valueInstances[m][0][n];
						}
						for(n=0;n<att.numInst[m][1];n++) {
							listKO2[countKO++]=att.valueInstances[m][1][n];
						}
					}
				}

				cleanTargetPacked ct=cleanTargetOfRule(att.valueInstances[target][0]
					,numOK1,att.valueInstances[target][1],numKO1);
				if(ct.maxNeg>maxNeg) {
					maxNeg=ct.maxNeg;
					splittedAtt=i;
					splittedValue=target;
					valueOrRest=1;
					conflicting.removeAllElements();
					for(q=0;q<numConflictingValues;q++) {
						if(targetConflicting[q]!=target) {
							conflicting.addElement(targetConflicting[q]);
						}
					}
					negAtt=ct.maxAtt;
					negValue=ct.maxValue;
				}

				ct=cleanTargetOfRule(listOK2,numOK2,listKO2,numKO2);
				if(ct.maxNeg>maxNeg) {
					maxNeg=ct.maxNeg;
					splittedAtt=i;
					splittedValue=target;
					valueOrRest=0;
					conflicting.removeAllElements();
					for(q=0;q<numConflictingValues;q++) {
						if(targetConflicting[q]!=target) {
							conflicting.addElement(targetConflicting[q]);
						}
					}
					negAtt=ct.maxAtt;
					negValue=ct.maxValue;
				}

				delete listOK2;
				delete listKO2;
			}
			removeInstOfAtt(att);
		}
	}

	aligned_uint *newRule=NULL;
	if(maxNeg>0) {
		newRule = new aligned_uint[tSymbolic->ruleSize];
		bcopy(ptr,newRule,sizeof(aligned_uint)*tSymbolic->ruleSize);

		int blockNum=splittedAtt/ATT_PER_BLOCK;
		int idx=splittedAtt%ATT_PER_BLOCK;
		for(i=0;i<conflicting.size();i++) {
			SET_COND(ptr[blockNum],idx,conflicting[i],0);
		}
		SET_COND(newRule[blockNum],idx,splittedValue,0);

		blockNum=negAtt/ATT_PER_BLOCK;
		idx=negAtt%ATT_PER_BLOCK;
		if(valueOrRest) {
			SET_COND(ptr[blockNum],idx,negValue,0);
		} else {
			SET_COND(newRule[blockNum],idx,negValue,0);
		}
		modif = 1;
	}

	removeActivationStats(act);
	delete listOK;
	delete listKO;
	return newRule;
}

attInstancesPacked classifier_gabil_packed::computeInstOfAtt(int attribute,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO) 
{
	unsigned long long numInst=numOK+numKO;
	instance **instances=is->getInstancesOfIteration();
	unsigned long long i,j,k;
	int numValues=2;

	attInstancesPacked att;
	att.numValues=numValues;
	att.valueInstances = new unsigned long long **[numValues];
	att.numInst = new unsigned long long *[numValues];
	for(i=0;i<numValues;i++) {
		att.numInst[i]= new unsigned long long[2];
		att.numInst[i][0]=0;
		att.numInst[i][1]=0;
		att.valueInstances[i]= new unsigned long long *[2];
		att.valueInstances[i][0] = new unsigned long long[numInst];
		att.valueInstances[i][1] = new unsigned long long[numInst];
	}

	for(i=0;i<numOK;i++) {
		unsigned long long inst=instOK[i];
		register int value=instances[inst]->nominalValues[attribute];
		att.valueInstances[value][0][att.numInst[value][0]++]=inst;
	}
	for(i=0;i<numKO;i++) {
		unsigned long long inst=instKO[i];
		register int value=instances[inst]->nominalValues[attribute];
		att.valueInstances[value][1][att.numInst[value][1]++]=inst;
	}

	return att;
}

void classifier_gabil_packed::removeInstOfAtt(attInstancesPacked att) 
{
	unsigned long long o,p;

	for(o=0;o<att.numValues;o++) {
		for(p=0;p<2;p++) {
			delete att.valueInstances[o][p];
		}
		delete att.valueInstances[o];
		delete att.numInst[o];
	}
	delete att.valueInstances;
	delete att.numInst;
}

cleanTargetPacked classifier_gabil_packed::cleanTargetOfRule(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO) 
{
	instance **instances=is->getInstancesOfIteration();
	unsigned long long i,j,k;

	int numValues=2;
	int numAttributes=tGlobals->numAttributesMC;
	unsigned long long countKO[numAttributes][numValues];
	unsigned long long countOK[numAttributes][numValues];
	int candidateAtt[numAttributes];
	int countCandidateValues[numAttributes];
	cleanTargetPacked ct;
	ct.maxValue=-1;
	ct.maxAtt=-1;
	ct.maxNeg=0;

	int numCandidateAtts=numAttributes;
	for(i=0;i<numAttributes;i++) {
		candidateAtt[i]=i;
		countCandidateValues[i]=numValues;
		for(j=0;j<numValues;j++) {
			countOK[i][j]=countKO[i][j]=0;
		}
	}

	for(i=0;i<numOK && numCandidateAtts>0;i++) {
		instance *ins=instances[instOK[i]];
		int indexAtt=0;
		while(indexAtt<numCandidateAtts) {
			register int att=candidateAtt[indexAtt];
			register int value=ins->nominalValues[att];
			if(countOK[att][value]==0) {
				countCandidateValues[att]--;
				if(countCandidateValues[att]==0) {
					candidateAtt[indexAtt]=candidateAtt[numCandidateAtts-1];
					numCandidateAtts--;
				} else {
					indexAtt++;
				}
			} else {
				indexAtt++;
			}
			countOK[att][value]++;
		}
	}

	if(numCandidateAtts==0) return ct;

	for(i=0;i<numKO;i++) {
		instance *ins=instances[instKO[i]];
		for(j=0;j<numCandidateAtts;j++) {
			register int att=candidateAtt[j];
			register int value=ins->nominalValues[att];
			countKO[att][value]++;
		}
	}

	for(j=0;j<numCandidateAtts;j++) {
		register int att=candidateAtt[j];
		for(i=0;i<numValues;i++) {
			if(countOK[att][i]==0 && countKO[att][i]>ct.maxNeg) {
				ct.maxAtt=att;
				ct.maxValue=i;
				ct.maxNeg=countKO[att][i];
			}
		}
	}

	return ct;
}

activationsAttPacked *classifier_gabil_packed::computeActivationStats(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO)
{
	unsigned long long i,j;
	instance **instances=is->getInstancesOfIteration();

	activationsAttPacked *act=new activationsAttPacked[tGlobals->numAttributesMC];
	for(i=0;i<tGlobals->numAttributesMC;i++) {
		int numValues=2;
		act[i].numValues=numValues;
		act[i].actValues=new unsigned long long *[numValues];
		for(j=0;j<numValues;j++) {
			act[i].actValues[j]=new unsigned long long[2];
			act[i].actValues[j][0]=act[i].actValues[j][1]=0;
		}
	}

	
	for(j=0;j<numOK;j++) {
		instance *ins=instances[instOK[j]];
		for(i=0;i<tGlobals->numAttributesMC;i++) {
			act[i].actValues[ins->nominalValues[i]][0]++;
		}
	}
	for(j=0;j<numKO;j++) {
		instance *ins=instances[instKO[j]];
		for(i=0;i<tGlobals->numAttributesMC;i++) {
			act[i].actValues[ins->nominalValues[i]][1]++;
		}
	}

	return act;
}

void classifier_gabil_packed::removeActivationStats(activationsAttPacked *act) 
{
	unsigned long long i,j;

	for(i=0;i<tGlobals->numAttributesMC;i++) {
		for(j=0;j<act[i].numValues;j++) {
			delete act[i].actValues[j];
		}
		delete act[i].actValues;
	}
	delete act;
}

void classifier_gabil_packed::generalizeRule(aligned_uint *rule,unsigned long long *candidates,unsigned long long numCand,unsigned long long *newInstances,unsigned long long &numNew)
{
	unsigned long long i,j,k;
	int len=tGlobals->numAttributes*2;
	int decoded[len];
	unsigned long long countPos[len];
	unsigned long long countNeg[len];
	JVector<unsigned long long> *candInst[len];
	instance **instances=is->getInstancesOfIteration();

	int countAtt=0;
	for(i=0,countAtt=0;i<tSymbolic->recodeBlocks;i++) {
		aligned_uint tmp=rule[i];
		for(j=0;j<ATT_PER_BLOCK && countAtt<tGlobals->numAttributesMC;j++,countAtt++) {
			for (k = 0; k < 2; k++) {
				decoded[countAtt*2+k]=GET_COND(tmp,j,k);
			}
		} 
	}
	int cl=rule[tSymbolic->recodeBlocks];

	for(i=0;i<len;i++) {
		countPos[i]=countNeg[i]=0;
		candInst[i]=NULL;
	}

	for(i=0;i<numCand;i++) {
		unsigned long long posInst=candidates[i];
		instance *ins=instances[posInst];
		unsigned long long numMatches=0;
		int whichPos;
		for(j=0;j<tGlobals->numAttributesMC;j++) {
			int value=ins->nominalValues[j];
			register int pos=j*2+value;
			if(decoded[pos]==0) {
				numMatches++;
				whichPos=pos;
				if(numMatches>=2) break;
			}
		}

		if(numMatches==1) {
			if(ins->instanceClass==cl) {
				countPos[whichPos]++;
				if(countNeg[whichPos]==0) {
					if(candInst[whichPos]==NULL) {
						candInst[whichPos] = new JVector<unsigned long long>((unsigned long long)(numCand*0.1));
					}
					candInst[whichPos]->addElement(posInst);
				}
			} else {
				countNeg[whichPos]++;
				if(candInst[whichPos]) {
					delete candInst[whichPos];
					candInst[whichPos]=NULL;
				}
			}
		}
	}
	
	unsigned long long max=0;
	int bestPos;
	for(i=0;i<len;i++) {
		if(countNeg[i]==0 && countPos[i]>max) {
			bestPos=i;
			max=countPos[i];
		}
	}

	numNew=max;
	if(max>0) {
		int att=bestPos/2;
		int val=bestPos%2;
		int blockNum=att/ATT_PER_BLOCK;
		int idx=att%ATT_PER_BLOCK;
		SET_COND(rule[blockNum],idx,val,1);
		modif = 1;

		for(i=0;i<max;i++) {
			newInstances[i]=candInst[bestPos]->elementAt(i);
		}
	}

	for(i=0;i<len;i++) {
		if(candInst[i]) delete candInst[i];
	}
}

void classifier_gabil_packed::doLocalSearch()
{
	unsigned long long numInstances = is->getNumInstancesOfIteration();	
	unsigned long long *remainingInstances = new unsigned long long[numInstances];
	unsigned long long *matched = new unsigned long long[numInstances];
	unsigned long long i;

	for(i=0;i<numInstances;i++) {
		remainingInstances[i]=i;
	}

	unsigned long long index=0;
	int lastSplitted=0;
	while(index<numRules) {
		aligned_uint *ptr=&chromosome[index*tSymbolic->ruleSize];	

		unsigned long long numMatches = computeMatches(ptr,remainingInstances,numInstances,matched);
		if(tGlobals->doRuleCleaning) {
			if(ruleCleaning(ptr,matched,numMatches)) {
				numMatches = computeMatches(ptr,remainingInstances,numInstances,matched);
			}
		}

		if(tGlobals->doRuleSplitting) {
			if(lastSplitted) {
				lastSplitted=0;
			} else {
				aligned_uint *newRule=ruleSplitting(ptr,matched,numMatches);
				if(newRule!=NULL) {
					aligned_uint *newChromosome = new aligned_uint[tSymbolic->ruleSize*(numRules+1)];
					bcopy(chromosome,newChromosome,(index+1)*tSymbolic->ruleSize*sizeof(aligned_uint));
					bcopy(newRule,&newChromosome[(index+1)*tSymbolic->ruleSize],tSymbolic->ruleSize*sizeof(aligned_uint));
					bcopy(&chromosome[(index+1)*tSymbolic->ruleSize],&newChromosome[(index+2)*tSymbolic->ruleSize],(numRules-index-1)*tSymbolic->ruleSize*sizeof(aligned_uint));
					delete chromosome;
					chromosome=newChromosome;
					numRules++;
					length=numRules*tSymbolic->ruleSize;
	
					ptr=&chromosome[index*tSymbolic->ruleSize];	
					numMatches = computeMatches(ptr,remainingInstances,numInstances,matched);
					delete newRule;
					lastSplitted=1;
				}
			}
		}

		numInstances=updateMatches(remainingInstances,numInstances,matched,numMatches);
		if(numInstances==0) {
			break;
		}

		if(tGlobals->doRuleGeneralizing) {
			generalizeRule(ptr,remainingInstances,numInstances,matched,numMatches);
			if(numMatches>0) {
				numInstances=updateMatches(remainingInstances,numInstances,matched,numMatches);
				if(numInstances==0) {
					break;
				}
			}
		}
		
		index++;	
	}

	delete remainingInstances;
	delete matched;
}

void classifier_gabil_packed::findOrder(JVector<matchProfileAgent *>&profiles
	,JVector<aligned_uint *>&ptrs,JVector<int> &ruleOrder) 
{
	unsigned long long numInst=is->getNumInstancesOfIteration();
	unsigned long long i,j;

	ruleOrderAgent *best=NULL;
	for(j=0;j<tCross->repetitionsRuleOrdering;j++) {
		int totalRules=profiles.size();
		Sampling samp(totalRules);
		ruleOrderAgent *roa = new ruleOrderAgent(defaultClass,numInst,totalRules);

		for(i=0;i<totalRules;i++) {
			int rule=samp.getSample();
			roa->insertRule(rule,profiles[rule]);
		}

		
		if(tGlobals->doRuleCleaning) {
			int num=roa->getNumRules()-1;
			for(i=0;i<num;i++) {
				matchProfileAgent *mpa=roa->generateActualMPA(i);
				if(cleanRule(ptrs[roa->getRule(i)],mpa)) {
					roa->removeMatchesOfRule(i,mpa);
				}
				delete mpa;
			}
		}
	
		
		if(tGlobals->doRuleSplitting) {
			for(i=0;i<roa->getNumRules()-1;i++) {
				matchProfileAgent *mpa=roa->generateActualMPA(i);
				splittedRulePacked spl;
				if(editRule(ptrs[roa->getRule(i)],mpa,spl)) {
					roa->removeRule(i);
					profiles.addElement(spl.mpa1);
					profiles.addElement(spl.mpa2);
					ptrs.addElement(spl.ptr1);
					ptrs.addElement(spl.ptr2);
					roa->insertRule(profiles.size()-2,profiles[profiles.size()-2]);
					roa->insertRule(profiles.size()-1,profiles[profiles.size()-1]);
				}
				delete mpa;
			}
		}
	
		if(tGlobals->doRuleGeneralizing) {
			unsigned long long numCand;
			unsigned long long candidates[numInst];	
			for(i=0;i<roa->getNumRules()-1;i++) {
				int whichRule=roa->getRule(i);
				aligned_uint *ptr=ptrs[whichRule];
				roa->getPossibleInstancesOfRule(i,candidates,numCand);
				unsigned long long newInstances[numCand];
				unsigned long long numNew;
				generalizeRule(ptr,candidates,numCand,newInstances,numNew);
				if(numNew>0) {
					delete profiles[whichRule];
					matchProfileAgent *mpa=evaluateRuleCX(ptr);
					profiles[whichRule]=mpa;
					roa->addPositiveExamples(i,newInstances,numNew,mpa);
				}
			}
		}
	
		roa->refineRuleSet();

		if(best==NULL) {
			best=roa;
		} else {
			double accOld=best->getAccuracy();
			double accNew=roa->getAccuracy();
			if(accNew>accOld) {
				delete best;
				best=roa;
			} else if(accNew==accOld 
				&& roa->getNumRules()<best->getNumRules()) {
				delete best;
				best=roa;
			} else {
				delete roa;
			}
		}
	}

	best->copyOrder(ruleOrder);
	if(ruleOrder.size()==0) {
		int pos=rnd(0,profiles.size()-1);
		ruleOrder.addElement(pos);
	}
	delete best;
}

matchProfileAgent *classifier_gabil_packed::evaluateRuleCX(aligned_uint *rule)
{
	unsigned long long i;
	int cl = rule[tSymbolic->recodeBlocks];
	unsigned long long max = is->getNumInstancesOfIteration();
	instance **instances=is->getInstancesOfIteration();

	matchProfileAgent *mpa=new matchProfileAgent(max,cl);

	for (i = 0; i < max; i++) {
		if (ruleMatches(rule, instances[i])) {
			if (instances[i]->instanceClass == cl) {
				mpa->addOK(i);
			} else {
				mpa->addKO(i);
			}
		}
	}
	mpa->generateProfiles();

	return mpa;
}


void classifier_gabil_packed::instOfAttAndValue(int attribute,int value,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO,unsigned long long *list)
{

	instance **instances=is->getInstancesOfIteration();
	unsigned long long i;

	unsigned long long num=0;

	for(i=0;i<numOK;i++) {
		int instValue=instances[instOK[i]]->nominalValues[attribute];
		if(instValue==value) {
			list[num++]=instOK[i];
		}
	}
	for(i=0;i<numKO;i++) {
		int instValue=instances[instKO[i]]->nominalValues[attribute];
		if(instValue==value) {
			list[num++]=instKO[i];
		}
	}
}


int classifier_gabil_packed::editRule(aligned_uint *ptr,matchProfileAgent *mpa, splittedRulePacked &spl)
{
	unsigned long long i,j,k,l,m,n,o,p,q;

	if(mpa->numKO==0) return 0;
	activationsAttPacked *act=computeActivationStats(mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO);

	unsigned long long maxNeg=0;
	int splittedAtt;
	int splittedValue;
	int valueOrRest;
	JVector<int>conflicting;
	int negAtt;
	int negValue;

	for(i=0;i<tGlobals->numAttributesMC;i++) {
		int numConflictingValues=0;	
		int targetConflicting[act[i].numValues];
		for(j=0;j<act[i].numValues;j++) {
			if(act[i].actValues[j][0]>0 && act[i].actValues[j][1]>0) {
				targetConflicting[numConflictingValues++]=j;
			}
		}

		if(numConflictingValues>1) {
			attInstancesPacked att=computeInstOfAtt(i,mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO);
			for(l=0;l<numConflictingValues;l++) {
				int target=targetConflicting[l];

				unsigned long long numOK1=att.numInst[target][0];
				unsigned long long numKO1=att.numInst[target][1];
				unsigned long long numOK2=mpa->numOK-numOK1;
				unsigned long long numKO2=mpa->numKO-numKO1;
				unsigned long long *listOK2 = new unsigned long long[numOK2];
				unsigned long long *listKO2 = new unsigned long long[numKO2];
				unsigned long long countOK=0;
				unsigned long long countKO=0;
				for(m=0;m<act[i].numValues;m++) {
					if(m!=target) {
						for(n=0;n<att.numInst[m][0];n++) {
							listOK2[countOK++]=att.valueInstances[m][0][n];
						}
						for(n=0;n<att.numInst[m][1];n++) {
							listKO2[countKO++]=att.valueInstances[m][1][n];
						}
					}
				}

				cleanTargetPacked ct=cleanTargetOfRule(att.valueInstances[target][0]
					,numOK1,att.valueInstances[target][1],numKO1);
				if(ct.maxNeg>maxNeg) {
					maxNeg=ct.maxNeg;
					splittedAtt=i;
					splittedValue=target;
					valueOrRest=1;
					conflicting.removeAllElements();
					for(q=0;q<numConflictingValues;q++) {
						if(targetConflicting[q]!=target) {
							conflicting.addElement(targetConflicting[q]);
						}
					}
					negAtt=ct.maxAtt;
					negValue=ct.maxValue;
				}

				ct=cleanTargetOfRule(listOK2,numOK2,listKO2,numKO2);
				if(ct.maxNeg>maxNeg) {
					maxNeg=ct.maxNeg;
					splittedAtt=i;
					splittedValue=target;
					valueOrRest=0;
					conflicting.removeAllElements();
					for(q=0;q<numConflictingValues;q++) {
						if(targetConflicting[q]!=target) {
							conflicting.addElement(targetConflicting[q]);
						}
					}
					negAtt=ct.maxAtt;
					negValue=ct.maxValue;
				}

				delete listOK2;
				delete listKO2;
			}
			removeInstOfAtt(att);
		}
	}

	int splitted=0;
	if(maxNeg>0) {
		splitted=1;
		aligned_uint *ptr1= new aligned_uint[tSymbolic->ruleSize];
		bcopy(ptr,ptr1,sizeof(aligned_uint)*tSymbolic->ruleSize);
		aligned_uint *ptr2= new aligned_uint[tSymbolic->ruleSize];
		bcopy(ptr,ptr2,sizeof(aligned_uint)*tSymbolic->ruleSize);

		int blockNum=splittedAtt/ATT_PER_BLOCK;
		int idx=splittedAtt%ATT_PER_BLOCK;
		for(i=0;i<conflicting.size();i++) {
			SET_COND(ptr1[blockNum],idx,conflicting[i],0);
		}
		SET_COND(ptr2[blockNum],idx,splittedValue,0);

		blockNum=negAtt/ATT_PER_BLOCK;
		idx=negAtt%ATT_PER_BLOCK;
		if(valueOrRest) {
			SET_COND(ptr1[blockNum],idx,negValue,0);
		} else {
			SET_COND(ptr2[blockNum],idx,negValue,0);
		}

		spl.ptr1=ptr1;
		spl.ptr2=ptr2;
		spl.mpa1=evaluateRuleCX(ptr1);
		spl.mpa2=evaluateRuleCX(ptr2);
	}

	removeActivationStats(act);
	return splitted;
}

int classifier_gabil_packed::cleanRule(aligned_uint *ptr,matchProfileAgent *mpa)
{
	if(mpa->numKO==0) return 0;

	cleanTargetPacked ct=cleanTargetOfRule(mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO);
	int changed=0;
	if(ct.maxNeg>0) {
		changed=1;

		int blockNum=ct.maxAtt/ATT_PER_BLOCK;
		int idx=ct.maxAtt%ATT_PER_BLOCK;
		SET_COND(ptr[blockNum],idx,ct.maxValue,0);

		unsigned long long removedExamples[ct.maxNeg];
		instOfAttAndValue(ct.maxAtt,ct.maxValue,mpa->listOK,mpa->numOK,mpa->listKO,mpa->numKO,removedExamples);
		mpa->removeMatched(removedExamples,ct.maxNeg);
	}

	return changed;
}	
	
void classifier_gabil_packed::crossover_smart(classifier **parents, int num) 
{
	int i, j, k;

	modif = 1;
	if(chromosome) delete chromosome;

	JVector<aligned_uint *>ptrs;
	JVector<matchProfileAgent *>profiles;
	int ruleCount=0;

	for(k=0;k<num;k++) {
		classifier_gabil_packed *parent=(classifier_gabil_packed *)parents[k];
		aligned_uint *ptr=parent->chromosome;
		for(i=0;i<parent->numRules;ptr+=tSymbolic->ruleSize,i++) {
			aligned_uint *newRule= new aligned_uint[tSymbolic->ruleSize];
			bcopy(ptr,newRule,sizeof(aligned_uint)*tSymbolic->ruleSize);
			matchProfileAgent *mpa = evaluateRuleCX(newRule);
			if(mpa->numOK>0) {
				ptrs.addElement(newRule);
				profiles.addElement(mpa);
			} else {
				delete mpa;
				delete newRule;
			}
		}
	}

	JVector<int> ruleOrder;
	findOrder(profiles,ptrs,ruleOrder);

	numRules = ruleOrder.size(); 
	length = numRules * tSymbolic->ruleSize;
	chromosome = new aligned_uint[length];
	for (i = 0; i < numRules; i++) {
		bcopy(ptrs[ruleOrder[i]], &chromosome[i * tSymbolic->ruleSize],
		      sizeof(aligned_uint) * tSymbolic->ruleSize);
	}

	for(i=0;i<profiles.size();i++) {
		delete profiles[i];
	}
	for(i=0;i<ptrs.size();i++) {
		delete ptrs[i];
	}
}

