#ifndef _CLASSIFIER_GABIL_PACKED_
#define _CLASSIFIER_GABIL_PACKED_

#include "classifier.h"
#include "agentPerformanceTraining.h"
#include "timerGlobals.h"
#include "timerEvolutionStats.h"
#include "timerSymbolicKR.h"
#include "matchProfileAgent.h"
#include <emmintrin.h>

extern timerGlobals *tGlobals;
extern timerSymbolicKR *tSymbolic;
extern timerEvolutionStats *tEvolStats;

typedef struct {
        aligned_uint *ptr1;
        matchProfileAgent *mpa1;
        aligned_uint *ptr2;
        matchProfileAgent *mpa2;
} splittedRulePacked;

typedef struct {
        int numValues;
        unsigned long long ***valueInstances;
        unsigned long long **numInst;
} attInstancesPacked;

typedef struct {
	int maxAtt;
	int maxValue;
	unsigned long long maxNeg;
} cleanTargetPacked;



typedef struct {
        int numValues;
        unsigned long long **actValues;
} activationsAttPacked;

#define COND_MASK 0x1U
#define GET_COND(BLK,IDX,COND) ((BLK&(COND_MASK<<(2*IDX+COND)))>>(2*IDX+COND))
#define SET_COND(BLK,IDX,COND,VAL) (BLK=(((ALL_MASK^(COND_MASK<<(2*IDX+COND)))&BLK)|(VAL<<(2*IDX+COND))))


class classifier_gabil_packed: public classifier   {
	aligned_uint *chromosome;
	int numRules;

	void crossover_1px(classifier_gabil_packed *in1
		,classifier_gabil_packed *in2
		,classifier_gabil_packed *out1
		,classifier_gabil_packed *out2);
	unsigned char getGene(short int rule,short int gene,short int value); 
	void setGene(short int rule,short int gene,short int value
		,unsigned char nfo);
	void gabilRuleCover(aligned_uint *rule,instance *ins,double prob);
	void initializeChromosome(void);

	int ruleCleaning(aligned_uint *ptr,unsigned long long *instances,unsigned long long numInstances);
	unsigned long long updateMatches(unsigned long long *instanceMatched,unsigned long long totalInstances,unsigned long long *instanceMatchedTemp,unsigned long long numMatched);
	unsigned long long computeMatches(aligned_uint *ptr,unsigned long long *instanceMatched,unsigned long long totalInstances,unsigned long long *instanceMatchedTemp);
	aligned_uint *ruleSplitting(aligned_uint *ptr,unsigned long long *instances,unsigned long long numInstances);
	cleanTargetPacked cleanTargetOfRule(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO);
        void instOfAttAndValue(int attribute,int value,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO,unsigned long long *list);
	attInstancesPacked computeInstOfAtt(int attribute,unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO);
	void removeInstOfAtt(attInstancesPacked att);
	activationsAttPacked *computeActivationStats(unsigned long long *instOK,unsigned long long numOK,unsigned long long *instKO,unsigned long long numKO);
	void removeActivationStats(activationsAttPacked *act);
	void findOrder(JVector<matchProfileAgent *>&profiles
		,JVector<aligned_uint *>&ptrs,JVector<int> &ruleOrder);
	matchProfileAgent *evaluateRuleCX(aligned_uint *rule);
	int cleanRule(aligned_uint *ptr,matchProfileAgent *mpa);
	int editRule(aligned_uint *ptr,matchProfileAgent *mpa,splittedRulePacked &spl);
	void generalizeRule(aligned_uint *rule,unsigned long long *candidates,unsigned long long numCand,unsigned long long *newInst,unsigned long long &numNew);
	
	inline int ruleMatches(aligned_uint *rule,instance *ins) 
	{
		tEvolStats->numExampleEvaluations++;
		register int match;
		/*register int i;

		for(i=0, match=1; i<tSymbolic->recodeBlocks && match;i++) {
			match = ((rule[i]&ins->packedRepresentation[i])
				== ins->packedRepresentation[i]);
		}*/
		register int i;
		__m128i vir,vii;

		for(i=0, match=1; i<tSymbolic->recodeBlocks && match;i+=4) {
			vir = _mm_load_si128((__m128i*)&rule[i]);
			vii = _mm_load_si128((__m128i*)&ins->packedRepresentation[i]);
			vir = _mm_and_si128(vir,vii);
			vii = _mm_cmpeq_epi32(vir,vii);
			match &= (-1 == (short int)_mm_movemask_epi8(vii));
		}

		return match;
	}

public:
	classifier_gabil_packed();
	classifier_gabil_packed(const classifier_gabil_packed &orig,int son=0);
	~classifier_gabil_packed();

	inline int getClass(int rule) {
		if(rule==numRules) return defaultClass;
		return chromosome[tSymbolic->ruleSize*rule+tSymbolic->recodeBlocks];
	}

	inline int getNumClassifiers() {
		int nC=numRules;
		if(tGlobals->defaultClassPolicy!=DISABLED) nC++;
		return nC;
	}

	void pruneClassifiers(int *classifiersToPrune,int numPruned);

	int classify(instance *ins);
	double computateTheoryLength(agentPerformanceTraining *ap);
	
	inline int hasSmartCrossover() {return 1;}
	void crossover_smart(classifier **parents, int num);

	inline double getHierarchicalSelectionLength()
	{
		if(tHierar->useMDL) return getTheoryLength();
		return aliveClassifiers;
	}

	inline int getNiche() {
		if(tGlobals->defaultClassPolicy!=AUTO) return 0;
		return getClass(getNumClassifiers()-1);
	}
	inline int getNumNiches() {
		if(tGlobals->defaultClassPolicy!=AUTO) return 1;
		return ai.getNumClasses();
	}

	void crossover(classifier *,classifier *,classifier *);
	void mutation();
	void dumpPhenotype(char *string);
	inline int numSpecialStages(){return 0;}
	void doSpecialStage(int stage){}

	//void classifyTraining(instanceSet *is,agentPerformanceTraining *ap);
	//int canDoSmartClassAsignment() {return 1;}
	//int canDoTrainAndClean() {return 1;}

	void doLocalSearch();
	int hasLocalSearch(){return 1;}
	

};

#endif

