#include "classifier_hyperrect.h"
#include "random.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "attributesInfo.h"
#include "timerGlobals.h"
#include "timerMutation.h"
#include "timerCrossover.h"
#include "timerMDL.h"
#include "timerHierar.h"
#include "timerRealKR.h"
#include "instanceSet.h"
#include "ruleOrderAgent.h"
#include "sampling.h"

extern attributesInfo ai;
extern timerGlobals *tGlobals;
extern timerHierar *tHierar;
extern timerMutation *tMut;
extern timerCrossover *tCross;
extern timerMDL *tMDL;
extern timerRealKR *tReal;
extern instanceSet *is;
extern Random rnd;

double classifier_hyperrect::computateTheoryLength(agentPerformanceTraining *ap)
{
	int i, j, k;
	double *ptr = chromosome;
	theoryLength = 0.0;

	for (i = 0; i < numRules; i++) {
		if(ap->getActivationsOfClassifier(i)) {
			for (j = 0; j < tGlobals->numAttributesMC; j++) {
				if (ai.getTypeOfAttribute(j) == REAL) {
					float vmin=ptr[0];
					float vmax=ptr[1];
					float size=ai.getSizeDomain(j);
					if(vmin<vmax && size>0) {
						theoryLength += 2.0 - (vmax-vmin) / size;
					}
				} else {
					int countFalses = 0;
					int numValues=tReal->attributeSize[j];
					for(k=0;k<numValues;k++) {
						if(!ptr[k]) countFalses++;
					}
					theoryLength += (double)countFalses/(double)numValues;
				}
				ptr += tReal->attributeSize[j];
			}
			ptr++;
		} else {
			ptr+=tReal->ruleSize;
		}
	}
	if(tGlobals->defaultClassPolicy!=DISABLED) {
		theoryLength+=tReal->defaultRuleTheoryLength;
	}

	return theoryLength;
}

classifier_hyperrect::classifier_hyperrect()
{
	numRules =
	    rnd(tGlobals->minClassifiersInit,
		tGlobals->maxClassifiersInit);
	length = tReal->ruleSize * numRules;
	chromosome = new double[tReal->ruleSize * numRules];
	if (chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	if(tGlobals->defaultClassPolicy==AUTO) {
		defaultClass=rnd(0,ai.getNumClasses()-1);
	} else  {
		defaultClass=tGlobals->defaultClass;
	}

	classifier_hyperrect::initializeChromosome();
}

classifier_hyperrect::classifier_hyperrect(const classifier_hyperrect & orig, int son)
{
	*this = orig;

	if (!son) {
		chromosome = new double[tReal->ruleSize * numRules];
		if (chromosome == NULL) {
			perror("out of memory");
			exit(1);
		}
		bcopy(orig.chromosome, chromosome,
		      numRules * tReal->ruleSize * sizeof(double));
	} else {
		chromosome = NULL;
	}
}

classifier_hyperrect::~classifier_hyperrect()
{
	delete chromosome;
}

void classifier_hyperrect::initializeChromosome()
{
	int i, j, k;

	for (i = 0; i < numRules; i++) {
		instance *ins=NULL;
		if(tGlobals->smartInit) {
			if(tGlobals->defaultClassPolicy!=DISABLED) {
				ins=is->getInstanceInit(defaultClass);
			} else {
				ins=is->getInstanceInit(ai.getNumClasses());
			}
		}

		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				float min,max;
				float sizeD=ai.getSizeDomain(j);
				float minD=ai.getMinDomain(j);
				float maxD=ai.getMaxDomain(j);
				if(!rnd<tReal->probIrr) {
					max=!rnd*sizeD+minD;
					min=!rnd*(maxD-max)+max;
				} else {
					//float size=(!rnd*0.10+0.90)*ai.getSizeDomain(j);
					float size=(!rnd*0.5+0.25)*sizeD;
					if(ins) {
						float val=ins->realValueOfAttribute(j);
						min=val-size/2.0;
						max=val+size/2.0;
						if(min<minD) {
							max+=(minD-min);
							min=minD;
						}
						if(max>maxD) {
							min-=(max-maxD);
							max=maxD;
						}
					} else {
						min=!rnd*(sizeD-size)+minD;
						max=min+size;
					}
				}
	
				setGene(i,j,0,min);
				setGene(i,j,1,max);
			} else {
				int value;
				if(ins) value=ins->valueOfAttribute(j);
				else value=-1;
				for(k=0;k<tReal->attributeSize[j];k++) {
					if(k!=value) {
						if(!rnd<tGlobals->probOne) {
							setGene(i,j,k,1);
						} else {
							setGene(i,j,k,0);
						}
					} else {
						setGene(i,j,k,1);
					}
				}
			}
		}
		int cl;
		do {
			cl=rnd(0,ai.getNumClasses()-1);
		} while(tGlobals->defaultClassPolicy!=DISABLED && cl==defaultClass);
		setGene(i, tGlobals->numAttributesMC, 0, cl);
	}
}

void classifier_hyperrect::setGene(short int rule, short int gene,
			      short int value, double nfo)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules) ||
	    (value < 0) || (value >= tReal->attributeSize[gene])) {
		fprintf(stderr,"Invalid access at setGene %d %d %d\n",
		       rule, gene, value);
		exit(1);
	}
#endif
	chromosome[rule * tReal->ruleSize +
		   tReal->attributeOffset[gene] + value] = nfo;
	modif = 1;
}

double classifier_hyperrect::getGene(short int rule,
				short int gene, short int value)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules) ||
	    (value < 0) || (value >= tReal->attributeSize[gene])) {
		fprintf(stderr,"Invalid access at getGene %d %d %d", rule,
		       gene, value);
		exit(1);
	}
#endif
	return chromosome[rule * tReal->ruleSize +
			  tReal->attributeOffset[gene] + value];
}

void classifier_hyperrect::crossover(classifier * in,
				classifier * out1, classifier * out2)
{
	switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this, (classifier_hyperrect *) in,
			      (classifier_hyperrect *) out1,
			      (classifier_hyperrect *) out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_hyperrect *) in,
			      (classifier_hyperrect *) out1,
			      (classifier_hyperrect *) out2);
	}
}


double classifier_hyperrect::mutationOffset(double geneValue, double offsetMin,
				       double offsetMax)
{
	double newValue;
	if (!rnd < 0.5) {
		newValue = geneValue + !rnd * offsetMax;
	} else {
		newValue = geneValue - !rnd * offsetMin;
	}
	return newValue;
}

void classifier_hyperrect::mutation()
{
	int i;
	int rule, attribute, value;

	modif = 1;
	tMut->mutationCounter++;

	rule = rnd(0,numRules-1);
	if(tGlobals->numClasses>1 && !rnd<0.10) {
		attribute = tGlobals->numAttributesMC;
		value=0;
	} else {
		attribute=rnd(0,tGlobals->numAttributesMC-1);
		value=rnd(0,tReal->attributeSize[attribute]-1);
	}

	if (attribute != tGlobals->numAttributesMC) {
		double oldValue=getGene(rule,attribute,value);
		if (ai.getTypeOfAttribute(attribute) == REAL) {
			double newValue;
			double minOffset, maxOffset;
			minOffset = maxOffset =
			    0.5 * ai.getSizeDomain(attribute);
			newValue = mutationOffset(oldValue, minOffset,
						  maxOffset);
			if (newValue < ai.getMinDomain(attribute))
				newValue = ai.getMinDomain(attribute);
			if (newValue > ai.getMaxDomain(attribute))
				newValue = ai.getMaxDomain(attribute);
			setGene(rule, attribute, value, newValue);
		} else {
			if(oldValue==1) setGene(rule,attribute,value,0);
			else setGene(rule,attribute,value,1);
		}
	} else {
		if(tGlobals->numClasses>1) {
			int newValue;
			int oldValue = (int) getGene(rule, attribute, value);
			do {
				newValue = rnd(0, ai.getNumClasses()-1);
			} while (newValue == oldValue || tGlobals->defaultClassPolicy!=DISABLED && newValue==defaultClass);
			setGene(rule, attribute, value, newValue);
		}
	}
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

void classifier_hyperrect::specialCrossover(double *parent1, double *parent2,
				       double *son1, double *son2)
{
	int i;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		if (ai.getTypeOfAttribute(i) == REAL) {
			tReal->specialCrossover(parent1[2], parent2[2],
						 son1[2], son2[2],
						 ai.getMinDomain(i),
						 ai.getMaxDomain(i));
			tReal->specialCrossover(parent1[3], parent2[3],
						 son1[3], son2[3], 0,
						 ai.getSizeDomain(i));
		}
		parent1 += tReal->attributeSize[i];
		parent2 += tReal->attributeSize[i];
		son1 += tReal->attributeSize[i];
		son2 += tReal->attributeSize[i];
	}
}

void classifier_hyperrect::crossover_2px(classifier_hyperrect * in1,
				    classifier_hyperrect * in2,
				    classifier_hyperrect * out1,
				    classifier_hyperrect * out2)
{
	int i1rule1, i1rule2, i2rule1, i2rule2;
	int cutPoint1, cutPoint2, sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	i1rule1 = rnd(0, in1->numRules - 1);
	i1rule2 = rnd(0, in1->numRules - 1);
	if (i1rule1 > i1rule2)
		swap(i1rule1, i1rule2);

	i2rule1 = rnd(0, in2->numRules - 1);
	i2rule2 = rnd(0, in2->numRules - 1);
	if (i2rule1 > i2rule2)
		swap(i2rule1, i2rule2);

	sizeOfSon1 = i1rule1 + i2rule2 - i2rule1 + in1->numRules - i1rule2;
	sizeOfSon2 = i2rule1 + i1rule2 - i1rule1 + in2->numRules - i2rule2;

	cutPoint1 = rnd(0, tReal->ruleSize - 1);
	cutPoint2 = rnd(0, tReal->ruleSize - 1);

	if (i1rule1 == i1rule2 || i2rule1 == i2rule2) {
		if (cutPoint1 > cutPoint2)
			swap(cutPoint1, cutPoint2);
	}

	out1->length = sizeOfSon1 * tReal->ruleSize;
	out2->length = sizeOfSon2 * tReal->ruleSize;

	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;

	out1->chromosome = new double[tReal->ruleSize * out1->numRules];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new double[tReal->ruleSize * out2->numRules];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	int cutPoint = cutPoint1;

	bcopy(chromosome, out1->chromosome,
	      i1rule1 * tReal->ruleSize * sizeof(double));
	bcopy(in2->chromosome, out2->chromosome,
	      i2rule1 * tReal->ruleSize * sizeof(double));

	bcopy(&in1->chromosome[i1rule1 * tReal->ruleSize],
	      &out1->chromosome[i1rule1 * tReal->ruleSize],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[i2rule1 * tReal->ruleSize],
	      &out2->chromosome[i2rule1 * tReal->ruleSize],
	      cutPoint * sizeof(double));

	bcopy(&in2->chromosome[i2rule1 * tReal->ruleSize + cutPoint],
	      &out1->chromosome[i1rule1 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));
	bcopy(&in1->chromosome[i1rule1 * tReal->ruleSize + cutPoint],
	      &out2->chromosome[i2rule1 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->
				 chromosome[i1rule1 * tReal->ruleSize],
				 &in2->chromosome[i2rule1 *
						  tReal->ruleSize],
				 &out1->chromosome[i1rule1 *
						   tReal->ruleSize],
				 &out2->chromosome[i2rule1 *
						   tReal->ruleSize]);

	if (i2rule1 != i2rule2)
		bcopy(&in2->chromosome[(i2rule1 + 1) * tReal->ruleSize],
		      &out1->chromosome[(i1rule1 + 1) * tReal->ruleSize],
		      (i2rule2 -
		       (i2rule1 + 1)) * tReal->ruleSize * sizeof(double));
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[(i1rule1 + 1) * tReal->ruleSize],
		      &out2->chromosome[(i2rule1 + 1) * tReal->ruleSize],
		      (i1rule2 -
		       (i1rule1 + 1)) * tReal->ruleSize * sizeof(double));


	cutPoint = cutPoint2;
	if (i2rule2 != i2rule1)
		bcopy(&in2->chromosome[i2rule2 * tReal->ruleSize],
		      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
					tReal->ruleSize],
		      cutPoint * sizeof(double));
	if (i1rule2 != i1rule1)
		bcopy(&in1->chromosome[i1rule2 * tReal->ruleSize],
		      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
					tReal->ruleSize],
		      cutPoint * sizeof(double));

	bcopy(&in1->chromosome[i1rule2 * tReal->ruleSize + cutPoint],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
				tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));
	bcopy(&in2->chromosome[i2rule2 * tReal->ruleSize + cutPoint],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
				tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->
				 chromosome[i1rule2 * tReal->ruleSize],
				 &in2->chromosome[i2rule2 *
						  tReal->ruleSize],
				 &out1->
				 chromosome[(i1rule1 + i2rule2 -
					     i2rule1) *
					    tReal->ruleSize],
				 &out2->
				 chromosome[(i2rule1 + i1rule2 -
					     i1rule1) * tReal->ruleSize]);

	bcopy(&in1->chromosome[(i1rule2 + 1) * tReal->ruleSize],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1 + 1) *
				tReal->ruleSize],
	      (in1->numRules -
	       (i1rule2 + 1)) * tReal->ruleSize * sizeof(double));
	bcopy(&in2->chromosome[(i2rule2 + 1) * tReal->ruleSize],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1 + 1) *
				tReal->ruleSize],
	      (in2->numRules -
	       (i2rule2 + 1)) * tReal->ruleSize * sizeof(double));
}

void classifier_hyperrect::dumpPhenotype(char *string)
{
	double *ptr = chromosome;
	char temp[10000];
	char temp2[10000];
	char tmp1[20];
	char tmp2[20];
	int i, j, k;

	strcpy(string, "");
	for (i = 0; i < numRules; i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		double *ptr2 = ptr;
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			sprintf(temp,"Att %s is " ,ai.getAttributeName(j)->cstr());
			int irr=1;

			if (ai.getTypeOfAttribute(j) == REAL) {
				double minD = ai.getMinDomain(j);
				double maxD = ai.getMaxDomain(j);
				double min = ptr2[0];
				double max = ptr2[1];

				if(min<max) {
					if(min==minD) {
						if(max==maxD) {
							// do nothing
						} else {
							irr=0;
							sprintf(temp2, "[<%f]", max);
							strcat(temp,temp2);
						}
					} else {
						if(max==maxD) {
							irr=0;
							sprintf(temp2, "[>%f]", min);
							strcat(temp,temp2);
						} else {
							irr=0;
							sprintf(tmp1, "%f", min);
							sprintf(tmp2, "%f", max);
							sprintf(temp2, "[%f,%f]", min,max);
							strcat(temp,temp2);
						}
					}
				}
			} else {
				for(k=0;k<tReal->attributeSize[j];k++) {
					if(ptr2[k]) {
						sprintf(temp2,"%s,",ai.getNominalValue(j,k)->cstr());
						strcat(temp,temp2);
					} else {
						irr=0;
					}
				}
				if(temp[strlen(temp) - 1] == ',') temp[strlen(temp) - 1] = 0;
			}
	
			if(!irr) {
				strcat(string, temp);
				strcat(string, "|");
			}
			ptr2 += tReal->attributeSize[j];
		}

		sprintf(temp, "%s\n", ai.getNominalValue(tGlobals->numAttributesMC,(int)ptr2[0])->cstr());
		strcat(string, temp);
		ptr += tReal->ruleSize;
	}

	if(tGlobals->defaultClassPolicy!=DISABLED) {
		sprintf(temp, "%d:Default rule -> %d\n", i,defaultClass);
		strcat(string, temp);
	}
	strcat(string, "\n");
}

void classifier_hyperrect::crossover_1px(classifier_hyperrect * in1,
				    classifier_hyperrect * in2,
				    classifier_hyperrect * out1,
				    classifier_hyperrect * out2)
{
	unsigned int rulep1, rulep2;
	int sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	rulep1 = rnd(0, in1->numRules - 1);
	rulep2 = rnd(0, in2->numRules - 1);
	sizeOfSon1 = rulep1 + in2->numRules - rulep2;
	sizeOfSon2 = rulep2 + in1->numRules - rulep1;

	int cutPoint = rnd(0, tReal->ruleSize - 1);

	out1->length = sizeOfSon1 * tReal->ruleSize;
	out2->length = sizeOfSon2 * tReal->ruleSize;
	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;
	out1->chromosome = new double[tReal->ruleSize * out1->numRules];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new double[tReal->ruleSize * out2->numRules];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	bcopy(in1->chromosome, out1->chromosome,
	      rulep1 * tReal->ruleSize * sizeof(double));

	bcopy(in2->chromosome, out2->chromosome,
	      rulep2 * tReal->ruleSize * sizeof(double));

	bcopy(&in1->chromosome[rulep1 * tReal->ruleSize],
	      &out1->chromosome[rulep1 * tReal->ruleSize],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[rulep2 * tReal->ruleSize],
	      &out2->chromosome[rulep2 * tReal->ruleSize],
	      cutPoint * sizeof(double));

	bcopy(&in1->chromosome[rulep1 * tReal->ruleSize + cutPoint],
	      &out2->chromosome[rulep2 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));
	bcopy(&in2->chromosome[rulep2 * tReal->ruleSize + cutPoint],
	      &out1->chromosome[rulep1 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->
				 chromosome[rulep1 * tReal->ruleSize],
				 &in2->chromosome[rulep2 *
						  tReal->ruleSize],
				 &out1->chromosome[rulep1 *
						   tReal->ruleSize],
				 &out2->chromosome[rulep2 *
						   tReal->ruleSize]);

	bcopy(&in2->chromosome[(rulep2 + 1) * tReal->ruleSize],
	      &out1->chromosome[(rulep1 + 1) * tReal->ruleSize],
	      (in2->numRules -
	       (rulep2 + 1)) * tReal->ruleSize * sizeof(double));

	bcopy(&in1->chromosome[(rulep1 + 1) * tReal->ruleSize],
	      &out2->chromosome[(rulep2 + 1) * tReal->ruleSize],
	      (in1->numRules -
	       (rulep1 + 1)) * tReal->ruleSize * sizeof(double));

}

void classifier_hyperrect::pruneClassifiers(int *classifiersToPrune,
				       int numPruned)
{
	double *newChromosome;
	int i;

	if (numRules == 1 || numPruned == 0) return;

	if(classifiersToPrune[numPruned-1]==numRules) numPruned--;

	newChromosome =
	    new double[(numRules - numPruned) * tReal->ruleSize];
	if (newChromosome == NULL) {
		fprintf(stderr,"Out of memory\n");
		exit(1);
	}

	int countPruned = 0;
	double *ptrOrig = chromosome;
	double *ptrDest = newChromosome;
	for (i = 0; i < numRules; i++) {
		if (countPruned < numPruned) {
			if (i != classifiersToPrune[countPruned]) {
				bcopy(ptrOrig, ptrDest,
				      tReal->ruleSize * sizeof(double));
				ptrDest += tReal->ruleSize;
			} else {
				countPruned++;
			}
		} else {
			bcopy(ptrOrig, ptrDest,
			      tReal->ruleSize * sizeof(double));
			ptrDest += tReal->ruleSize;
		}
		ptrOrig += tReal->ruleSize;
	}

	delete chromosome;
	chromosome = newChromosome;

	numRules -= numPruned;
	length = numRules * tReal->ruleSize;
}

int classifier_hyperrect::classify(instance * ins)
{
	if (tGlobals->ignoreMissingValues && ins->hasMissingValues()) {
		return classifyIgnoreMissing(ins);
	}
	return classifyNoIgnoreMissing(ins);
}

int classifier_hyperrect::classifyIgnoreMissing(instance * ins)
{
	double *ptr = chromosome;
	int i, j, match;
	int numAtributs = tGlobals->numAttributesMC;

	for (i = 0; i < numRules; i++) {
		match = 1;
		double *cAtr = ptr;
		for (j = 0; j < numAtributs && match; j++) {
			if (!ins->isMissing(j)) {
				if (ai.getTypeOfAttribute(j) == REAL) {
					double valueAtr = ins->realValueOfAttribute(j);
					double min = cAtr[0];
					double max = cAtr[1];
					if (min<max && ((valueAtr < min) || (valueAtr > max))) match=0;
				} else {
					int valueAtr
						= ins->valueOfAttribute(j);
					if (cAtr[valueAtr] == 0) {
						match = 0;
					}
				}
			}
			cAtr += tReal->attributeSize[j];
		}
		if (match) return i;
		ptr += tReal->ruleSize;
	}

	if(tGlobals->defaultClassPolicy==DISABLED) return -1;
	return numRules;
}


int classifier_hyperrect::classifyNoIgnoreMissing(instance * ins)
{
	double *ptr = chromosome;
	int i, j, match;
	int numAtributs = tGlobals->numAttributesMC;

	for (i = 0; i < numRules; i++) {
		match = 1;
		double *cAtr = ptr;
		for (j = 0; j < numAtributs && match; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				double valueAtr = ins->realValueOfAttribute(j);
				double min = cAtr[0];
				double max = cAtr[1];
				if (min<max && ((valueAtr < min) || (valueAtr > max))) match=0;
			} else {
				int valueAtr = ins->valueOfAttribute(j);
				if (cAtr[valueAtr] == 0) {
					match = 0;
				}
			}
			cAtr += tReal->attributeSize[j];
		}
		if (match) return i;
		ptr += tReal->ruleSize;
	}

	if(tGlobals->defaultClassPolicy==DISABLED) return -1;
	return numRules;
}

int classifier_hyperrect::ruleMatches(double *rule,instance *ins)
{
	int j, match;
	int numAtributs = tGlobals->numAttributesMC;

	match = 1;
	double *cAtr = rule;
	for (j = 0; j < tGlobals->numAttributesMC && match; j++) {
		if (ai.getTypeOfAttribute(j) == REAL) {
			double valueAtr = ins->realValueOfAttribute(j);
			double min = cAtr[0];
			double max = cAtr[1];
			if (min<max && ((valueAtr < min) || (valueAtr > max))) return 0;
		} else {
			int valueAtr = ins->valueOfAttribute(j);
			if (cAtr[valueAtr] == 0) {
				return 0;
			}
		}
		cAtr += tReal->attributeSize[j];
	}

	return 1;
}


void classifier_hyperrect::crossover_smart(classifier **parents, int num)
{
	int i, j, k;

	modif = 1;
	if(chromosome) delete chromosome;

	JVector<double *>ptrs;
	JVector<matchProfileAgent *>profiles;
	int ruleCount=0;

	for(k=0;k<num;k++) {
		classifier_hyperrect *parent=(classifier_hyperrect *)parents[k];
		double *ptr=parent->chromosome;
		for(i=0;i<parent->numRules;ptr+=tReal->ruleSize,i++) {
			double *newRule= new double[tReal->ruleSize];
			bcopy(ptr,newRule,sizeof(double)*tReal->ruleSize);

			matchProfileAgent *mpa = evaluateRuleCX(newRule);
			ptrs.addElement(newRule);
			profiles.addElement(mpa);
		}
	}

	JVector<int> ruleOrder;
	findOrder(profiles,ptrs,ruleOrder);

	numRules = ruleOrder.size();
	chromosome = new double[numRules * tReal->ruleSize];
	length = tReal->ruleSize * numRules;
	for (i = 0; i < numRules; i++) {
		bcopy(ptrs[ruleOrder[i]], &chromosome[i * tReal->ruleSize],
		      sizeof(double) * tReal->ruleSize);
	}

	for(i=0;i<profiles.size();i++) {
		delete profiles[i];
	}
	for(i=0;i<ptrs.size();i++) {
		delete ptrs[i];
	}
}

void classifier_hyperrect::findOrder(JVector<matchProfileAgent *>&profiles
	,JVector<double *>&ptrs,JVector<int> &ruleOrder)
{
	unsigned long long numInst=is->getNumInstancesOfIteration();
	int i,j;

	ruleOrderAgent *best=NULL;
	for(j=0;j<tCross->repetitionsRuleOrdering;j++) {
		int totalRules=profiles.size();
		Sampling samp(totalRules);
		ruleOrderAgent *roa = new ruleOrderAgent(defaultClass,numInst,totalRules);

		for(i=0;i<totalRules;i++) {
			int rule=samp.getSample();
			roa->insertRule(rule,profiles[rule]);
		}
		roa->refineRuleSet();

		if(best==NULL) {
			best=roa;
		} else {
			double accOld=best->getAccuracy();
			double accNew=roa->getAccuracy();
			if(accNew>accOld) {
				delete best;
				best=roa;
			} else if(accNew==accOld
				&& roa->getNumRules()<best->getNumRules()) {
				delete best;
				best=roa;
			} else {
				delete roa;
			}
		}
	}

	best->copyOrder(ruleOrder);
	if(ruleOrder.size()==0) {
		int pos=rnd(0,profiles.size()-1);
		ruleOrder.addElement(pos);
	}
	delete best;
}

matchProfileAgent *classifier_hyperrect::evaluateRuleCX(double *rule)
{
	unsigned long long i;
	int cl = (int)rule[tReal->ruleSize-1];
	unsigned long long max = is->getNumInstancesOfIteration();
	instance **instances=is->getInstancesOfIteration();

	matchProfileAgent *mpa=new matchProfileAgent(max,cl);

	for (i = 0; i < max; i++) {
		if (ruleMatches(rule, instances[i])) {
			if (instances[i]->getClass() == cl) {
				mpa->addOK(i);
			} else {
				mpa->addKO(i);
			}
		}
	}
	mpa->generateProfiles();

	return mpa;
}

