#include "classifier_instances.h"
#include "random.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "timerGlobals.h"
#include "timerHierar.h"
#include "timerRealKR.h"
#include "timerMutation.h"
#include "attributesInfo.h"

extern attributesInfo ai;
extern timerGlobals *tGlobals;
extern timerHierar *tHierar;
extern timerMutation *tMut;
extern timerRealKR *tReal;
extern Random rnd;

double classifier_instances::computateTheoryLength(agentPerformanceTraining *ap)
{
	theoryLength=tReal->instanceTheoryLength * aliveClassifiers;
	return theoryLength;
}

classifier_instances::classifier_instances()
{
	numInstances =
	    rnd(tGlobals->minClassifiersInit,
		tGlobals->maxClassifiersInit);
	length = tGlobals->numAttributes * numInstances;
	chromosome = new double[length];
	if (chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	classifier_instances::initializeChromosome();
}

classifier_instances::
classifier_instances(const classifier_instances & orig, int son)
{
	// Copiem les dades esttiques
	*this = orig;

	if (!son) {
		chromosome = new double[length];
		if (chromosome == NULL) {
			perror("out of memory");
			exit(1);
		}
		bcopy(orig.chromosome, chromosome,
		      length * sizeof(double));
	}
}

classifier_instances::~classifier_instances()
{
	delete chromosome;
}

void classifier_instances::initializeChromosome()
{
	int i, j, k;

	for (i = 0; i < numInstances; i++) {
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				setGene(i, j,
					!rnd * ai.getSizeDomain(j) +
					ai.getMinDomain(j));
			} else {
				setGene(i, j,
					rnd(0,
					    ai.getNumValuesAttribute(j) -
					    1));
			}
		}

		setGene(i, tGlobals->numAttributesMC,
			rnd(0, ai.getNumClasses() - 1));
	}
}

void classifier_instances::setGene(short int rule, short int gene,
				   double nfo)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numInstances)) {
		fprintf(stderr,"Invalid access at setGene %d %d\n", rule, gene);
		exit(1);
	}
#endif
	chromosome[rule * tGlobals->numAttributes + gene] = nfo;
	modif = 1;
}

double classifier_instances::getGene(short int rule, short int gene)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numInstances)) {
		fprintf(stderr,"Invalid access at getGene %d %d", rule, gene);
		exit(1);
	}
#endif
	return chromosome[rule * tGlobals->numAttributes + gene];
}

void classifier_instances::crossover(classifier * in,
				     classifier * out1, classifier * out2)
{
	switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this, (classifier_instances *) in,
			      (classifier_instances *) out1,
			      (classifier_instances *) out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_instances *) in,
			      (classifier_instances *) out1,
			      (classifier_instances *) out2);
	}
}

void classifier_instances::mutation()
{
	int i;
	int rule, attribute, valor;

	modif = 1;
	tMut->mutationCounter++;

        rule = rnd(0,numInstances-1);
        if(!rnd<0.10) {
                attribute = tGlobals->numAttributesMC;
        } else {
                attribute=rnd(0,tGlobals->numAttributesMC-1);
        }

	if (attribute != tGlobals->numAttributesMC) {
		if (ai.getTypeOfAttribute(attribute) == REAL) {
			setGene(rule, attribute,
				!rnd * ai.getSizeDomain(attribute) +
				ai.getMinDomain(attribute));
		} else {
			setGene(rule, attribute,
				rnd(0,
				    ai.getNumValuesAttribute(attribute) -
				    1));
		}
	} else {
		int oldValue = (int) getGene(rule, attribute);
		int newValue;
		do {
			newValue = rnd(0, ai.getNumClasses() - 1);
		} while (oldValue == newValue);
		setGene(rule, attribute, newValue);
	}
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

void classifier_instances::specialCrossover(double *parent1,
					    double *parent2, double *son1,
					    double *son2)
{
	int i;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		if (ai.getTypeOfAttribute(i) == REAL) {
			tReal->specialCrossover(parent1[i], parent2[i],
						son1[i], son2[i],
						ai.getMinDomain(i),
						ai.getMaxDomain(i));
		}
	}
}

void classifier_instances::crossover_2px(classifier_instances * in1,
					 classifier_instances * in2,
					 classifier_instances * out1,
					 classifier_instances * out2)
{
	int i1rule1, i1rule2, i2rule1, i2rule2;
	int cutPoint1, cutPoint2, sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	i1rule1 = rnd(0, in1->numInstances - 1);
	i1rule2 = rnd(0, in1->numInstances - 1);
	if (i1rule1 > i1rule2)
		swap(i1rule1, i1rule2);

	i2rule1 = rnd(0, in2->numInstances - 1);
	i2rule2 = rnd(0, in2->numInstances - 1);
	if (i2rule1 > i2rule2)
		swap(i2rule1, i2rule2);

	sizeOfSon1 =
	    i1rule1 + i2rule2 - i2rule1 + in1->numInstances - i1rule2;
	sizeOfSon2 =
	    i2rule1 + i1rule2 - i1rule1 + in2->numInstances - i2rule2;

	cutPoint1 = rnd(0, tGlobals->numAttributesMC);
	cutPoint2 = rnd(0, tGlobals->numAttributesMC);

	if (i1rule1 == i1rule2 || i2rule1 == i2rule2) {
		if (cutPoint1 > cutPoint2)
			swap(cutPoint1, cutPoint2);
	}

	out1->length = sizeOfSon1 * tGlobals->numAttributes;
	out2->length = sizeOfSon2 * tGlobals->numAttributes;

	out1->numInstances = sizeOfSon1;
	out2->numInstances = sizeOfSon2;
	out1->chromosome = new double[out1->length];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new double[out2->length];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	int cutPoint = cutPoint1;

	bcopy(chromosome, out1->chromosome,
	      i1rule1 * tGlobals->numAttributes * sizeof(double));
	bcopy(in2->chromosome, out2->chromosome,
	      i2rule1 * tGlobals->numAttributes * sizeof(double));

	bcopy(&in1->chromosome[i1rule1 * tGlobals->numAttributes],
	      &out1->chromosome[i1rule1 * tGlobals->numAttributes],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[i2rule1 * tGlobals->numAttributes],
	      &out2->chromosome[i2rule1 * tGlobals->numAttributes],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[i2rule1 * tGlobals->numAttributes + cutPoint],
	      &out1->chromosome[i1rule1 * tGlobals->numAttributes + cutPoint],
	      (tGlobals->numAttributes - cutPoint) * sizeof(double));
	bcopy(&in1->chromosome[i1rule1 * tGlobals->numAttributes + cutPoint],
	      &out2->chromosome[i2rule1 * tGlobals->numAttributes + cutPoint],
	      (tGlobals->numAttributes - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->chromosome[i1rule1 * tGlobals->numAttributes],
				 &in2->chromosome[i2rule1 * tGlobals->numAttributes],
				 &out1->chromosome[i1rule1 *
						   tGlobals->numAttributes],
				 &out2->chromosome[i2rule1 *
						   tGlobals->numAttributes]);

	if (i2rule1 != i2rule2)
		bcopy(&in2->chromosome[(i2rule1 + 1) * tGlobals->numAttributes],
		      &out1->chromosome[(i1rule1 + 1) * tGlobals->numAttributes],
		      (i2rule2 -
		       (i2rule1 + 1)) * tGlobals->numAttributes * sizeof(double));
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[(i1rule1 + 1) * tGlobals->numAttributes],
		      &out2->chromosome[(i2rule1 + 1) * tGlobals->numAttributes],
		      (i1rule2 -
		       (i1rule1 + 1)) * tGlobals->numAttributes * sizeof(double));


	cutPoint = cutPoint2;

	if (i2rule2 != i2rule1)
		bcopy(&in2->chromosome[i2rule2 * tGlobals->numAttributes],
		      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
					tGlobals->numAttributes],
		      cutPoint * sizeof(double));
	if (i1rule2 != i1rule1)
		bcopy(&in1->chromosome[i1rule2 * tGlobals->numAttributes],
		      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
					tGlobals->numAttributes],
		      cutPoint * sizeof(double));

	bcopy(&in1->chromosome[i1rule2 * tGlobals->numAttributes + cutPoint],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
				tGlobals->numAttributes + cutPoint],
	      (tGlobals->numAttributes - cutPoint) * sizeof(double));
	bcopy(&in2->chromosome[i2rule2 * tGlobals->numAttributes + cutPoint],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
				tGlobals->numAttributes + cutPoint],
	      (tGlobals->numAttributes - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->chromosome[i1rule2 * tGlobals->numAttributes],
				 &in2->chromosome[i2rule2 * tGlobals->numAttributes],
				 &out1->
				 chromosome[(i1rule1 + i2rule2 -
					     i2rule1) * tGlobals->numAttributes],
				 &out2->
				 chromosome[(i2rule1 + i1rule2 -
					     i1rule1) * tGlobals->numAttributes]);

	bcopy(&in1->chromosome[(i1rule2 + 1) * tGlobals->numAttributes],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1 + 1) *
				tGlobals->numAttributes],
	      (in1->numInstances -
	       (i1rule2 + 1)) * tGlobals->numAttributes * sizeof(double));
	bcopy(&in2->chromosome[(i2rule2 + 1) * tGlobals->numAttributes],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1 + 1) *
				tGlobals->numAttributes],
	      (in2->numInstances -
	       (i2rule2 + 1)) * tGlobals->numAttributes * sizeof(double));

}

void classifier_instances::dumpPhenotype(char *string)
{
	double *ptr = chromosome;
	char temp[10000];
	int i, j;

	strcpy(string, "");
	for (i = 0; i < numInstances; i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				sprintf(temp, "%f", ptr[j]);
			} else {
				sprintf(temp, "%s",
					ai.getNominalValue(j,
							   (int) ptr[j])->cstr());
			}
			strcat(string, temp);
			strcat(string, "|");
		}
		sprintf(temp, "%d\n", (int) ptr[tGlobals->numAttributesMC]);
		strcat(string, temp);
		ptr += tGlobals->numAttributes;
	}
	strcat(string, "\n");
}

void classifier_instances::crossover_1px(classifier_instances * in1,
					 classifier_instances * in2,
					 classifier_instances * out1,
					 classifier_instances * out2)
{
	unsigned int rulep1, rulep2;
	int sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	rulep1 = rnd(0, in1->numInstances - 1);
	rulep2 = rnd(0, in2->numInstances - 1);
	sizeOfSon1 = rulep1 + in2->numInstances - rulep2;
	sizeOfSon2 = rulep2 + in1->numInstances - rulep1;

	int cutPoint = rnd(0, tGlobals->numAttributesMC);

	out1->length = sizeOfSon1 * tGlobals->numAttributes;
	out2->length = sizeOfSon2 * tGlobals->numAttributes;
	out1->numInstances = sizeOfSon1;
	out2->numInstances = sizeOfSon2;
	out1->chromosome = new double[out1->length];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new double[out2->length];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	bcopy(in1->chromosome, out1->chromosome,
	      rulep1 * tGlobals->numAttributes * sizeof(double));

	bcopy(in2->chromosome, out2->chromosome,
	      rulep2 * tGlobals->numAttributes * sizeof(double));

	bcopy(&in1->chromosome[rulep1 * tGlobals->numAttributes],
	      &out1->chromosome[rulep1 * tGlobals->numAttributes],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[rulep2 * tGlobals->numAttributes],
	      &out2->chromosome[rulep2 * tGlobals->numAttributes],
	      cutPoint * sizeof(double));

	bcopy(&in1->chromosome[rulep1 * tGlobals->numAttributes + cutPoint],
	      &out2->chromosome[rulep2 * tGlobals->numAttributes + cutPoint],
	      (tGlobals->numAttributes - cutPoint) * sizeof(double));
	bcopy(&in2->chromosome[rulep2 * tGlobals->numAttributes + cutPoint],
	      &out1->chromosome[rulep1 * tGlobals->numAttributes + cutPoint],
	      (tGlobals->numAttributes - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->chromosome[rulep1 * tGlobals->numAttributes],
				 &in2->chromosome[rulep2 * tGlobals->numAttributes],
				 &out1->chromosome[rulep1 * tGlobals->numAttributes],
				 &out2->chromosome[rulep2 *
						   tGlobals->numAttributes]);

	bcopy(&in2->chromosome[(rulep2 + 1) * tGlobals->numAttributes],
	      &out1->chromosome[(rulep1 + 1) * tGlobals->numAttributes],
	      (in2->numInstances -
	       (rulep2 + 1)) * tGlobals->numAttributes * sizeof(double));

	bcopy(&in1->chromosome[(rulep1 + 1) * tGlobals->numAttributes],
	      &out2->chromosome[(rulep2 + 1) * tGlobals->numAttributes],
	      (in1->numInstances -
	       (rulep1 + 1)) * tGlobals->numAttributes * sizeof(double));
}

void classifier_instances::pruneClassifiers(int *classifiersToPrune,
					    int numPruned)
{
	double *newChromosome;
	int i;

	if (numInstances == 1 || numPruned == 0)
		return;

	newChromosome =
	    new double[(numInstances - numPruned) * tGlobals->numAttributes];
	if (newChromosome == NULL) {
		fprintf(stderr,"Out of memory\n");
		exit(1);
	}

	int countPruned = 0;
	double *ptrOrig = chromosome;
	double *ptrDest = newChromosome;
	for (i = 0; i < numInstances; i++) {
		if (countPruned < numPruned) {
			if (i != classifiersToPrune[countPruned]) {
				bcopy(ptrOrig, ptrDest,
				      tGlobals->numAttributes * sizeof(double));
				ptrDest += tGlobals->numAttributes;
			} else {
				countPruned++;
			}
		} else {
			bcopy(ptrOrig, ptrDest,
			      tGlobals->numAttributes * sizeof(double));
			ptrDest += tGlobals->numAttributes;
		}
		ptrOrig += tGlobals->numAttributes;
	}

	delete chromosome;
	chromosome = newChromosome;

	numInstances -= numPruned;
	length = numInstances * tGlobals->numAttributes;
}

int classifier_instances::classify(instance * ins)
{
	double *ptr = chromosome;
	int i, j;
	double distMin;
	int instMin;

	if (tGlobals->ignoreMissingValues && ins->hasMissingValues()) {
		for (i = 0; i < numInstances; i++) {
			double dist = 0;
			for (j = 0; j < tGlobals->numAttributesMC; j++) {
				if (!ins->isMissing(j)) {
					if (ai.getTypeOfAttribute(j) ==
					    REAL) {
						double distTmp =
						    ins->
						    realValueOfAttribute(j)
						    - ptr[j];
						distTmp /= ai.
						    getSizeDomain(j);
						distTmp *= distTmp;
						dist += distTmp;
					} else {
						if (ptr[j] !=
						    ins->
						    valueOfAttribute(j)) {
							dist++;
						}
					}
				}
			}

			if (i) {
				if (dist < distMin) {
					distMin = dist;
					instMin = i;
				}
			} else {
				distMin = dist;
				instMin = 0;
			}

			ptr += tGlobals->numAttributes;
		}
	} else {
		for (i = 0; i < numInstances; i++) {
			double dist = 0;
			for (j = 0; j < tGlobals->numAttributesMC; j++) {
				if (ai.getTypeOfAttribute(j) == REAL) {
					double dom=ai.getSizeDomain(j);
					if(dom>0) {
						double distTmp =
						    ins->realValueOfAttribute(j) -
						    ptr[j];
						distTmp /= ai.getSizeDomain(j);
						distTmp *= distTmp;
						dist += distTmp;
					}
				} else {
					if (ptr[j] !=
					    ins->valueOfAttribute(j)) {
						dist++;
					}
				}
			}

			if (i) {
				if (dist < distMin) {
					distMin = dist;
					instMin = i;
				}
			} else {
				distMin = dist;
				instMin = 0;
			}

			ptr += tGlobals->numAttributes;
		}

	}

	return instMin;
}

int classifier_instances::getClass(int rule)
{
	return (int) chromosome[rule * tGlobals->numAttributes + tGlobals->numAttributesMC];
}
