#include "classifier_lcs.h"

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>

#include "attributesInfo.h"
#include "random.h"
#include "timerGlobals.h"
#include "timerHierar.h"
#include "timerMutation.h"
#include "timerMDL.h"
#include "timerSymbolicKR.h"

extern attributesInfo ai;
extern timerSymbolicKR *tSymbolic;
extern timerGlobals *tGlobals;
extern timerHierar *tHierar;
extern timerMutation *tMut;
extern timerMDL *tMDL;
extern Random rnd;

double classifier_lcs::computateTheoryLength(agentPerformanceTraining *ap)
{
	int i,j;
	unsigned char *ptr=chromosome;
	theoryLength=0;

	for(i=0;i<numRules;i++) {
		for(j=0;j<tGlobals->numAttributesMC;j++) {
			theoryLength++;
			if(ptr[j]<ai.getNumValuesAttribute(j))
				theoryLength += 3.0;
		}
		ptr+=tGlobals->numAttributes;
	}
        if(tGlobals->defaultClassPolicy!=DISABLED)
                //if(ap->getActivationsOfClassifier(numRules))
                        theoryLength+=tSymbolic->defaultRuleTheoryLength;

	return theoryLength;
}

classifier_lcs::classifier_lcs()
{
	numRules = rnd (tGlobals->minClassifiersInit, tGlobals->maxClassifiersInit);
	length = tGlobals->numAttributes * numRules;
	chromosome = new unsigned char[length];
	if (chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

        if(tGlobals->defaultClassPolicy==AUTO) {
                defaultClass=rnd(0,ai.getNumClasses()-1);
        } else  {
                defaultClass=tGlobals->defaultClass;
        }


	initializeChromosome();
}


classifier_lcs::
classifier_lcs(const classifier_lcs & orig,int son)
{
	// Copiem les dades esttiques
	*this = orig;

	if(!son) {
		chromosome = new unsigned char[length];
		if (chromosome == NULL) {
			perror("out of memory");
			exit(1);
		}

		bcopy(orig.chromosome, chromosome, length);
	}
}

classifier_lcs::~classifier_lcs()
{
	delete chromosome;
}

void classifier_lcs::initializeChromosome()
{
	int i, j;

	for (i = 0; i < numRules; i++) {
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if(!rnd<tSymbolic->probSharp)
				setGene(i,j,ai.getNumValuesAttribute(j));
			else
				setGene(i,j,rnd(0,ai.getNumValuesAttribute(j)-1));
		}
		int cl;
		do {
			cl=rnd(0,ai.getNumClasses()-1);
		} while(tGlobals->defaultClassPolicy!=DISABLED && cl==defaultClass);
		setGene(i, tGlobals->numAttributesMC, cl);
	}
}

void classifier_lcs::setGene(short int rule, short int gene, unsigned char nfo)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules)) {
		fprintf(stderr,"Invalid access to setGene %d %d", rule, gene);
		exit(1);
	}
#endif
	chromosome[rule * tGlobals->numAttributes + gene] = nfo;
	modif = 1;
}

unsigned char classifier_lcs::getGene(short int rule, short int gene)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules)) {
		fprintf(stderr,"Invalid access to getGene %d %d", rule,
		       gene);
		exit(1);
	}
#endif
	return chromosome[rule * tGlobals->numAttributes + gene];
}

void classifier_lcs::crossover(classifier * in,
				      classifier * out1,
				      classifier * out2)
{
	switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this,(classifier_lcs *)in, (classifier_lcs *)out1, (classifier_lcs *)out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_lcs *)in, (classifier_lcs *)out1, (classifier_lcs *)out2);
	}


}

void classifier_lcs::mutation()
{
	int i;
	unsigned short int rule, attribute,valor;

	// Modificarem el chromosome
	modif = 1;
	tMut->mutationCounter++;

        rule = rnd(0,numRules-1);
        if(tGlobals->numClasses>1 && !rnd<0.10) {
                attribute = tGlobals->numAttributesMC;
        } else {
                attribute=rnd(0,tGlobals->numAttributesMC-1);
        }

	if (attribute < tGlobals->numAttributesMC) {
		int newValue;
		if(!rnd < tSymbolic->probSharp) {
			newValue = ai.getNumValuesAttribute(attribute);
		} else {
			newValue = rnd (0, ai.getNumValuesAttribute(attribute)-1);
		}
		setGene(rule, attribute, newValue);
	} else {
		if(tGlobals->numClasses>1) {
			int oldValue = getGene(rule,attribute);
			int newValue;
			do {
				newValue = rnd (0, ai.getNumClasses()-1);
			} while(newValue==oldValue || tGlobals->defaultClassPolicy!=DISABLED && newValue==defaultClass);
			setGene(rule, attribute, newValue);
		}
	}
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

void classifier_lcs::crossover_2px(classifier_lcs * in1,classifier_lcs * in2,
					     classifier_lcs * out1,
					     classifier_lcs * out2)
{
	int i1rule1, i1rule2, i2rule1, i2rule2;
	int cutPoint1, cutPoint2, sizeSon1, sizeSon2;

	out1->modif = out2->modif = 1;

	i1rule1 = rnd (0, in1->numRules - 1);
	i1rule2 = rnd (0, in1->numRules - 1);
	if (i1rule1 > i1rule2)
		swap(i1rule1, i1rule2);

	i2rule1 = rnd (0, in2->numRules - 1);
	i2rule2 = rnd (0, in2->numRules - 1);
	if (i2rule1 > i2rule2)
		swap(i2rule1, i2rule2);

	sizeSon1 = i1rule1 + i2rule2 - i2rule1 + in1->numRules - i1rule2;
	sizeSon2 = i2rule1 + i1rule2 - i1rule1 + in2->numRules - i2rule2;

	cutPoint1 = rnd (0, tGlobals->numAttributesMC);
	cutPoint2 = rnd (0, tGlobals->numAttributesMC);

	if(i1rule1==i1rule2 || i2rule1==i2rule2) {
		if(cutPoint1>cutPoint2)
			swap(cutPoint1,cutPoint2);
	}

	out1->length = sizeSon1 * tGlobals->numAttributes;
	out2->length = sizeSon2 * tGlobals->numAttributes;
	out1->numRules = sizeSon1;
	out2->numRules = sizeSon2;

	out1->chromosome = new unsigned char[out1->length];
	out2->chromosome = new unsigned char[out2->length];

	// Rules before the first cut point
	bcopy(in1->chromosome, out1->chromosome, i1rule1 * tGlobals->numAttributes);
	bcopy(in2->chromosome, out2->chromosome, i2rule1 * tGlobals->numAttributes);

	// Attributes before first cut point in affected rule
	bcopy(&in1->chromosome[i1rule1 * tGlobals->numAttributes],
	      &out1->chromosome[i1rule1 * tGlobals->numAttributes], cutPoint1);
	bcopy(&in2->chromosome[i2rule1 * tGlobals->numAttributes],
	      &out2->chromosome[i2rule1 * tGlobals->numAttributes], cutPoint1);
	// Attributes after first cut point in affected rule
	bcopy(&in2->chromosome[i2rule1 * tGlobals->numAttributes + cutPoint1],
	      &out1->chromosome[i1rule1 * tGlobals->numAttributes + cutPoint1], tGlobals->numAttributes - cutPoint1);
	bcopy(&in1->chromosome[i1rule1 * tGlobals->numAttributes + cutPoint1],
	      &out2->chromosome[i2rule1 * tGlobals->numAttributes + cutPoint1], tGlobals->numAttributes - cutPoint1);


	// Rules between the two cut points
	if(i2rule1!=i2rule2)
	bcopy(&in2->chromosome[(i2rule1 + 1) * tGlobals->numAttributes],
	      &out1->chromosome[(i1rule1 + 1) * tGlobals->numAttributes],
	      (i2rule2 - (i2rule1 + 1)) * tGlobals->numAttributes);
	if(i1rule1!=i1rule2)
	bcopy(&in1->chromosome[(i1rule1 + 1) * tGlobals->numAttributes],
	      &out2->chromosome[(i2rule1 + 1) * tGlobals->numAttributes],
	      (i1rule2 - (i1rule1 + 1)) * tGlobals->numAttributes);


	// Attributes before second cut point in affected rule
	if(i2rule1!=i2rule2)
	bcopy(&in2->chromosome[i2rule2 * tGlobals->numAttributes],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) * tGlobals->numAttributes],
	      cutPoint2);
	if(i1rule1!=i1rule2)
	bcopy(&in1->chromosome[i1rule2 * tGlobals->numAttributes],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) * tGlobals->numAttributes],
	      cutPoint2);
	// Attributes after second cut point in affected rule
	bcopy(&in1->chromosome[i1rule2 * tGlobals->numAttributes + cutPoint2],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) * tGlobals->numAttributes +
			       cutPoint2], tGlobals->numAttributes - cutPoint2);
	bcopy(&in2->chromosome[i2rule2 * tGlobals->numAttributes + cutPoint2],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) * tGlobals->numAttributes +
			       cutPoint2], tGlobals->numAttributes - cutPoint2);

	// Rules after second cut point 
	bcopy(&in1->chromosome[(i1rule2 + 1) * tGlobals->numAttributes],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1 + 1) *
			       tGlobals->numAttributes], (in1->numRules - (i1rule2 + 1)) * tGlobals->numAttributes);
	bcopy(&in2->chromosome[(i2rule2 + 1) * tGlobals->numAttributes],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1 + 1) *
			       tGlobals->numAttributes],
	      (in2->numRules - (i2rule2 + 1)) * tGlobals->numAttributes);

}

void classifier_lcs::dumpPhenotype(char *string)
{
	unsigned char *ptr=chromosome;
	char temp[10000];
	int i,j;

	strcpy(string, "");
	for(i=0;i<numRules;i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if(ptr[j]==ai.getNumValuesAttribute(j))
				strcpy(temp,"#|");
			else
				sprintf(temp,"%d|",ptr[j]);
			strcat(string, temp);
		}
                int cl=(int)ptr[j];
                sprintf(temp, "%d\n", cl);

		strcat(string, temp);
		ptr+=tGlobals->numAttributes;
	}
        if(tGlobals->defaultClassPolicy!=DISABLED) {
                sprintf(temp, "%d:Default rule -> %d\n", i,defaultClass);
                strcat(string, temp);
        }
	strcat(string, "\n");
}

void classifier_lcs::crossover_1px(
	classifier_lcs * in1,
	classifier_lcs * in2,
	classifier_lcs * out1,
	classifier_lcs * out2)
{
	unsigned int cutPoint, ruleParent1, ruleParent2;
	int sizeSon1, sizeSon2;

	out1->modif = out2->modif = 1;

	ruleParent1 = rnd (0, in1->numRules - 1);
	ruleParent2 = rnd (0, in2->numRules - 1);
	
	sizeSon1 = (ruleParent1) + 1 + (in2->numRules - (ruleParent2 + 1));
	sizeSon2 = (ruleParent2) + 1 + (in1->numRules - (ruleParent1 + 1));

	cutPoint = rnd (0, tGlobals->numAttributesMC);

	out1->length = sizeSon1 * tGlobals->numAttributes;
	out2->length = sizeSon2 * tGlobals->numAttributes;
	out1->numRules = sizeSon1;
	out2->numRules = sizeSon2;
	
	out1->chromosome = new unsigned char[out1->length];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new unsigned char[out2->length];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	// Rules before cut point
	bcopy(in1->chromosome, out1->chromosome, ruleParent1 * tGlobals->numAttributes);

	bcopy(in2->chromosome, out2->chromosome, ruleParent2 * tGlobals->numAttributes);

	// Attributes before cut point in affected rule
	bcopy(&in1->chromosome[ruleParent1 * tGlobals->numAttributes],
	      &out1->chromosome[ruleParent1 * tGlobals->numAttributes], cutPoint);
	bcopy(&in2->chromosome[ruleParent2 * tGlobals->numAttributes],
	      &out2->chromosome[ruleParent2 * tGlobals->numAttributes], cutPoint);
	// Attributes after cut point in affected rule
	bcopy(&in1->chromosome[ruleParent1 * tGlobals->numAttributes + cutPoint],
	      &out2->chromosome[ruleParent2 * tGlobals->numAttributes + cutPoint], tGlobals->numAttributes - cutPoint);
	bcopy(&in2->chromosome[ruleParent2 * tGlobals->numAttributes + cutPoint],
	      &out1->chromosome[ruleParent1 * tGlobals->numAttributes + cutPoint], tGlobals->numAttributes - cutPoint);


	// Rules after cut point
	bcopy(&in2->chromosome[(ruleParent2 + 1) * tGlobals->numAttributes],
	      &out1->chromosome[(ruleParent1 + 1) * tGlobals->numAttributes],
	      (in2->numRules - (ruleParent2 + 1)) * tGlobals->numAttributes);
	bcopy(&in1->chromosome[(ruleParent1 + 1) * tGlobals->numAttributes],
	      &out2->chromosome[(ruleParent2 + 1) * tGlobals->numAttributes],
	      (in1->numRules - (ruleParent1 + 1)) * tGlobals->numAttributes);
}


void classifier_lcs::pruneClassifiers(int *classifiersToPrune,int numPruned)
{
	unsigned char *newChromosome;
	int i;

	if (numRules == 1 || numPruned==0) return;

	newChromosome = new unsigned char[(numRules - numPruned) * tGlobals->numAttributes];
	if (newChromosome == NULL) {
		fprintf(stderr,"Out of memory\n");
		exit(1);
	}

	int countPruned=0;
	unsigned char *ptrOrig=chromosome;
	unsigned char *ptrDest=newChromosome;
	for(i=0;i<numRules;i++) {
                if (countPruned < numPruned) {
                        if (i != classifiersToPrune[countPruned]) {
                                bcopy(ptrOrig, ptrDest, tGlobals->numAttributes);
                                ptrDest += tGlobals->numAttributes;
                        } else {
                                countPruned++;
                        }
                } else {
                        bcopy(ptrOrig, ptrDest, tGlobals->numAttributes);
                        ptrDest += tGlobals->numAttributes;
                }
		ptrOrig+=tGlobals->numAttributes;
	}

	delete chromosome;
	chromosome = newChromosome;

	numRules-=numPruned;
	length = numRules * tGlobals->numAttributes;
}

int classifier_lcs::classify(instance *ins)
{
	unsigned char *cRegla = chromosome;
        int i, j, match;

	if(tGlobals->ignoreMissingValues && ins->hasMissingValues()) {
                for(i=0;i<numRules;i++) {
                        match=1;
                        for(j = 0; j < tGlobals->numAttributesMC && match; j++) {
                                if (!ins->isMissing(j)
					&& cRegla[j]<ai.getNumValuesAttribute(j)
					&& cRegla[j]!= ins->valueOfAttribute(j)) {
                                        match = 0;
                                }
                        }
                        if(match) return i;
                        cRegla+=tGlobals->numAttributes;
                }
        } else {
                for(i=0;i<numRules;i++) {
                        match=1;
                        for(j = 0; j < tGlobals->numAttributesMC && match; j++) {
                                if (cRegla[j]<ai.getNumValuesAttribute(j)
					&& cRegla[j]!= ins->valueOfAttribute(j)) {
                                        match = 0;
                                }
                        }
                        if(match) return i;
                        cRegla+=tGlobals->numAttributes;
                }

        }

        if(tGlobals->defaultClassPolicy==DISABLED) return -1;
        return numRules;
}

int classifier_lcs::getClass(int rule)
{
        if(rule==numRules) return defaultClass;
        return (int) chromosome[rule * tGlobals->numAttributes +
                tGlobals->numAttributesMC];
}


