#include "classifier_ubr.h"
#include "random.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <strings.h>
#include "attributesInfo.h"
#include "timerGlobals.h"
#include "timerMutation.h"
#include "timerCrossover.h"
#include "timerMDL.h"
#include "timerHierar.h"
#include "timerRealKR.h"
#include "instanceSet.h"
#include "ruleOrderAgent.h"
#include "sampling.h"

extern attributesInfo ai;
extern timerGlobals *tGlobals;
extern timerHierar *tHierar;
extern timerMutation *tMut;
extern timerCrossover *tCross;
extern timerMDL *tMDL;
extern timerRealKR *tReal;
extern instanceSet *is;
extern Random rnd;

static void swap(double &a,double &b)
{
	double temp=a;
	a=b;
	b=temp;
}

double classifier_ubr::computateTheoryLength(agentPerformanceTraining *ap)
{
	int i, j, k;
	double *ptr = chromosome;
	theoryLength = 0.0;

	for (i = 0; i < numRules; i++) {
		if(ap->getActivationsOfClassifier(i)) {
			for (j = 0; j < tGlobals->numAttributesMC; j++) {
				if (ai.getTypeOfAttribute(j) == REAL) {
					double min=ptr[0];
					double max=ptr[1];
					double vmin=ptr[2];
					double vmax=ptr[3];
					if(vmin>vmax) swap(vmin,vmax);
					double dmin=ai.getMinDomain(j);
					double dmax=ai.getMaxDomain(j);

					if(min) vmin=dmin;
					if(max) vmax=dmax;

					theoryLength+=5.0;
					if(dmax-dmin>0) 
						theoryLength += 5.0 - 
							(vmax-vmin)
							/ (dmax-dmin) * 5.0;
				} else {
					double countFalses = 0;
					int numValues=tReal->attributeSize[j];
					for(k=0;k<numValues;k++) {
						if(!ptr[k]) countFalses++;
					}
					theoryLength+=numValues + countFalses;
				}
				ptr += tReal->attributeSize[j];
			}
			ptr++;
		} else {
			ptr+=tReal->ruleSize;
		}
	}
	if(tGlobals->defaultClassPolicy!=DISABLED) {
		//if(ap->getActivationsOfClassifier(numRules))
			theoryLength+=tReal->defaultRuleTheoryLength;
	}

	return theoryLength;
}

classifier_ubr::classifier_ubr(int useMPI,int originNode)
{
	numRules =
	    rnd(tGlobals->minClassifiersInit,
		tGlobals->maxClassifiersInit);
	length = tReal->ruleSize * numRules;
	chromosome = new double[tReal->ruleSize * numRules];
	if (chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	if(tGlobals->defaultClassPolicy==AUTO) {
		defaultClass=rnd(0,ai.getNumClasses()-1);
	} else  {
		defaultClass=tGlobals->defaultClass;
	}



	classifier_ubr::initializeChromosome();
}

classifier_ubr::classifier_ubr(const classifier_ubr & orig, int son)
{
	*this = orig;

	if (!son) {
		chromosome = new double[tReal->ruleSize * numRules];
		if (chromosome == NULL) {
			perror("out of memory");
			exit(1);
		}
		bcopy(orig.chromosome, chromosome,
		      numRules * tReal->ruleSize * sizeof(double));
	} else {
		chromosome = NULL;
	}
}

classifier_ubr::~classifier_ubr()
{
	delete chromosome;
}

void classifier_ubr::initializeChromosome()
{
	int i, j, k;

	for (i = 0; i < numRules; i++) {
		instance *ins=NULL;
		if(tGlobals->smartInit) {
			if(tGlobals->defaultClassPolicy!=DISABLED) {
				ins=is->getInstanceInit(defaultClass);
			} else {
				ins=is->getInstanceInit(ai.getNumClasses());
			}
		}

		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				if(!rnd<0.75) setGene(i,j,0,1);
				else setGene(i,j,0,0);

				if(!rnd<0.75) setGene(i,j,1,1);
				else setGene(i,j,1,0);

				double min,max;
				double size=(!rnd*0.25+0.50)*ai.getSizeDomain(j);
				if(ins) {
					double minD=ai.getMinDomain(j);
					double maxD=ai.getMaxDomain(j);
					double val=ins->realValueOfAttribute(j);
					min=val-size/2.0;
					max=val+size/2.0;
					if(min<minD) {
						min=minD;
					}
					if(max>maxD) {
						max=maxD;
					}
				} else {
					min=!rnd * (ai.getSizeDomain(j) - size)
						+ ai.getMinDomain(j);
					max=min+size;
				}
				
				if(!rnd<0.5) {
					setGene(i,j,2,min);
					setGene(i,j,3,max);
				} else {
					setGene(i,j,3,min);
					setGene(i,j,2,max);
				}
			} else {
				int value;
				if(ins) value=ins->valueOfAttribute(j);
				else value=-1;
				for(k=0;k<tReal->attributeSize[j];k++) {
					if(k!=value) {
						if(!rnd<tGlobals->probOne) {
							setGene(i,j,k,1);
						} else {
							setGene(i,j,k,0);
						}
					} else {
						setGene(i,j,k,1);
					}
				}
			}
		}
		int cl;
		do {
			cl=rnd(0,ai.getNumClasses()-1);
		} while(tGlobals->defaultClassPolicy!=DISABLED && cl==defaultClass);
		setGene(i, tGlobals->numAttributesMC, 0, cl);
	}
}

void classifier_ubr::setGene(short int rule, short int gene,
			      short int value, double nfo)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules) ||
	    (value < 0) || (value >= tReal->attributeSize[gene])) {
		fprintf(stderr,"Invalid access at setGene %d %d %d\n",
		       rule, gene, value);
		exit(1);
	}
#endif
	chromosome[rule * tReal->ruleSize +
		   tReal->attributeOffset[gene] + value] = nfo;
	modif = 1;
}

double classifier_ubr::getGene(short int rule,
				short int gene, short int value)
{
#ifdef DEBUG
	if ((gene < 0) || (gene >= tGlobals->numAttributes) ||
	    (rule < 0) || (rule >= numRules) ||
	    (value < 0) || (value >= tReal->attributeSize[gene])) {
		fprintf(stderr,"Invalid access at getGene %d %d %d", rule,
		       gene, value);
		exit(1);
	}
#endif
	return chromosome[rule * tReal->ruleSize +
			  tReal->attributeOffset[gene] + value];
}

void classifier_ubr::crossover(classifier * in,
				classifier * out1, classifier * out2)
{
	switch (tGlobals->cxOperator) {
	case CROSS_2P:
		crossover_2px(this, (classifier_ubr *) in,
			      (classifier_ubr *) out1,
			      (classifier_ubr *) out2);
		break;
	case CROSS_1P:
	default:

		crossover_1px(this, (classifier_ubr *) in,
			      (classifier_ubr *) out1,
			      (classifier_ubr *) out2);
	}
}

void classifier_ubr::searchPosition(int pos, int &attribute, int &value)
{
	int i;

	for (i = 0;
	     i < tGlobals->numAttributes && pos >= tReal->attributeOffset[i]; i++);
	attribute = i - 1;
	value = pos - tReal->attributeOffset[attribute];
}

double classifier_ubr::mutationOffset(double geneValue, double offsetMin,
				       double offsetMax)
{
	double newValue;
	if (!rnd < 0.5) {
		newValue = geneValue + !rnd * offsetMax;
	} else {
		newValue = geneValue - !rnd * offsetMin;
	}
	return newValue;
}

void classifier_ubr::mutation()
{
	int i;
	int rule, attribute, value;

	modif = 1;
	tMut->mutationCounter++;

	rule = rnd(0,numRules-1);
	if(tGlobals->numClasses>1 && !rnd<0.10) {
		attribute = tGlobals->numAttributesMC;
		value=0;
	} else {
		attribute=rnd(0,tGlobals->numAttributesMC-1);
		value=rnd(0,tReal->attributeSize[attribute]-1);
	}

	if (attribute != tGlobals->numAttributesMC) {
		double oldValue=getGene(rule,attribute,value);
		if (ai.getTypeOfAttribute(attribute) == REAL) {
			double newValue;
			if(value<2) {
				if(oldValue==0) newValue=1;
				else newValue=0;
			} else {
				double minOffset, maxOffset;
				minOffset = maxOffset =
				    0.7 * ai.getSizeDomain(attribute);
				if (oldValue - minOffset <
					    ai.getMinDomain(attribute))
					minOffset = oldValue -
					    ai.getMinDomain(attribute);
				if (oldValue + maxOffset >
					    ai.getMaxDomain(attribute))
					maxOffset = ai.getMaxDomain(attribute) -
					    oldValue;
				newValue = mutationOffset(oldValue, minOffset,
							  maxOffset);
			}
			setGene(rule, attribute, value, newValue);
		} else {
			if(oldValue==1) setGene(rule,attribute,value,0);
			else setGene(rule,attribute,value,1);
		}
	} else {
		if(tGlobals->numClasses>1) {
			int newValue;
			int oldValue = (int) getGene(rule, attribute, value);
			do {
				newValue = rnd(0, ai.getNumClasses()-1);
			} while (newValue == oldValue || tGlobals->defaultClassPolicy!=DISABLED && newValue==defaultClass);
			setGene(rule, attribute, value, newValue);
		}
	}
}

static void swap(int &a, int &b)
{
	int tmp;

	tmp = a;
	a = b;
	b = tmp;
}

void classifier_ubr::specialCrossover(double *parent1, double *parent2,
				       double *son1, double *son2)
{
	int i;

	for (i = 0; i < tGlobals->numAttributesMC; i++) {
		if (ai.getTypeOfAttribute(i) == REAL) {
			tReal->specialCrossover(parent1[2], parent2[2],
						 son1[2], son2[2],
						 ai.getMinDomain(i),
						 ai.getMaxDomain(i));
			tReal->specialCrossover(parent1[3], parent2[3],
						 son1[3], son2[3], 0,
						 ai.getSizeDomain(i));
		}
		parent1 += tReal->attributeSize[i];
		parent2 += tReal->attributeSize[i];
		son1 += tReal->attributeSize[i];
		son2 += tReal->attributeSize[i];
	}
}

void classifier_ubr::crossover_2px(classifier_ubr * in1,
				    classifier_ubr * in2,
				    classifier_ubr * out1,
				    classifier_ubr * out2)
{
	int i1rule1, i1rule2, i2rule1, i2rule2;
	int cutPoint1, cutPoint2, sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	i1rule1 = rnd(0, in1->numRules - 1);
	i1rule2 = rnd(0, in1->numRules - 1);
	if (i1rule1 > i1rule2)
		swap(i1rule1, i1rule2);

	i2rule1 = rnd(0, in2->numRules - 1);
	i2rule2 = rnd(0, in2->numRules - 1);
	if (i2rule1 > i2rule2)
		swap(i2rule1, i2rule2);

	sizeOfSon1 = i1rule1 + i2rule2 - i2rule1 + in1->numRules - i1rule2;
	sizeOfSon2 = i2rule1 + i1rule2 - i1rule1 + in2->numRules - i2rule2;

	cutPoint1 = rnd(0, tReal->ruleSize - 1);
	cutPoint2 = rnd(0, tReal->ruleSize - 1);

	if (i1rule1 == i1rule2 || i2rule1 == i2rule2) {
		if (cutPoint1 > cutPoint2)
			swap(cutPoint1, cutPoint2);
	}

	out1->length = sizeOfSon1 * tReal->ruleSize;
	out2->length = sizeOfSon2 * tReal->ruleSize;

	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;

	out1->chromosome = new double[tReal->ruleSize * out1->numRules];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new double[tReal->ruleSize * out2->numRules];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}

	int cutPoint = cutPoint1;

	bcopy(chromosome, out1->chromosome,
	      i1rule1 * tReal->ruleSize * sizeof(double));
	bcopy(in2->chromosome, out2->chromosome,
	      i2rule1 * tReal->ruleSize * sizeof(double));

	bcopy(&in1->chromosome[i1rule1 * tReal->ruleSize],
	      &out1->chromosome[i1rule1 * tReal->ruleSize],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[i2rule1 * tReal->ruleSize],
	      &out2->chromosome[i2rule1 * tReal->ruleSize],
	      cutPoint * sizeof(double));

	bcopy(&in2->chromosome[i2rule1 * tReal->ruleSize + cutPoint],
	      &out1->chromosome[i1rule1 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));
	bcopy(&in1->chromosome[i1rule1 * tReal->ruleSize + cutPoint],
	      &out2->chromosome[i2rule1 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->
				 chromosome[i1rule1 * tReal->ruleSize],
				 &in2->chromosome[i2rule1 *
						  tReal->ruleSize],
				 &out1->chromosome[i1rule1 *
						   tReal->ruleSize],
				 &out2->chromosome[i2rule1 *
						   tReal->ruleSize]);

	if (i2rule1 != i2rule2)
		bcopy(&in2->chromosome[(i2rule1 + 1) * tReal->ruleSize],
		      &out1->chromosome[(i1rule1 + 1) * tReal->ruleSize],
		      (i2rule2 -
		       (i2rule1 + 1)) * tReal->ruleSize * sizeof(double));
	if (i1rule1 != i1rule2)
		bcopy(&in1->chromosome[(i1rule1 + 1) * tReal->ruleSize],
		      &out2->chromosome[(i2rule1 + 1) * tReal->ruleSize],
		      (i1rule2 -
		       (i1rule1 + 1)) * tReal->ruleSize * sizeof(double));


	cutPoint = cutPoint2;
	if (i2rule2 != i2rule1)
		bcopy(&in2->chromosome[i2rule2 * tReal->ruleSize],
		      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
					tReal->ruleSize],
		      cutPoint * sizeof(double));
	if (i1rule2 != i1rule1)
		bcopy(&in1->chromosome[i1rule2 * tReal->ruleSize],
		      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
					tReal->ruleSize],
		      cutPoint * sizeof(double));

	bcopy(&in1->chromosome[i1rule2 * tReal->ruleSize + cutPoint],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1) *
				tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));
	bcopy(&in2->chromosome[i2rule2 * tReal->ruleSize + cutPoint],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1) *
				tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->
				 chromosome[i1rule2 * tReal->ruleSize],
				 &in2->chromosome[i2rule2 *
						  tReal->ruleSize],
				 &out1->
				 chromosome[(i1rule1 + i2rule2 -
					     i2rule1) *
					    tReal->ruleSize],
				 &out2->
				 chromosome[(i2rule1 + i1rule2 -
					     i1rule1) * tReal->ruleSize]);

	bcopy(&in1->chromosome[(i1rule2 + 1) * tReal->ruleSize],
	      &out1->chromosome[(i1rule1 + i2rule2 - i2rule1 + 1) *
				tReal->ruleSize],
	      (in1->numRules -
	       (i1rule2 + 1)) * tReal->ruleSize * sizeof(double));
	bcopy(&in2->chromosome[(i2rule2 + 1) * tReal->ruleSize],
	      &out2->chromosome[(i2rule1 + i1rule2 - i1rule1 + 1) *
				tReal->ruleSize],
	      (in2->numRules -
	       (i2rule2 + 1)) * tReal->ruleSize * sizeof(double));
}

void classifier_ubr::dumpPhenotype(char *string)
{
	double *ptr = chromosome;
	char temp[10000];
	char tmp1[20];
	char tmp2[20];
	double averageWidth=0;
	int numIntervals=0;
	int i, j, k;

	strcpy(string, "");
	for (i = 0; i < numRules; i++) {
		sprintf(temp, "%d:", i);
		strcat(string, temp);
		double *ptr2 = ptr;
		for (j = 0; j < tGlobals->numAttributesMC; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				double minD = ai.getMinDomain(j);
				double maxD = ai.getMaxDomain(j);
				double min = ptr2[2];
				double max = ptr2[3];
				double size = ai.getSizeDomain(j);

				if(min>max) swap(min,max);
				if (ptr2[0] || min <= minD) {
					min=minD;
				}
				sprintf(tmp1, "%f", min);
				if (ptr2[1] || max >= maxD) {
					max=maxD;
				}
				sprintf(tmp2, "%f", max);
				averageWidth+=(max-min)/size;
				numIntervals++;
				sprintf(temp, "[%s,%s]", tmp1, tmp2);
			} else {
				for(k=0;k<tReal->attributeSize[j];k++) {
					temp[k]=(int)ptr2[k]+'0';
				}
				temp[k]=0;
			}
			strcat(string, temp);
			strcat(string, "|");
			ptr2 += tReal->attributeSize[j];
		}
		int cl=(int) (*ptr2);
		sprintf(temp, "%d\n", cl);
		strcat(string, temp);
		ptr += tReal->ruleSize;
	}
	if(tGlobals->defaultClassPolicy!=DISABLED) {
		sprintf(temp, "%d:Default rule -> %d\n", i,defaultClass);
		strcat(string, temp);
	}
	strcat(string, "\n");
	sprintf(temp,"Average interval size %f\n"
		,averageWidth/(double)numIntervals);
	strcat(string,temp);
}

void classifier_ubr::crossover_1px(classifier_ubr * in1,
				    classifier_ubr * in2,
				    classifier_ubr * out1,
				    classifier_ubr * out2)
{
	unsigned int rulep1, rulep2;
	int sizeOfSon1, sizeOfSon2;

	out1->modif = out2->modif = 1;

	rulep1 = rnd(0, in1->numRules - 1);
	rulep2 = rnd(0, in2->numRules - 1);
	sizeOfSon1 = rulep1 + in2->numRules - rulep2;
	sizeOfSon2 = rulep2 + in1->numRules - rulep1;

	int cutPoint = rnd(0, tReal->ruleSize - 1);

	out1->length = sizeOfSon1 * tReal->ruleSize;
	out2->length = sizeOfSon2 * tReal->ruleSize;
	out1->numRules = sizeOfSon1;
	out2->numRules = sizeOfSon2;
	out1->chromosome = new double[tReal->ruleSize * out1->numRules];
	if (out1->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	out2->chromosome = new double[tReal->ruleSize * out2->numRules];
	if (out2->chromosome == NULL) {
		perror("out of memory");
		exit(1);
	}
	bcopy(in1->chromosome, out1->chromosome,
	      rulep1 * tReal->ruleSize * sizeof(double));

	bcopy(in2->chromosome, out2->chromosome,
	      rulep2 * tReal->ruleSize * sizeof(double));

	bcopy(&in1->chromosome[rulep1 * tReal->ruleSize],
	      &out1->chromosome[rulep1 * tReal->ruleSize],
	      cutPoint * sizeof(double));
	bcopy(&in2->chromosome[rulep2 * tReal->ruleSize],
	      &out2->chromosome[rulep2 * tReal->ruleSize],
	      cutPoint * sizeof(double));

	bcopy(&in1->chromosome[rulep1 * tReal->ruleSize + cutPoint],
	      &out2->chromosome[rulep2 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));
	bcopy(&in2->chromosome[rulep2 * tReal->ruleSize + cutPoint],
	      &out1->chromosome[rulep1 * tReal->ruleSize + cutPoint],
	      (tReal->ruleSize - cutPoint) * sizeof(double));

	if (tReal->thereIsSpecialCrossover)
		specialCrossover(&in1->
				 chromosome[rulep1 * tReal->ruleSize],
				 &in2->chromosome[rulep2 *
						  tReal->ruleSize],
				 &out1->chromosome[rulep1 *
						   tReal->ruleSize],
				 &out2->chromosome[rulep2 *
						   tReal->ruleSize]);

	bcopy(&in2->chromosome[(rulep2 + 1) * tReal->ruleSize],
	      &out1->chromosome[(rulep1 + 1) * tReal->ruleSize],
	      (in2->numRules -
	       (rulep2 + 1)) * tReal->ruleSize * sizeof(double));

	bcopy(&in1->chromosome[(rulep1 + 1) * tReal->ruleSize],
	      &out2->chromosome[(rulep2 + 1) * tReal->ruleSize],
	      (in1->numRules -
	       (rulep1 + 1)) * tReal->ruleSize * sizeof(double));

}

void classifier_ubr::pruneClassifiers(int *classifiersToPrune,
				       int numPruned)
{
	double *newChromosome;
	int i;

	if (numRules == 1 || numPruned == 0) return;

	if(classifiersToPrune[numPruned-1]==numRules) numPruned--;

	newChromosome =
	    new double[(numRules - numPruned) * tReal->ruleSize];
	if (newChromosome == NULL) {
		fprintf(stderr,"Out of memory\n");
		exit(1);
	}

	int countPruned = 0;
	double *ptrOrig = chromosome;
	double *ptrDest = newChromosome;
	for (i = 0; i < numRules; i++) {
		if (countPruned < numPruned) {
			if (i != classifiersToPrune[countPruned]) {
				bcopy(ptrOrig, ptrDest,
				      tReal->ruleSize * sizeof(double));
				ptrDest += tReal->ruleSize;
			} else {
				countPruned++;
			}
		} else {
			bcopy(ptrOrig, ptrDest,
			      tReal->ruleSize * sizeof(double));
			ptrDest += tReal->ruleSize;
		}
		ptrOrig += tReal->ruleSize;
	}

	delete chromosome;
	chromosome = newChromosome;

	numRules -= numPruned;
	length = numRules * tReal->ruleSize;
}

int classifier_ubr::classify(instance * ins)
{
	if (tGlobals->ignoreMissingValues && ins->hasMissingValues()) {
		return classifyIgnoreMissing(ins);
	}
	return classifyNoIgnoreMissing(ins);
}

int classifier_ubr::classifyIgnoreMissing(instance * ins)
{
	double *ptr = chromosome;
	int i, j, match;
	int numAtributs = tGlobals->numAttributesMC;

	for (i = 0; i < numRules; i++) {
		match = 1;
		double *cAtr = ptr;
		for (j = 0; j < numAtributs && match; j++) {
			if (!ins->isMissing(j)) {
				if (ai.getTypeOfAttribute(j) == REAL) {
					double valueAtr =
					    ins->realValueOfAttribute(j);
					double min = cAtr[2];
					double max = cAtr[3];
					if(min>max) swap(min,max);
					if ((!cAtr[0] && valueAtr < min)
					    || (!cAtr[1] && valueAtr > max))
						match = 0;
				} else {
					int valueAtr
						= ins->valueOfAttribute(j);
					if (cAtr[valueAtr] == 0) {
						match = 0;
					}
				}
			}
			cAtr += tReal->attributeSize[j];
		}
		if (match) return i;
		ptr += tReal->ruleSize;
	}

	if(tGlobals->defaultClassPolicy==DISABLED) return -1;
	return numRules;
}


int classifier_ubr::classifyNoIgnoreMissing(instance * ins)
{
	double *ptr = chromosome;
	int i, j, match;
	int numAtributs = tGlobals->numAttributesMC;

	for (i = 0; i < numRules; i++) {
		match = 1;
		double *cAtr = ptr;
		for (j = 0; j < numAtributs && match; j++) {
			if (ai.getTypeOfAttribute(j) == REAL) {
				double valueAtr = ins->realValueOfAttribute(j);
				double min = cAtr[2];
				double max = cAtr[3];
				if(min>max) swap(min,max);
				if ((!cAtr[0] && valueAtr < min)
				    || (!cAtr[1] && valueAtr > max))
					match = 0;
			} else {
				int valueAtr = ins->valueOfAttribute(j);
				if (cAtr[valueAtr] == 0) {
					match = 0;
				}
			}
			cAtr += tReal->attributeSize[j];
		}
		if (match) return i;
		ptr += tReal->ruleSize;
	}

	if(tGlobals->defaultClassPolicy==DISABLED) return -1;
	return numRules;
}

int classifier_ubr::ruleMatches(double *rule,instance *ins)
{
	int j, match;
	int numAtributs = tGlobals->numAttributesMC;

	match = 1;
	double *cAtr = rule;
	for (j = 0; j < tGlobals->numAttributesMC && match; j++) {
		if (ai.getTypeOfAttribute(j) == REAL) {
			double valueAtr = ins->realValueOfAttribute(j);
			double min = cAtr[2];
			double max = cAtr[3];
			if(min>max) swap(min,max);
			if ((!cAtr[0] && valueAtr < min)
			    || (!cAtr[1] && valueAtr > max))
				return 0;
		} else {
			int valueAtr = ins->valueOfAttribute(j);
			if (cAtr[valueAtr] == 0) {
				return 0;
			}
		}
		cAtr += tReal->attributeSize[j];
	}

	return 1;
}


void classifier_ubr::crossover_smart(classifier **parents, int num)
{
        int i, j, k;

        modif = 1;
        if(chromosome) delete chromosome;

        JVector<double *>ptrs;
        JVector<matchProfileAgent *>profiles;
        int ruleCount=0;

        for(k=0;k<num;k++) {
                classifier_ubr *parent=(classifier_ubr *)parents[k];
                double *ptr=parent->chromosome;
                for(i=0;i<parent->numRules;ptr+=tReal->ruleSize,i++) {
                        double *newRule= new double[tReal->ruleSize];
                        bcopy(ptr,newRule,sizeof(double)*tReal->ruleSize);

                        matchProfileAgent *mpa = evaluateRuleCX(newRule);
                        ptrs.addElement(newRule);
                        profiles.addElement(mpa);
                }
        }

        JVector<int> ruleOrder;
        findOrder(profiles,ptrs,ruleOrder);

        numRules = ruleOrder.size();
        chromosome = new double[numRules * tReal->ruleSize];
	length = tReal->ruleSize * numRules;
        for (i = 0; i < numRules; i++) {
                bcopy(ptrs[ruleOrder[i]], &chromosome[i * tReal->ruleSize],
                      sizeof(double) * tReal->ruleSize);
        }

        for(i=0;i<profiles.size();i++) {
                delete profiles[i];
        }
        for(i=0;i<ptrs.size();i++) {
                delete ptrs[i];
        }
}

void classifier_ubr::findOrder(JVector<matchProfileAgent *>&profiles
        ,JVector<double *>&ptrs,JVector<int> &ruleOrder)
{
        unsigned long long numInst=is->getNumInstancesOfIteration();
        int i,j;

        ruleOrderAgent *best=NULL;
        for(j=0;j<tCross->repetitionsRuleOrdering;j++) {
        	int totalRules=profiles.size();
                Sampling samp(totalRules);
                ruleOrderAgent *roa = new ruleOrderAgent(defaultClass,numInst,totalRules);

                for(i=0;i<totalRules;i++) {
                        int rule=samp.getSample();
                        roa->insertRule(rule,profiles[rule]);
                }
                roa->refineRuleSet();

                if(best==NULL) {
                        best=roa;
                } else {
                        double accOld=best->getAccuracy();
                        double accNew=roa->getAccuracy();
                        if(accNew>accOld) {
                                delete best;
                                best=roa;
                        } else if(accNew==accOld
                                && roa->getNumRules()<best->getNumRules()) {
                                delete best;
                                best=roa;
                        } else {
                                delete roa;
                        }
                }
        }

        best->copyOrder(ruleOrder);
        if(ruleOrder.size()==0) {
                int pos=rnd(0,profiles.size()-1);
                ruleOrder.addElement(pos);
        }
        delete best;
}

matchProfileAgent *classifier_ubr::evaluateRuleCX(double *rule)
{
        unsigned long long i;
        int cl = (int)rule[tReal->ruleSize-1];
        unsigned long long max = is->getNumInstancesOfIteration();
	instance **instances=is->getInstancesOfIteration();

        matchProfileAgent *mpa=new matchProfileAgent(max,cl);

        for (i = 0; i < max; i++) {
                if (ruleMatches(rule, instances[i])) {
                        if (instances[i]->getClass() == cl) {
                                mpa->addOK(i);
                        } else {
                                mpa->addKO(i);
                        }
                }
        }
        mpa->generateProfiles();

        return mpa;
}

