#ifndef _CLASSIFIER_UBR_
#define _CLASSIFIER_UBR_

#include "classifier.h"
#include "timerRealKR.h"
#include "timerGlobals.h"
#include "agentPerformanceTraining.h"
#include "matchProfileAgent.h"

extern timerRealKR *tReal;
extern timerGlobals *tGlobals;

class classifier_ubr: public classifier   {
	double *chromosome;
	int numRules;

	/* Operadors de crossover i mutation */
	void crossover_1px(classifier_ubr *in1,classifier_ubr *in2
		,classifier_ubr *out1,classifier_ubr *out2);
	void crossover_2px(classifier_ubr *in1,classifier_ubr *in2, classifier_ubr *out1
		,classifier_ubr *out2);
	double getGene(short int rule,short int attr,short int value); 
	void setGene(short int rule,short int attr,short int value
		,double nfo);

	void specialCrossover(double *parent1,double *parent2,double *son1,double *son2);

	double mutationOffset(double geneValue,double offsetMin,double offsetMax);

	void searchPosition(int pos,int &attribute,int &value);

	void initializeChromosome(void);
	int classifyIgnoreMissing(instance *i);
	int classifyNoIgnoreMissing(instance *i);

	void findOrder(JVector<matchProfileAgent *>&profiles,JVector<double *>&ptrs,JVector<int> &ruleOrder);
	matchProfileAgent *evaluateRuleCX(double *rule);
	int ruleMatches(double *rule,instance *ins);

public:
	classifier_ubr(int useMPI=0,int originNode=0);
	classifier_ubr(const classifier_ubr &orig,int son=0);
	~classifier_ubr();

	inline int getClass(int rule) {
		if(rule==numRules) return defaultClass;
        	return (int) chromosome[rule * tReal->ruleSize +
			tReal->ruleSize - 1];
	}

	inline double getHierarchicalSelectionLength()
	{
		if(tHierar->useMDL) return getTheoryLength();
		return aliveClassifiers; //numRules+1;
	}

        inline int getNiche() {
                if(tGlobals->defaultClassPolicy!=AUTO) return 0;
                return getClass(getNumClassifiers()-1);
        }
        inline int getNumNiches() {
                if(tGlobals->defaultClassPolicy!=AUTO) return 1;
                return ai.getNumClasses();
        }

	int getNumClassifiers(){
		int nC=numRules;
		if(tGlobals->defaultClassPolicy!=DISABLED) nC++;
		return nC;
	}

	void pruneClassifiers(int *classifiersToPrune,int numPruned);
	int classify(instance *i);
	double computateTheoryLength(agentPerformanceTraining *ap);

	void crossover(classifier *,classifier *,classifier *);
	void mutation();
	
	void dumpPhenotype(char *string);

	inline int hasSmartCrossover() {return 1;}
        void crossover_smart(classifier **parents, int num);

	
	inline int numSpecialStages(){return 0;}
	void doSpecialStage(int stage){}
};

#endif

