#include "timerPruning.h"
#include "timerCrossover.h"
#include "timerGlobals.h"
#include "sampling.h"

extern timerCrossover *tCross;
extern timerPruning *tPruning;
extern timerGlobals *tGlobals;

void geneticAlgorithm::crossTwoParents(int parent1, int parent2, int son1,
				       int son2)
{
	offspringPopulation[son1] =
	    cf->cloneClassifier(population[parent1], 1);
	offspringPopulation[son2] =
	    cf->cloneClassifier(population[parent2], 1);

	population[parent1]->crossover(population[parent2]
				       , offspringPopulation[son1]
				       , offspringPopulation[son2]);

	if (tCross->checkCrossover) {
		tPruning->disable();
		population[parent1]->fitnessComputation();
		population[parent2]->fitnessComputation();
		offspringPopulation[son1]->fitnessComputation();
		offspringPopulation[son2]->fitnessComputation();
		offspringPopulation[son1]->activateModified();
		offspringPopulation[son2]->activateModified();
		tPruning->enable();

		double p1 = population[parent1]->getAccuracy();
		double p2 = population[parent2]->getAccuracy();
		double s1 = offspringPopulation[son1]->getAccuracy();
		double s2 = offspringPopulation[son2]->getAccuracy();
		/*if(s1>p1 || s1>p2 || s2>p1 || s2>p2) {
		   tCross->offspringWin++;
		   } else {
		   tCross->parentsWin++;
		   } */
		double aveAccP = (p1 + p2) / 2.0;
		double aveAccO = (s1 + s2) / 2.0;
		if (aveAccP >= aveAccO) {
			tCross->parentsWin++;
		} else {
			tCross->offspringWin++;
		}
	}
}

void geneticAlgorithm::crossSmart(int *parentsPos, int numParents, int son)
{
	classifier *parents[numParents];
	int i;
	for (i = 0; i < numParents; i++) {
		parents[i] = population[parentsPos[i]];
	}

	offspringPopulation[son] =
	    cf->cloneClassifier(population[parentsPos[0]], 1);
	offspringPopulation[son]->crossover_smart(parents, numParents);
}


void geneticAlgorithm::crossOneParent(int parent, int son)
{
	offspringPopulation[son] =
	    cf->cloneClassifier(population[parent], 0);
}

void geneticAlgorithm::crossover()
{
	int i, j, k, countCross = 0;
	int numNiches = population[0]->getNumNiches();
	int hasSmartX = population[0]->hasSmartCrossover() 
		&& !lastIteration && (tCross->smartCrossoverProb>0);

	JVector < int >parents[numNiches];

	for (i = 0; i < numNiches; i++) {
		parents[i].ensureCapacity(popSize);
	}

	for (i = 0; i < popSize; i++) {
		int niche = population[i]->getNiche();
		parents[niche].addElement(i);
	}

	for (i = 0; i < numNiches; i++) {
		int size = parents[i].size();
		Sampling samp(size);
		int p1 = -1;
		for (j = 0; j < size; j++) {
			if (!rnd < tCross->crossoverProb) {
				if (hasSmartX
				    && !rnd < tCross->smartCrossoverProb) {
					int num = tCross->numParentsSmartX;
					int posParents[num];
					for (k = 0; k < num; k++) {
						posParents[k] = parents[i][samp.getSample()];
					}
					crossSmart(posParents, num,
						   countCross++);
				} else {
					if (p1 == -1) {
						p1 = samp.getSample();
					} else {
						int p2 = samp.getSample();
						crossTwoParents(parents[i][p1],
								parents[i][p2],
								countCross,
								countCross
								+ 1);
						countCross += 2;
						p1 = -1;
					}
				}
			} else {
				crossOneParent(parents[i][samp.getSample()],
					       countCross++);
			}
		}
		if (p1 != -1) {
			crossOneParent(parents[i][p1], countCross++);
		}
	}
}
