#include <stdio.h>
#include <string.h>
#include "discretizationManager.h"

discretizationManager::discretizationManager() 
{
	numAttributes=-1;
	ready=0;
}

void discretizationManager::setNumAttributes(int num)
{
	int i;

	numAttributes=num;
	numD=discNames.size();
	discretizers = new discretizer *[numD];
	for(i=0;i<numD;i++) {
		createDiscretizer(i,discNames[i]);
	}
}

void discretizationManager::insertDiscretizer(char *name)
{
	char *copy=new char[strlen(name)+1];
	strcpy(copy,name);
	discNames.addElement(copy);
}

void discretizationManager::createDiscretizer(int num,char *name) 
{
	int i;

	if(numAttributes==-1) {
		fprintf(stderr,"Number of attributes undefined at"
			" insertDiscretizer:%s\n", name);
		exit(1);
	}

	if(!strncasecmp(name,"uniform",7)) {
		int numBins;

		if(sscanf(name+8,"%d",&numBins)!=1) {
			fprintf(stderr,"Parse error at insertDiscretizer:%s\n",
				name);
			exit(1);
		}

		discretizers[num]=new uniformDiscretizer(numBins,numAttributes);
	} else if(!strncasecmp(name,"equalfreq",9)) {
		int numBins;

		if(sscanf(name+10,"%d",&numBins)!=1)
		{
			fprintf(stderr,"Parse error at insertDiscretizer:%s\n",name);
			exit(1);
		}
		discretizers[num] = new equalfreqDiscretizer(numBins,numAttributes);
	} else if(!strncasecmp(name,"random",6)) {
		discretizers[num] = new randDiscretizer(numAttributes);
	} else if(!strncasecmp(name,"chimerge",8)) {
		float siglevel;
		if(sscanf(name+9,"%f",&siglevel)!=1)
		{
			fprintf(stderr,"Parse error at insertDiscretizer:%s\n",name);
			exit(1);
		}
		discretizers[num] = new chimergeDiscretizer((double)siglevel,numAttributes);
	} else if(!strncasecmp(name,"mdlp",4)) {
		discretizers[num] = new mdlpDiscretizer(numAttributes);
	} else if(!strncasecmp(name,"mantaras",8)) {
		discretizers[num] = new mantarasDiscretizer(numAttributes);
	} else if(!strncasecmp(name,"id3",3)) {
		discretizers[num] = new id3Discretizer(numAttributes);
	} else if(!strncasecmp(name,"fayyad",3)) {
		discretizers[num] = new fayyadDiscretizer(numAttributes);
	} else if(!strncasecmp(name,"usd",3)) {
		discretizers[num] = new usdDiscretizer(numAttributes);
	} else {
		fprintf(stderr,"Unknown discretizer %s\n",name);
		exit(1);
	}
}

void discretizationManager::setBounds(double *minDomains,double *maxDomains)
{
	int i;

	for(i=0;i<numD;i++) {
		discretizers[i]->setBounds(minDomains,maxDomains);
	}

	ready=1;
}

void discretizationManager::updateDiscretizers(char *dataFile)
{
	int i;

	for(i=0;i<numD;i++) {
		discretizers[i]->updateDiscretizer(dataFile);
	}
}

