/************************************************************************
 * MDLP Discretizer					*
 *						 Pere Miralles Ribas	*
 ************************************************************************/

#include "equalfreqDiscretizer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "messageBuffer.h"
#include "random.h"

extern Random rnd;

extern messageBuffer mb;

void equalfreqDiscretizer::results()
{
	int i;

	double sum=0;
	double count=0;
	for(i=0;i<numAttributes;i++) {
		if(realAttributes[i]) {
			sum+=numCutPoints[i];
			count++;
                        //mb.printf("Atribut: %d Numero de punts de tall: %d\n",i,numCutPoints[i]);
                        //for(int j=0;j<numCutPoints[i];j++)
                        //        mb.printf("[%d] : %f\n",j,(float)cutPoints[i][j]);

		}
	}
	mb.printf("Numero de punts de tall Discretitzacio : %g\n",sum/count);
}

void equalfreqDiscretizer::readFile(char *datafile)
{
	char buffer[10000];
	int i,j;

	realAttributes=new bool[numAttributes];
	cutPoints = new double *[numAttributes];
	numCutPoints = new int[numAttributes];
	int attributeCount=0;
	for(i=0;i<numAttributes;i++) {
		cutPoints[i]=NULL;
	}
	FILE *fp=fopen(datafile,"r");

	fgets(buffer,9999,fp);
	while(!feof(fp) && strncasecmp(buffer,"@data",5)) {
		if(!strncasecmp(buffer,"@attribute ",10)) {
			char *ptr=buffer+10;
			if(attributeCount==numAttributes) {
				while(*ptr==' ' || *ptr=='\t') ptr++;
				while(*ptr!=' ' && *ptr!='\t') ptr++;
				while(*ptr==' ' || *ptr=='\t') ptr++;
				if(*ptr!='{') {
					fprintf(stderr,"Wrong format !!\n");
					exit(1);
				}
				ptr++;
				numClasses=0;
				for(i=0;i<strlen(ptr);i++) {
					if(ptr[i]==',' || ptr[i]=='}') numClasses++;
				}
				classes=new char *[numClasses];
				for(i=0;i<numClasses;i++) {
					char tmp[1000];
					int index=0;
					while(*ptr!=',' && *ptr!='}') {
						tmp[index++]=*ptr++;
					}
					tmp[index]=0;
					classes[i]=new char[strlen(tmp)+1];
					strcpy(classes[i],tmp);
					ptr++;
				}
			} else {
				while(*ptr==' ' || *ptr=='\t') ptr++;
				while(*ptr!=' ' && *ptr!='\t') ptr++;
				while(*ptr==' ' || *ptr=='\t') ptr++;
				if(!strncasecmp(ptr,"real",4) || !strncasecmp(ptr,"integer",7)) {
					realAttributes[attributeCount]=true;
				} else {
					realAttributes[attributeCount]=false;
				}
			}
			attributeCount++;
		}
		fgets(buffer,9999,fp);
	}

	int pos=ftell(fp);
	numInstances=0;
	fgets(buffer,9999,fp);
	while(!feof(fp)) {
		numInstances++;
		fgets(buffer,9999,fp);
	}
	fseek(fp,pos,SEEK_SET);

	missing = new bool *[numInstances];
	realValues = new double *[numInstances];
	classOfInstances = new int[numInstances];

	fgets(buffer,9999,fp);
	int instanceCount=0;
	while(!feof(fp)) {
		buffer[strlen(buffer)-1]=0;
		if(buffer[0] && buffer[0]!='%') {		
			missing[instanceCount] = new bool[numAttributes];
			realValues[instanceCount] = new double[numAttributes];
			char *ptr=buffer;
			for(j=0;j<numAttributes;j++) {
				char tmp[1000];
				int index=0;
				while(*ptr!=',') {
					tmp[index++]=*ptr++;
				}
				tmp[index]=0;
				if(realAttributes[j]) {
					if(tmp[0]=='?') {
						missing[instanceCount][j]=true;
					} else {
						missing[instanceCount][j]=false;
						float tempFloat;
						if(sscanf(tmp,"%f",&tempFloat)!=1) {
							fprintf(stderr,"Wrong format 2 %s!!!\n",ptr);
							exit(1);
						}	
						realValues[instanceCount][j]=tempFloat;
					}
				}
				ptr++;
			}
			bool found=false;
			for(j=0;j<numClasses && !found;j++) {
				if(!strcasecmp(classes[j],ptr)) {
					classOfInstances[instanceCount]=j;
					found=true;
				}
			}
			if(!found) {
				fprintf(stderr,"Wrong format 3 !!!\n");
				exit(1);
			}
			instanceCount++;
		}
		fgets(buffer,9999,fp);
	}

	fclose(fp);
}
	
	
void equalfreqDiscretizer::doDiscretize() 
{
	int i,j;
	int points[numInstances];
	double cp[numInstances];

	for(i=0;i<numAttributes;i++) {
		if(realAttributes[i]) {
			int numPoints=0;
			for(j=0;j<numInstances;j++) {
				if(!missing[j][i]) {
					points[numPoints++]=j;
				}
			}

			if(numPoints) {
				sortValues(i,points,0,numPoints-1);
	
				int numCP=0;
				discretizeAttribute(i,points,0,numPoints-1,cp,numCP);
				numCutPoints[i]=numCP;
				if(numCP>0) {
					cutPoints[i]=new double[numCP];
					for(j=0;j<numCP;j++) {
						cutPoints[i][j]=cp[j];
					}
				}
			} else {
				numCutPoints[i]=0;
			}
		} else {
			numCutPoints[i]=0;
		}
	}
}

void equalfreqDiscretizer::sortValues(int attribute,int *values,int begin,int end) 
{
        double pivot;
        int temp;
        int i,j;

        i=begin;j=end;
        pivot=realValues[values[(i+j)/2]][attribute];
        do {
                while(realValues[values[i]][attribute]<pivot) i++;
                while(realValues[values[j]][attribute]>pivot) j--;
                if(i<=j) {
                        if(i<j) {
                                temp=values[i];
                                values[i]=values[j];
                                values[j]=temp;
                        }
                        i++; j--;
                }
        } while(i<=j);
        if(begin<j) sortValues(attribute,values,begin,j);
        if(i<end) sortValues(attribute,values,i,end);
}

void equalfreqDiscretizer::discretizeAttribute(int attribute,int *values,int begin,int end,double *cp,int &numCP) {
	int i;

	double quota=(end-begin+1)/(double)numBins;
	double dBound=0.0;

	int oldBound=0;
	for(i=0;i<numBins-1;i++) {
		dBound+=quota;
		int iBound=(int)round(dBound);
		if(iBound<oldBound) continue;
		if(realValues[values[iBound-1]][attribute]!=realValues[values[iBound]][attribute]) {
                	double cutPoint=(realValues[values[iBound-1]][attribute]+realValues[values[iBound]][attribute])/2.0;
			cp[numCP++]=cutPoint;
		} else {
			double val=realValues[values[iBound]][attribute];
			int numFW=1;
			while(iBound+numFW<=end && realValues[values[iBound+numFW]][attribute]==val) numFW++;
			if(iBound+numFW>end) numFW=end-begin+2;
			int numBW=1;
			while(iBound-numBW>oldBound && realValues[values[iBound-numBW]][attribute]==val) numBW++;
			if(iBound-numBW==oldBound) numBW=end-begin+2;

			if(numFW<numBW) {
				iBound+=numFW;
			} else if(numBW<numFW) {
				iBound-=numBW;
				iBound++;
			} else {
				if(numFW==end-begin+2) {
					return;
				} 
				if(!rnd<0.5) {
					iBound+=numFW;
				} else {
					iBound-=numBW;
					iBound++;
				}
			}
	               	double cutPoint=(realValues[values[iBound-1]][attribute]+realValues[values[iBound]][attribute])/2.0;
			cp[numCP++]=cutPoint;
		}
		oldBound=iBound;
	}
}	
		
equalfreqDiscretizer::equalfreqDiscretizer(int bins,int attributes)
{ 
	numBins=bins;
	numAttributes = attributes;                             // Nombre d'atributs
	demanaVectorDouble(minDomain,numAttributes);		// Valor mínim
	demanaVectorDouble(maxDomain,numAttributes);		// Valor màxim

}

void equalfreqDiscretizer::setBounds(double *min, double *max)
{ 
	for (int i = 0; i < numAttributes; i++)
	{
		minDomain[i] = min[i];
		maxDomain[i] = max[i];
	}
}

void equalfreqDiscretizer::updateDiscretizer(char *dataFile)
{
	readFile(dataFile);
	doDiscretize();
	cleanMemory();
	results();
		
}
int equalfreqDiscretizer::getNumIntervals(int attr)
{
	return numCutPoints[attr]+1;
}

int equalfreqDiscretizer::discretize(int attribute, double value)
{
	if (numCutPoints[attribute] == 0)
		return 0;
	for(int i=0;i<numCutPoints[attribute];i++) {
		if(value<cutPoints[attribute][i]) return i;
	}
	return numCutPoints[attribute];
}

void equalfreqDiscretizer::cleanMemory()
{
	int i;

	for(i=0;i<numInstances;i++) {
		delete missing[i];
		delete realValues[i];
	}
	delete missing;
	delete realValues;
	delete classOfInstances;

	for(i=0;i<numClasses;i++) {
		delete classes[i];
	}
	delete classes;	
}

equalfreqDiscretizer::~equalfreqDiscretizer()
{
	int i;

	for(i=0;i<numAttributes;i++) {
		if(cutPoints[i]) delete cutPoints[i];
	}
	delete realAttributes;
	delete cutPoints;
}
