#include "factory.h"

#include "classifier_gabil.h"
#include "classifier_gabil_packed.h"
#include "classifier_lcs.h"
#include "classifier_adaptive.h"
#include "classifier_hyperrect.h"
#include "classifier_instances.h"
#include "configManagement.h"
#include "instanceSet.h"
#include "attributesInfo.h"

extern configManagement cm;
extern instanceSet *is;
extern attributesInfo ai;

classifierFactory::classifierFactory()
{
	if (cm.thereIsParameter(KR_ADI))
		classifierType = KR_ADI;
	else if (cm.thereIsParameter(KR_HYPERRECT))
		classifierType = KR_HYPERRECT;
	else if (cm.thereIsParameter(KR_INSTANCE_SET))
		classifierType = KR_INSTANCE_SET;
	else if (cm.thereIsParameter(KR_LCS))
		classifierType = KR_LCS;
	else
		classifierType = KR_GABIL;

	if(classifierType == KR_GABIL) {
		usePacked=ai.allAttributesBinary();
	}
}


classifier *classifierFactory::createClassifier()
{
	if (classifierType == KR_ADI)
		return new classifier_adaptive();
	if (classifierType == KR_HYPERRECT)
		return new classifier_hyperrect();
	if (classifierType == KR_INSTANCE_SET)
		return new classifier_instances();
	if (classifierType == KR_LCS)
		return new classifier_lcs();
	if(usePacked) {
		return new classifier_gabil_packed();
	} else {
		return new classifier_gabil();
	}
}

classifier *classifierFactory::cloneClassifier(classifier * orig, int son)
{
	if (classifierType == KR_ADI)
		return new classifier_adaptive(
			*((classifier_adaptive *) orig), son);

	if (classifierType == KR_HYPERRECT)
		return new classifier_hyperrect(
			*((classifier_hyperrect *) orig), son);

	if (classifierType == KR_INSTANCE_SET)
		return new classifier_instances(
			*((classifier_instances *) orig), son);

	if (classifierType == KR_LCS)
		return new classifier_lcs(
			*((classifier_lcs *) orig), son);

	if(usePacked) {
		return new classifier_gabil_packed(*((classifier_gabil_packed *)orig),son);
	} else {
		return new classifier_gabil(*((classifier_gabil *)orig),son);
	}
}

void classifierFactory::deleteClassifier(classifier * ind)
{
	if (classifierType == KR_ADI)
		delete(classifier_adaptive *) ind;
	else if (classifierType == KR_HYPERRECT)
		delete(classifier_hyperrect *) ind;
	else if (classifierType == KR_INSTANCE_SET)
		delete(classifier_instances *) ind;
	else if (classifierType == KR_LCS)
		delete(classifier_lcs *) ind;
	else {
		if(usePacked) {
			delete(classifier_gabil_packed *) ind;
		} else {
			delete(classifier_gabil *) ind;
		}
	}
}
