/************************************************************************
 * ID3 Discretizer					    		*
 *						 Pere Miralles Ribas	*
 ************************************************************************/

#include "id3Discretizer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "messageBuffer.h"

extern messageBuffer mb;

void id3Discretizer::results()
{
        double numpttotals=0;
        double numAttr=0;
        for(int i=0;i<numatributs;i++)
        { // Per a tots els atributs
                if(tipusatributs[i]==3 || tipusatributs[i]==2)
                {
                        numAttr++;
                        numpttotals=numpttotals+numCutPoints[i];
                }
        }
        mb.printf("Numero de punts de tall Discretitzacio : %g\n",numpttotals/numAttr);

}

int id3Discretizer::calculclassesinterval(int inici,int final,int atr)
{ 
	int classesinterval;
	int *classemaj;
	int pos;
	
	//Recompte de les frequencies de les classes a l'interval
	demanaVectorInt(classemaj,numclasses);
	inicontador(classemaj);
	for(int i=inici;i<=final;i++)
	{
		pos=buscanumclasse(classemaj,classes[atr][i]);
		classemaj[pos]=classemaj[pos]+1;
	}
	classesinterval=0;
	for(int i=0;i<numclasses;i++)
	{
		if(classemaj[i]>0)
			classesinterval++;
	}

	return classesinterval;
}

double id3Discretizer::calculentropia(int inici,int final,int atr)
{
	// Calcul de l'entropia
	
	int pos;
	double entropia;
	int *classemaj;
	double prob;
	
	demanaVectorInt(classemaj,numclasses);
	inicontador(classemaj);	
	
	for(int i=inici;i<=final;i++)
	{ 
		pos=buscanumclasse(classemaj,classes[atr][i]);
	        classemaj[pos]=classemaj[pos]+1;
	}
					
	entropia=0;
	for(int classe=0;classe<numclasses;classe++)
	{
		prob=((double)classemaj[classe]) / ((double)(final-inici+1));
		if(prob>0)
			entropia=entropia+(prob*(log10(prob)/log10(2.0)));
	}

	return -entropia;
}
	
void id3Discretizer::inicialitzapt(double *ptpos,int *espt,int atribut,int limit)
{
	for(int i=0;i<limit;i++)
	{
		ptpos[i]=(dadesc[atribut][i]+dadesc[atribut][i+1])/2;
		espt[i]=0; // començem sense cap punt de tall
	}	
}

void id3Discretizer::discretitzador()
{
	int limit;
	double *ptpos;		// Possible pt
	int *espt;		// Indica si el possible pt es pt (1) o no (0)
	double *ent;		// Valor de entropia associat al punt de tall que divideix S en S1 i S2
	int numpt;
		
	for(int i=0;i<numatributs;i++)
	{ // Per a tots els atributs
		if(tipusatributs[i]==3 || tipusatributs[i]==2)
		{ // Nomes per als atributs numerics
			limit=contavalors(i);
			demanaVectorDouble(ptpos,limit-1);
			demanaVectorInt(espt,limit-1);
			demanaVectorDouble(ent,limit-1);
			// Entre tots els valors reals pot existit un punt de tall (valreals-1)
			inicialitzapt(ptpos,espt,i,limit-1);
			assignapt(i,0,numdades-1,ptpos,espt,ent,limit-1);

			// Guardem els pt creats
			numpt=0;
			for(int j=0;j<limit-1;j++)
				if(espt[j])
					numpt++;
			demanaVectorDouble(cutPoints[i],numpt);
			numCutPoints[i]=numpt;
			numpt=0;
			for(int j=0;j<limit-1;j++)
			{
				if(espt[j])
				{
					cutPoints[i][numpt]=ptpos[j];
					numpt++;
				}
			}			
		}
	}
}

void id3Discretizer::assignapt(int atr,int inici,int fi,double *ptpos,int *espt,double *ent,int limit)
{
	int pivot;
	double S,S1,S2;         // Longituds del intervals S,S1 i S2;
	double entS1,entS2,entS;
	int posminim;
	double minim;
	int tall;
	int k,k1,k2;

	if(dades[atr][inici]!=dades[atr][fi] && calculclassesinterval(inici,fi,atr)>1)
	{ // Existeixen pt en l'interval i hi ha més d'una classe
		for(int i=0;i<limit;i++)
		{ // Evaluem tots els possibles pt
			if(ptpos[i]>dades[atr][inici] && ptpos[i]<dades[atr][fi] && !espt[i])
			{// Els que estan dins del domini i no son pt emcara
				// Calcul del pivot
				pivot=inici;
				while(pivot<fi && dades[atr][pivot]<ptpos[i])
					pivot++;
				pivot--;
				S1=(double)(pivot-inici+1);
			       	S2=(double)(fi-(pivot+1)+1);
				S=(double)(fi-inici+1);
				entS1=calculentropia(inici,pivot,atr);
				entS2=calculentropia(pivot+1,fi,atr);
				ent[i]= (S1/S)*entS1 +(S2/S)*entS2;							
			}
		}
		// Busquem el pt que minimitza l'entropia
		minim=MAX_DOUBLE;
		for(int j=0;j<limit;j++)
		{
			if(ptpos[j]>dades[atr][inici] && ptpos[j]<dades[atr][fi] && 
					!espt[j] && ent[j]<minim)
			{
				minim=ent[j];
				posminim=j;
			}
			
		}
		pivot=inici;
		while(pivot<fi && dades[atr][pivot]<ptpos[posminim])
			pivot++;
		pivot--;
		
		espt[posminim]=1;
		assignapt(atr,inici,pivot,ptpos,espt,ent,limit);
		assignapt(atr,pivot+1,fi,ptpos,espt,ent,limit);
		
	}
		
}	

id3Discretizer::id3Discretizer(int attributes)
{ 
	numAttributes = attributes;                             // Nombre d'atributs
	demanaVectorInt(numCutPoints,attributes);               // Numero de punts de tall per atribut
	cutPoints = new double *[numAttributes];                // Punts de tall per atribut
	demanaVectorDouble(minDomain,numAttributes);		// Valor mínim
	demanaVectorDouble(maxDomain,numAttributes);		// Valor màxim
}

void id3Discretizer::setBounds(double *min, double *max)
{ 
	for (int i = 0; i < numAttributes; i++)
	{
		minDomain[i] = min[i];
		maxDomain[i] = max[i];
	}
}

void id3Discretizer::updateDiscretizer(char *dataFile)
{
	parserFitxer(dataFile);
	carregarDades(dataFile);
	compactardades();
	discretitzador();
	results();

}
int id3Discretizer::getNumIntervals(int attr)
{
	return numCutPoints[attr]+1;
}

int id3Discretizer::discretize(int attribute, double value)
{
	if (numCutPoints[attribute] == 0)
		return 0;
	for(int i=0;i<numCutPoints[attribute];i++)
	{
		if(value<cutPoints[attribute][i])
			return i;
	}
	return numCutPoints[attribute];
}
