#include "instance.h"
#include "JVector.h"
#include "JString.h"
#include <math.h>
#include "discretizationManager.h"
#include "attributesInfo.h"
#include "timerSymbolicKR.h"

extern discretizationManager disc;
extern attributesInfo ai;
extern timerSymbolicKR *tSymbolic;

instance::instance(char *string,int pTrainTest)
{
	int i,j;

	traintest=pTrainTest;
	missingValues=0;
	numAttributes=ai.getNumAttributesMC();

	noRealAtt=1;
	if(ai.thereAreRealValuedAttributes()) {
		noRealAtt=0;
		realValues=new double[numAttributes];
	}

	missing=NULL;
	nominalValues=new unsigned char[numAttributes];

	if(ai.allAttributesBinary()) {
		packedRepresentation=new aligned_uint[ai.getRecodeBlocks()];
	} else {
		packedRepresentation=NULL;
	}


	if(!noRealAtt) {
		int num=disc.getNumDiscretizers();
		if(num) {
			nominalValuesFromDiscretizers=new unsigned char[numAttributes*num];
		}
	}

	parseInstance(string);
}

instance::instance()
{
	missingValues=0;
	numAttributes=ai.getNumAttributesMC();
	noRealAtt=1;
	if(ai.thereAreRealValuedAttributes()) {
		noRealAtt=0;
		realValues=new double[numAttributes];
	}
	missing=NULL;
	nominalValues=new unsigned char[numAttributes];

	if(ai.allAttributesBinary()) {
		packedRepresentation=new aligned_uint[ai.getRecodeBlocks()];
	} else {
		packedRepresentation=NULL;
	}

	if(!noRealAtt) {
		int num=disc.getNumDiscretizers();
		if(num) {
			nominalValuesFromDiscretizers=new unsigned char[numAttributes*num];
		}
	}
}


int instance::extractNominalValue(char *instance,int attribute)
{
	if(attribute<numAttributes && missingValues && missing[attribute]) return -1;
	int value=ai.valueOfNominalAttribute(attribute,instance);

	/*if(value==-1) {
		fprintf(stderr,"Instance:%d. Wrong nominal value of attribute %d:|%s|\n",id,attribute,instance);
		exit(1);
	}*/

	return value;
} 

void instance::generatePackedRepresentation()
{
	int recodeBlocks=ai.getRecodeBlocks();
	int i,j,countAtt;
	
	for(i=0,countAtt=0;i<recodeBlocks;i++) {
		unsigned int tmp;
		for(j=0;j<ATT_PER_BLOCK;j++,countAtt++) {
			if(countAtt<numAttributes) {
				if(missingValues && missing[countAtt]) {
					SET_ATT(tmp,j,0);
				} else {
					SET_ATT(tmp,j,nominalValues[countAtt]+1);
				}
			} else {
				SET_ATT(tmp,j,0);
			}
		}
		packedRepresentation[i]=tmp;
	}

	//delete nominalValues;
	//nominalValues=NULL;
}
 
int instance::extractRealValue(double instance,int attribute)
{
	if(!disc.isReady()) return -1;

	if(traintest==TEST) {
		double min=ai.getMinDomain(attribute);
		double max=ai.getMaxDomain(attribute);

		if(instance<min) {
			//printf("Value smaller than current minimum: %f (min %f)\n"
			//	,instance,min);
			instance=min;
		}
		if(instance>max) {
			//printf("Value greater than current maximum: %f (max %f)\n"
			//	,instance,max);
			instance=max;
		}
	}

	int i;
	int num=disc.getNumDiscretizers();
	int defaultInterval;
	for(i=0;i<num;i++) {
		int value=disc.getDiscretizer(i)->discretize(attribute,instance);
		if(!i) defaultInterval=value;
		nominalValuesFromDiscretizers[i*numAttributes+attribute]=value;
	}

	return defaultInterval;
}

void instance::parseInstance(char *string)
{
	int i,j;
	char *token;

	token=strtok(string,",");
	for(i=0;i<=numAttributes && token;i++) {
		if(ai.getTypeOfAttribute(i)==NOMINAL) {
			if(token[0]=='?') {
				if(missing==NULL) {
					missing = new char[numAttributes];
					for(j=0;j<numAttributes;j++) {
						missing[j]=0;
					}
				}
				missing[i]=1;
				missingValues=1;
			}
			if(i<numAttributes) {
				nominalValues[i]=extractNominalValue(token,i);
			} else {
				instanceClass=extractNominalValue(token,numAttributes);
			}
		} else {
			if(token[0]=='?') {
				if(missing==NULL) {
					missing = new char[numAttributes];
					for(j=0;j<numAttributes;j++) {
						missing[j]=0;
					}
				}

				missing[i]=1;
				missingValues=1;
			} else {
				float temp;
				if(sscanf(token,"%f",&temp)==0) {
					fprintf(stderr,
					  "Incorrect real value:%s\n",token);
					exit(1);
				}
				realValues[i]=temp;
			}
		}
		token=strtok(NULL,",");
	}
	if(token!=NULL) {
		fprintf(stderr,"Incorrect instance:%s\n",string);
		exit(1);
	}
}

instance::~instance()
{
	if(nominalValues) delete nominalValues;
	if(missing) delete missing;
	if(!noRealAtt) {
		delete realValues;
		if(disc.getNumDiscretizers()) {
			delete nominalValuesFromDiscretizers;
		}
	}
	if(packedRepresentation) delete packedRepresentation;
}

void instance::updateRealValuedAttributes()
{
	int i;

	if(!missingValues) {
		for(i=0;i<numAttributes;i++) {
			if(ai.getTypeOfAttribute(i)==REAL) {
				nominalValues[i]=extractRealValue(realValues[i],i);
			}
		}
	} else {
		for(i=0;i<numAttributes;i++) {
			if(ai.getTypeOfAttribute(i)==REAL && !missing[i]) {
				nominalValues[i]=extractRealValue(realValues[i],i);
			}
		}
	}
}

void instance::updateMissing()
{
	int i;

	if(!missingValues) return;

	for(i=0;i<numAttributes;i++) {
		if(missing[i]) {
			if(ai.getTypeOfAttribute(i)==REAL) {
				double val=ai.getAverageOfAttribute(instanceClass,i);
				realValues[i]=val;
				//printf("Value for attribute %d %g\n",i,val);
			} else if(ai.getTypeOfAttribute(i)==NOMINAL) {
				int val=ai.getMostFrequentValueOfAttribute(instanceClass,i);
				nominalValues[i]=val;
				//printf("Value for attribute %d %d\n",i,val);
			}
			missing[i]=0;
		}
	}

	missingValues=0;
	delete missing;
	missing=NULL;
}

double instance::distance(instance *ins) 
{
	int i;
	double dist=0;
	for(i=0;i<numAttributes;i++) {
		if (ai.getTypeOfAttribute(i) == REAL) {
			double dom=ai.getSizeDomain(i);
			if(dom>0) {
				double distTmp =
				    ins->realValueOfAttribute(i) -
				    realValues[i];
				distTmp /= dom;
				distTmp *= distTmp;
				dist += distTmp;
			}
		} else {
			if (nominalValues[i] !=
			    ins->valueOfAttribute(i)) {
				dist++;
			}
		}
	}
	return dist;
}

void instance::dumpInstance() 
{
        int i;
        for(i=0;i<numAttributes;i++) {
                if(ai.getTypeOfAttribute(i)==NOMINAL) {
                        printf("%s,",ai.getNominalValue(i, nominalValues[i])->cstr());
                } else {
                        printf("%.3f,",realValues[i]);
                }
        }
        printf("%s\n",ai.getNominalValue(numAttributes, instanceClass)->cstr());
}

