#include "instanceSetFile.h"
#include "configManagement.h"
#include "instance.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "JVector.h"
#include "JString.h"
#include "attributesInfo.h"
#include "utils.h"
#include "windowingILAS.h"
#include "windowingGWS.h"
#include "random.h"
#include "messageBuffer.h"
#include "timeManagement.h"
#include "discretizationManager.h"


extern messageBuffer mb;
extern attributesInfo ai;
extern Random rnd;
extern configManagement cm;
extern timeManagement tm;
extern discretizationManager disc;

void instanceSetFile::readFile(char fileName[], unsigned long long &numInstances, int traintest)
{
	FILE *fp;
	char string[100000];
	JVector<instance *> tempSet(1000,100000);
	unsigned long long num = 0;

	fp = fopen(fileName, "r");
	if (!fp) {
		fprintf(stderr,"Can't open %s\n",fileName);
		exit(1);
	}

	parseHeader(fp, traintest);
	if(traintest==TRAIN) {
		disc.setNumAttributes(ai.getNumAttributesMC());
		disc.updateDiscretizers(fileName);
		ai.finish();
	}

	fgets(string, 99999, fp);
	while (!feof(fp)) {
		string[strlen(string) - 1] = 0;
		if(string[strlen(string) - 1]==13) 
			string[strlen(string) - 1] = 0;
		if (string[0] != '%' && strlen(string) > 0) {
			instance *tmp=new instance(string,traintest);
			if(traintest==TRAIN) ai.insertInstance(tmp);
			tempSet.addElement(tmp);
			num++;
		}
		fgets(string, 9999, fp);
	}
	fclose(fp);

	numInstances = num;
	set=new instance*[num];
	for(int i=0;i<num;i++) {
		set[i]=tempSet[i];
	}
	tempSet.removeAllElements();
	if(traintest==TRAIN) ai.calculateAverages();
}

void instanceSetFile::parseRelation(char *string)
{
	char cadena[5000];

	string += 9;
	while (*string == ' ' || *string == '\t')
		string++;
	mb.printf("Dataset name: %s\n", string);
}

void instanceSetFile::parseReal(int numAttr)
{
	ai.setTypeOfAttribute(numAttr,REAL);
	mb.printf("Attribute %d real valued\n", numAttr);
}

void instanceSetFile::parseInteger(char *string, int numAttr)
{
	int min, max;
	if (sscanf(string+8, "[%d,%d]", &min, &max) != 2) {
		fprintf(stderr,"Parse error:%s\n", string+8);
		exit(1);
	}
	if (min >= max) {
		fprintf(stderr,"Attribute %d inconsistent:%d %d\n",
		       numAttr, min, max);
		exit(1);
	}
	ai.setTypeOfAttribute(numAttr,REAL);
	mb.printf("Attribute %d integer [%d:%d]\n", numAttr, min, max);

}

void instanceSetFile::parseNominal(char *string, int numAttr)
{
	mb.printf("Attribute %d nominal\n", numAttr);
	ai.setTypeOfAttribute(numAttr,NOMINAL);
	int numValues = 0;
	char *ptr = string;
	ptr++;
	while (*ptr != '}') {
		char value[500];
		int size;
		while (*ptr == ' ' || *ptr == '\t') ptr++;
		for (size = 0;
		     !(*ptr == 0 || *ptr == ',' || *ptr == '}'); 
		     value[size++] = *ptr++);
		if (*ptr == 0) {
			fprintf(stderr,"Parse error: %s\n", string);
			exit(1);
		}
		value[size] = 0;
		if (size > 0) {
			JString *str = new JString(value);
			ai.insertNominalValue(numAttr,str);
			mb.printf("Value %d of attribute %d: %s\n", numValues,
			       numAttr, value);
			numValues++;
		}
		if (*ptr == ',') ptr++;
	}
}

void instanceSetFile::parseAttribute(char *string, int numAttr)
{
	char name[5000];

	string += 10;
	while (*string == ' ' || *string == '\t')
		string++;

	if(*string=='\'') {
		string++;
		int count=0;
		while(*string!='\'') {
			name[count++]=*string++;
		}
		name[count]=0;
		string++;
	} else {
		if (sscanf(string, "%s", name) != 1) {
			fprintf(stderr,"Parse error:%s\n", string);
			exit(1);
		}
		string += strlen(name);
	}


	while (*string == ' ' || *string == '\t')
		string++;

	mb.printf("Attribute %d:Name %s Def:%s\n", numAttr, name, string);

	ai.insertAttributeName(new JString(name));

	if (!strcasecmp(string, "real") || !strcasecmp(string, "numeric")) {
		parseReal(numAttr);
	} else if (!strncasecmp(string, "integer", 7)) {
		parseInteger(string, numAttr);
	} else if (string[0] == '{') {
		parseNominal(string, numAttr);
	} else {
		fprintf(stderr,"Unknown attribute type %s\n", string);
		exit(1);
	}
}

void instanceSetFile::parseHeader(FILE * fp,int traintest)
{
	JVector<char *>header;

	char string[10000];
	int end = 0;
	int numAttr = 0;

	fgets(string, 9999, fp);
	while (!feof(fp) && !end) {
		string[strlen(string)-1]=0;
		if(string[strlen(string) - 1]==13) 
			string[strlen(string) - 1] = 0;
		if (string[0] != '%' && strlen(string) > 1) {
			if (!strncasecmp(string, "@relation", 9)) {
				if (traintest == TRAIN) parseRelation(string);
			} else if (!strncasecmp(string, "@attribute", 10)) {
				if (traintest == TRAIN) {
					char *tmp=new char[strlen(string)+1];
					strcpy(tmp,string);
					header.addElement(tmp);
					numAttr++;
				}
			} else if (!strncasecmp(string, "@data", 5)) {
				end = 1;
			} else {
				fprintf(stderr,"Unknown header element:%s\n"
					, string);
				exit(1);
			}
		}
		if(!end) 
			fgets(string, 9999, fp);
	}

	if (traintest == TRAIN) {
		ai.setNumAttributes(numAttr);
		for(int i=0;i<numAttr;i++) {
			parseAttribute(header[i], i);
			delete header[i];
		}
		if (ai.getTypeOfAttribute(numAttr - 1) != NOMINAL) {
			fprintf(stderr,"Class attribute (last) should be nominal\n");
			exit(1);
		}
	}
}

void instanceSetFile::initializeWindowing(int traintest)
{
	if(traintest == TRAIN) {
		if(cm.thereIsParameter(WINDOWING_ILAS)) {
			windowingEnabled=1;
			win=new windowingILAS;
		} else if(cm.thereIsParameter(WINDOWING_GWS)) {
			windowingEnabled=1;
			win=new windowingGWS;
		} else {
			windowingEnabled=0;
			window=NULL;
		}

		if(windowingEnabled) {
			win->setInstances(set,numInstances);
			win->newIteration(window,windowSize);
		}
	} else {
		windowingEnabled=0;
		window=NULL;
	}
}

instanceSetFile::instanceSetFile(char fileName[], int traintest)
{
	unsigned long long i;

	readFile(fileName, numInstances, traintest);
	if (numInstances==0) {
		fprintf(stderr,"Instances file %s is empty\n",fileName);
		exit(1);
	}

	for(i=0;i<numInstances;i++) {
		if(!cm.thereIsParameter(IGNORE_MISSING_VALUES)) 
			set[i]->updateMissing();
		set[i]->updateRealValuedAttributes();
		if(ai.allAttributesBinary()) 
			set[i]->generatePackedRepresentation();
	}
	initializeWindowing(traintest);

	if(traintest==TRAIN) {
		numClasses=ai.getNumClasses();
		classWiseInit=cm.thereIsParameter(CLASS_WISE_INIT);
		initInstanceLists();
	} else {
		initSamplings=NULL;
		countsByClass=NULL;
		instByClass=NULL;
	}
}

instanceSetFile::~instanceSetFile()
{
	unsigned long long i;

	for (i = 0; i < numInstances; i++)
		delete set[i];
	delete set;
	if(initSamplings) {
		for(i=0;i<numClasses;i++) {
			delete initSamplings[i];
			delete instByClass[i];
		}
		delete initSamplings;
		delete instByClass;
		delete countsByClass;
	}
	if(window) delete window;
}

unsigned long long instanceSetFile::getNumInstancesOfIteration()
{
	if(windowingEnabled) return windowSize;
	return numInstances;
}

int instanceSetFile::newIteration(int isLast)
{
	if(isLast) {
		//mb.printf("Windowing disabled for last iteration\n");
		windowingEnabled=0;
		return 1;
	}

	if(windowingEnabled) {
		win->newIteration(window,windowSize);
		if(win->needReEval()) return 1;
		return 0;
	}

	return 0;
}

void instanceSetFile::initInstanceLists()
{
	int i;

	countsByClass = new int[numClasses];
	initSamplings = new Sampling *[numClasses];
	instByClass = new int *[numClasses];

	if(windowingEnabled) {
		int numInst=(int)getNumInstancesOfIteration();

		for(i=0;i<numClasses;i++) {
			int num=ai.getInstancesOfClass(i);
			instByClass[i] = new int[num];
			countsByClass[i]=0;
		}
	
		for(i=0;i<numInst;i++) {
			int cl=window[i]->getClass();
			instByClass[cl][countsByClass[cl]++]=i;
		}

		for(i=0;i<numClasses;i++) {
			initSamplings[i] = new Sampling(countsByClass[i]);
		}
	} else {
		perror("Not supported");
		exit(1);
	}
}
	
instance *instanceSetFile::getInstanceInit(int forbiddenCL) 
{
	if(classWiseInit) {
		int cl;
		do {
			cl=rnd(0,numClasses-1);
		} while(cl==forbiddenCL || countsByClass[cl]==0);

		int pos=initSamplings[cl]->getSample();
		int insIndex=instByClass[cl][pos];
		instance *ins;

		if(windowingEnabled) ins=window[insIndex];
		else ins=set[insIndex];

		return ins;
	} else {
		int nc=numClasses;
		int count[nc];
		int total=0;
		int i;
		if(forbiddenCL!=numClasses) nc--;
	
		for(i=0;i<nc;i++) {
			if(i<forbiddenCL) 
				count[i]=initSamplings[i]->numSamplesLeft();
			else 
				count[i]=initSamplings[i+1]->numSamplesLeft();
			total+=count[i];
		}
		int pos=rnd(0,total-1);
		int acum=0;
		int found=0;
		for(i=0;i<nc && !found;i++) {
			acum+=count[i];
			if(pos<acum) {
				found=1;
			}
		}
		i--;
		if(i>=forbiddenCL) i++;

		pos=initSamplings[i]->getSample();
		int insIndex=instByClass[i][pos];

		instance *ins;
		if(windowingEnabled) ins=window[insIndex];
		else ins=set[insIndex];

		return ins;
	}
}

int instanceSetFile::numVersions()
{
        if(windowingEnabled) return win->numVersions();
        return 1;
}

int instanceSetFile::getCurrentVersion()
{
        if(windowingEnabled) return win->getCurrentVersion();
        return 0;
}

