#ifndef _INSTANCE_SET_FILE_
#define _INSTANCE_SET_FILE_

#include <stdio.h>
#include "JVector.h"
#include "windowing.h"
#include "sampling.h"
#include "instanceSet.h"

#define TRAIN 1
#define TEST 2

class instance;

class instanceSetFile : public instanceSet {
private:
	Sampling **initSamplings;
	int *countsByClass;
	int **instByClass;
	void initInstanceLists();
	int numClasses;
	int classWiseInit;

	instance **set;
	unsigned long long numInstances;

	int windowingEnabled;
	instance **window;
	int windowSize;
	windowing *win;

	void readFile(char fileName[],unsigned long long &numInstances,int traintest);
	void parseHeader(FILE *,int traintest);
	void parseRelation(char *string);
	void parseReal(int numAtr);
	void parseInteger(char *string,int numAtr);
	void parseNominal(char *string,int numAtr);
	void parseAttribute(char *string,int numAtr);
	void initializeWindowing(int traintest);

public:
	instanceSetFile(char *fileName,int traintest);
	~instanceSetFile();

	inline instance *getInstanceTest(unsigned long long index) {
		return set[index];
	}
	inline instance *getInstance(unsigned long long index) {
		if(windowingEnabled) return window[index];
		return set[index];
	}
	inline instance **getInstancesOfIteration() {
		if(windowingEnabled) return window;
		return set;
	}

	inline instance **getAllInstances() {
		return set;
	}


	instance *getInstanceInit(int forbiddenCL);

	inline int isWindowingEnabled() {return windowingEnabled;}
	unsigned long long getNumInstancesOfIteration();
	int newIteration(int isLast);
	inline unsigned long long getNumInstances(){return numInstances;}
	int numVersions();
	int getCurrentVersion();
};



#endif
