#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>

#include "configManagement.h"
#include "populationWrapper.h"
#include "classifierFitness.h"
#include "random.h"
#include "instanceSetFile.h"
#include "timersManagement.h"
#include "discretizationManager.h"
#include "attributesInfo.h"
#include "lex_conf.h"
#include "timeManagement.h"
#include "messageBuffer.h"

int stop = 0;
messageBuffer mb;
discretizationManager disc;
attributesInfo ai;
configManagement cm;
instanceSet *is;
timeManagement timeM;
Random rnd;
double percentageOfLearning=0;
int lastIteration=0;
int nodeRank;
int numTasks;

void handler(int sig)
{
	stop = 1;
}

int main(int argc, char *argv[])
{
	int rc;

        if (argc < 4) {
		fprintf(stderr,"Incorrect parameters\n"
                       "%s: <Config file> <Train file> <Test file>\n",
		       argv[0]);
		exit(1);
	}

        //if(argc==5) mb.setFile(argv[4]);


	parseConfig(argv[1]);
	rnd.dumpSeed();

	is=new instanceSetFile(argv[2], TRAIN);
	timersManagement timers;
	timeM.resetTime();

	populationWrapper pw((int)cm.getParameter(POP_SIZE));

	timers.setPW(&pw);
	signal(SIGINT, handler);

	double ave,dev;
	pw.getAverageDevAccuracy(ave,dev);
	mb.printf("Initial population average accuracy:%f deviation:%f\n",ave,dev);
	double totalTime=-1;
	if(cm.thereIsParameter(TOTAL_TIME)) 
		totalTime=cm.getParameter(TOTAL_TIME);

	int countIt=0,numIterations = (int) cm.getParameter(ITERATIONS);
	if(numIterations==0) stop=1;

	int dumpGenotype=-1;
	if(cm.thereIsParameter(DUMP_GENOTYPE_ITERATIONS)) {
		dumpGenotype=(int)cm.getParameter(DUMP_GENOTYPE_ITERATIONS);
	}

	for (;!stop;) {
		timers.incIteration(lastIteration);

		if (is->newIteration(lastIteration)) {
			pw.activateModifiedFlag();
		}

		if(dumpGenotype!=-1 && !(countIt%dumpGenotype)) {
			classifier *ind=pw.getBestPopulation();
			char string[100000];
			mb.printf("Dumping genotype\n");
			ind->dumpGenotype(string);
			mb.printf("%s\n",string);
		}

		pw.gaIteration();

		//classifier *ind=pw.getBestOverall();
		//if(classifierTest(*ind,is)==1.0) {
		//	stop=1;
		//}

		timers.dumpStats();
		countIt++;
		if(totalTime==-1) {
			if(countIt==numIterations) stop=1;
			else if(countIt==numIterations-1) lastIteration=1;
			percentageOfLearning=(double)countIt
				/(double)(numIterations);
		} else {
			if(lastIteration) stop=1;
			else {
				double currentTime=timeM.totalTime();
				percentageOfLearning=currentTime/totalTime;
				if(percentageOfLearning>1) 
					percentageOfLearning=1;
				if(currentTime>=totalTime) {
					mb.printf("Total number of iterations in time %d : %d\n",(int)totalTime,countIt+1);
					lastIteration=1;
				}
			}
		}
		//mb.printf("Total time: %g\n",timeM.totalTime());
	}

	classifier *ind = pw.getBestOverall();
	//ind->postProcessIndividual();

	char phenotype[1000000];
	ind->dumpPhenotype(phenotype);
	mb.printf("Phenotype: \n%s\n", phenotype);

	classifierStats(*ind, argv[2], "Train");
	classifierStats(*ind, argv[3], "Test");

	//delete is;
	return 0;
}
