#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>

#include "configManagement.h"
#include "populationWrapper.h"
#include "classifierFitness.h"
#include "random.h"
#include "instanceSetVirtual.h"
#include "timersManagement.h"
#include "discretizationManager.h"
#include "attributesInfo.h"
#include "lex_conf.h"
#include "timeManagement.h"
#include "messageBuffer.h"

int stop = 0;
messageBuffer mb;
discretizationManager disc;
attributesInfo ai;
configManagement cm;
instanceSet *is;
timeManagement timeM;
Random rnd;
double percentageOfLearning=0;
int lastIteration=0;
int nodeRank;
int numTasks;

void handler(int sig)
{
	stop = 1;
}

int main(int argc, char *argv[])
{
	int rc;

	if (argc < 2) {
		fprintf(stderr,"Incorrect parameters\n"
		       "%s: <Config file>\n",
		       argv[0]);
		exit(1);
	}

	if(argc==3) mb.setFile(argv[2]);

	parseConfig(argv[1]);
	rnd.dumpSeed();
	is=new instanceSetVirtual(TRAIN);
	timersManagement timers;
	timeM.resetTime();

	populationWrapper pw((int)cm.getParameter(POP_SIZE));

	timers.setPW(&pw);
	signal(SIGINT, handler);

	double ave,dev;
	pw.getAverageDevAccuracy(ave,dev);
	mb.printf("Initial population average accuracy:%f deviation:%f\n",ave,dev);
	double totalTime=-1;
	if(cm.thereIsParameter(TOTAL_TIME)) 
		totalTime=cm.getParameter(TOTAL_TIME);

	int countIt=0,numIterations = (int) cm.getParameter(ITERATIONS);
	if(numIterations==0) stop=1;

	for (;!stop;) {
		timers.incIteration(lastIteration);

		if (is->newIteration(lastIteration)) {
			pw.activateModifiedFlag();
		}

		pw.gaIteration();

		//classifier *ind=pw.getBestOverall();
		//if(classifierTestVirtual(*ind,is)==1.0) {
		//	stop=1;
		//}

		timers.dumpStats();
		countIt++;

		if(countIt==numIterations) stop=1;
		else if(countIt==numIterations-1) lastIteration=1;
		percentageOfLearning=(double)countIt/(double)numIterations;

		mb.printf("Total time: %g\n",timeM.totalTime());
	}

	classifier *ind = pw.getBestOverall();

	char phenotype[1000000];
	ind->dumpPhenotype(phenotype);
	mb.printf("Phenotype: \n%s\n", phenotype);

	//classifierStatsVirtual(*ind, (instanceSetVirtual *)is);

	delete is;
	return 0;
}
