/************************************************************************
 * Mantaras Distance Discretizer					*
 ************************************************************************/

#include "mantarasDiscretizer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "messageBuffer.h"

extern messageBuffer mb;

void mantarasDiscretizer::results()
{
	int i;

	double sum=0;
	double count=0;
	for(i=0;i<numAttributes;i++) {
		if(realAttributes[i]) {
			sum+=numCutPoints[i];
			count++;
                        //mb.printf("Atribut: %d Numero de punts de tall: %d\n",i,numCutPoints[i]);
                        //for(int j=0;j<numCutPoints[i];j++)
                        //        mb.printf("[%d] : %f\n",j,(float)cutPoints[i][j]);

		}
	}
	mb.printf("Numero de punts de tall Discretitzacio : %g\n",sum/count);
}

void mantarasDiscretizer::readFile(char *datafile)
{
	char buffer[10000];
	int i,j;

	realAttributes=new bool[numAttributes];
	cutPoints = new double *[numAttributes];
	numCutPoints = new int[numAttributes];
	int attributeCount=0;
	for(i=0;i<numAttributes;i++) {
		cutPoints[i]=NULL;
	}
	FILE *fp=fopen(datafile,"r");

	fgets(buffer,9999,fp);
	while(!feof(fp) && strncasecmp(buffer,"@data",5)) {
		if(!strncasecmp(buffer,"@attribute",10)) {
			char *ptr=buffer+10;
			if(attributeCount==numAttributes) {
				while(*ptr==' ' || *ptr=='\t') ptr++;
				while(*ptr!=' ' && *ptr!='\t') ptr++;
				while(*ptr==' ' || *ptr=='\t') ptr++;
				if(*ptr!='{') {
					fprintf(stderr,"Wrong format !!\n");
					exit(1);
				}
				ptr++;
				numClasses=0;
				for(i=0;i<strlen(ptr);i++) {
					if(ptr[i]==',' || ptr[i]=='}') numClasses++;
				}
				classes=new char *[numClasses];
				cd=new int[numClasses];
				classCount=new int[numClasses];
				for(i=0;i<numClasses;i++) {
					char tmp[1000];
					int index=0;
					while(*ptr!=',' && *ptr!='}') {
						tmp[index++]=*ptr++;
					}
					tmp[index]=0;
					classes[i]=new char[strlen(tmp)+1];
					strcpy(classes[i],tmp);
					ptr++;
				}
			} else {
				while(*ptr==' ' || *ptr=='\t') ptr++;
				while(*ptr!=' ' && *ptr!='\t') ptr++;
				while(*ptr==' ' || *ptr=='\t') ptr++;
				if(!strncasecmp(ptr,"real",4) || !strncasecmp(ptr,"integer",7)) {
					realAttributes[attributeCount]=true;
				} else {
					realAttributes[attributeCount]=false;
				}
			}
			attributeCount++;
		}
		fgets(buffer,9999,fp);
	}

	int pos=ftell(fp);
	numInstances=0;
	fgets(buffer,9999,fp);
	while(!feof(fp)) {
		numInstances++;
		fgets(buffer,9999,fp);
	}
	fseek(fp,pos,SEEK_SET);

	missing = new bool *[numInstances];
	realValues = new double *[numInstances];
	classOfInstances = new int[numInstances];
	candidateCutPoints = new int[numInstances];
	cutPointsMask = new int[numInstances];

	fgets(buffer,9999,fp);
	int instanceCount=0;
	while(!feof(fp)) {
		buffer[strlen(buffer)-1]=0;
		if(buffer[0] && buffer[0]!='%') {		
			missing[instanceCount] = new bool[numAttributes];
			realValues[instanceCount] = new double[numAttributes];
			char *ptr=buffer;
			for(j=0;j<numAttributes;j++) {
				char tmp[1000];
				int index=0;
				while(*ptr!=',') {
					tmp[index++]=*ptr++;
				}
				tmp[index]=0;
				if(realAttributes[j]) {
					if(tmp[0]=='?') {
						missing[instanceCount][j]=true;
					} else {
						missing[instanceCount][j]=false;
						float tempFloat;
						if(sscanf(tmp,"%f",&tempFloat)!=1) {
							fprintf(stderr,"Wrong format 2 %s!!!\n",ptr);
							exit(1);
						}	
						realValues[instanceCount][j]=tempFloat;
					}
				}
				ptr++;
			}
			bool found=false;
			for(j=0;j<numClasses && !found;j++) {
				if(!strcasecmp(classes[j],ptr)) {
					classOfInstances[instanceCount]=j;
					found=true;
				}
			}
			if(!found) {
				fprintf(stderr,"Wrong format 3 !!!\n");
				exit(1);
			}
			instanceCount++;
		}
		fgets(buffer,9999,fp);
	}

	fclose(fp);
}
	
	
void mantarasDiscretizer::doDiscretize() 
{
	int i,j;
	int points[numInstances];
	double cp[numInstances];

	for(i=0;i<numAttributes;i++) {
		if(realAttributes[i]) {
			int numPoints=0;
			for(j=0;j<numInstances;j++) {
				if(!missing[j][i]) {
					points[numPoints++]=j;
				}
			}

			if(numPoints) {
				sortValues(i,points,0,numPoints-1);
	
				int numCP=0;
				discretizeAttribute(i,points,0,numPoints-1,cp,numCP);
				numCutPoints[i]=numCP;
				//printf("Attribute %d, num cut points : %d\n",i,numCutPoints[i]);
				if(numCP>0) {
					cutPoints[i]=new double[numCP];
					for(j=0;j<numCP;j++) {
						cutPoints[i][j]=cp[j];
						//printf("Attribute %d, cut point %d : %f\n",i,j,cutPoints[i][j]);
					}
				}
			} else {
				numCutPoints[i]=0;
			}
		} else {
			numCutPoints[i]=0;
		}
	}
}

void mantarasDiscretizer::sortValues(int attribute,int *values,int begin,int end) 
{
        double pivot;
        int temp;
        int i,j;

        i=begin;j=end;
        pivot=realValues[values[(i+j)/2]][attribute];
        do {
                while(realValues[values[i]][attribute]<pivot) i++;
                while(realValues[values[j]][attribute]>pivot) j--;
                if(i<=j) {
                        if(i<j) {
                                temp=values[i];
                                values[i]=values[j];
                                values[j]=temp;
                        }
                        i++; j--;
                }
        } while(i<=j);
        if(begin<j) sortValues(attribute,values,begin,j);
        if(i<end) sortValues(attribute,values,i,end);
}

void mantarasDiscretizer::discretizeAttribute(int attribute,int *values,int begin,int end,double *cp,int &numCP) {
	int i,j,k;

	double entClass=computeClassEntropy(attribute,values,begin,end);
	if(entClass==0) return;
	getCandidateCutPoints(attribute,values,begin,end);
        if(numCandidates==0) return;

	int pos=getBestCandidate(attribute,values,begin,end,entClass);
	while(doCutPoint(attribute,values,begin,end,pos)) {
		cutPointsMask[pos]=1;
		int posMin=candidateCutPoints[pos];
                double cutPoint=(realValues[values[posMin-1]][attribute]+realValues[values[posMin]][attribute])/2.0;
		for(j=0;j<numCP && cp[j]<cutPoint;j++);
		for(k=numCP-1;k>=j;k--) {
			cp[k+1]=cp[k];
		}
		cp[j]=cutPoint;
		numCP++;
	
		pos=getBestCandidate(attribute,values,begin,end,entClass);
	}
}	

int mantarasDiscretizer::getBestCandidate(int attribute,int *values,int begin,int end,double entClass) {
	int first=1;
	int i;
	double distMin;
	int posMin;

	for(i=0;i<numCandidates;i++) {
		if(cutPointsMask[i]==0) {
			
			double dist=computePartitionDistance(attribute,values,begin,end,i,entClass);
			if(first) {
				distMin=dist;
				posMin=i;
				first=0;
			} else {
				if(dist-distMin<-0.0000001) {
					distMin=dist;
					posMin=i;
				}
			}
		}
	}
	return posMin;
}
		
double mantarasDiscretizer::computePartitionDistance(int attribute,int *values,int begin,int end,int newCP,double entClass) {
	int i,j,k;

	double entInt=0;
	double entDisc=0;

	cutPointsMask[newCP]=1;
	int index=0;
	int partB=begin;
	while(index<numCandidates) {
		while(index<numCandidates && cutPointsMask[index]==0) index++;
		int partE;
		if(index==numCandidates) partE=end;
		else partE=candidateCutPoints[index]-1;

		classDistribution(attribute,values,partB,partE,cd,numC,numV);
		entInt+=computeEntropy(cd,numC,end-begin+1);
		double prop=(double)numV/(double)(end-begin+1);
		entDisc+=prop*log2(prop);

		partB=partE+1;
		index++;
	}
	entDisc=-entDisc;
	double res=2.0-(entDisc+entClass)/entInt;
	cutPointsMask[newCP]=0;
	return res;
}

double mantarasDiscretizer::log2(double val)
{
	return log(val)/log(2.0);
}

int mantarasDiscretizer::doCutPoint(int attribute,int *values,int begin,int end,int cutP) {
	double lengthOrig=doCutPoint2(attribute,values,begin,end);
	cutPointsMask[cutP]=1;
	double lengthNew=doCutPoint2(attribute,values,begin,end);
	cutPointsMask[cutP]=0;

	if(lengthNew<lengthOrig) return 1;
	return 0;
}

double mantarasDiscretizer::doCutPoint2(int attribute,int *values,int begin,int end) {
	int i,j,k;
	double entAcum=0;
	int index=0;
	int partB=begin;
	int numCP=0;

	while(index<numCandidates) {
		while(index<numCandidates && cutPointsMask[index]==0) index++;
		int partE;
		if(index==numCandidates) partE=end;
		else {
			partE=candidateCutPoints[index]-1;
			numCP++;
		}

		classDistribution(attribute,values,partB,partE,cd,numC,numV);

		double ent=computeEntropy(cd,numC,numV);
		entAcum+=ent*(double)(numC+numV);

		partB=partE+1;
		index++;
	}
	entAcum+=(double)numCP*log2(end-begin)+(double)((numCP+1)*numClasses);
	return entAcum;
}

double mantarasDiscretizer::computeEntropy(int *cd,int numC,int numV) {
        double ent=0;

        for(int i=0;i<numC;i++) {
                double prob=cd[i]/(double)numV;
                ent+=prob*log(prob)/log(2.0);
        }
        return -ent;
}


void mantarasDiscretizer::getCandidateCutPoints(int attribute,int *values,int begin,int end) {
	numCandidates=0;
        double valueAnt=realValues[values[begin]][attribute];

        for(int i=begin;i<=end;i++) {
                double val=realValues[values[i]][attribute];
                if(val!=valueAnt) {
			cutPointsMask[numCandidates]=0;
			candidateCutPoints[numCandidates++]=i;
		}
                valueAnt=val;
        }
}

void mantarasDiscretizer::classDistribution(int attribute,int *values,int begin,int end,int *cd,int &numC,int &numV) {
	int i;

        for(i=0;i<numClasses;i++) classCount[i]=0;
        for(i=begin;i<=end;i++) classCount[classOfInstances[values[i]]]++;

	numV=numC=0;
        for(i=0;i<numClasses;i++) {
                if(classCount[i]>0) {
			cd[numC]=classCount[i];
			numV+=classCount[i];
			numC++;
		}
        }
}

double mantarasDiscretizer::computeClassEntropy(int attribute,int *values,int begin,int end) {
	classDistribution(attribute,values,begin,end,cd,numC,numV);
	return computeEntropy(cd,numC,numV);
}


mantarasDiscretizer::mantarasDiscretizer(int attributes)
{ 
	numAttributes = attributes;                             // Nombre d'atributs
	demanaVectorDouble(minDomain,numAttributes);		// Valor mínim
	demanaVectorDouble(maxDomain,numAttributes);		// Valor màxim

}

void mantarasDiscretizer::setBounds(double *min, double *max)
{ 
	for (int i = 0; i < numAttributes; i++)
	{
		minDomain[i] = min[i];
		maxDomain[i] = max[i];
	}
}

void mantarasDiscretizer::updateDiscretizer(char *dataFile)
{
	readFile(dataFile);
	doDiscretize();
	cleanMemory();
	results();
		
}
int mantarasDiscretizer::getNumIntervals(int attr)
{
	return numCutPoints[attr]+1;
}

int mantarasDiscretizer::discretize(int attribute, double value)
{
	if (numCutPoints[attribute] == 0)
		return 0;
	for(int i=0;i<numCutPoints[attribute];i++) {
		if(value<cutPoints[attribute][i]) return i;
	}
	return numCutPoints[attribute];
}

void mantarasDiscretizer::cleanMemory()
{
	int i;

	for(i=0;i<numInstances;i++) {
		delete missing[i];
		delete realValues[i];
	}
	delete missing;
	delete realValues;
	delete classOfInstances;
	delete candidateCutPoints;

	for(i=0;i<numClasses;i++) {
		delete classes[i];
	}
	delete classes;	
	delete cd;
	delete classCount;
}

mantarasDiscretizer::~mantarasDiscretizer()
{
	int i;

	for(i=0;i<numAttributes;i++) {
		if(cutPoints[i]) delete cutPoints[i];
	}
	delete realAttributes;
	delete cutPoints;
}
