/************************************************************************
 * MDLP Discretizer					*
 *						 Pere Miralles Ribas	*
 ************************************************************************/

#include "mdlpDiscretizer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "messageBuffer.h"

extern messageBuffer mb;

void mdlpDiscretizer::results()
{
	int i;

	double sum=0;
	double count=0;
	for(i=0;i<numAttributes;i++) {
		if(realAttributes[i]) {
			sum+=numCutPoints[i];
			count++;
                        //mb.printf("Atribut: %d Numero de punts de tall: %d\n",i,numCutPoints[i]);
                        //for(int j=0;j<numCutPoints[i];j++)
                        //        mb.printf("[%d] : %f\n",j,(float)cutPoints[i][j]);

		}
	}
	mb.printf("Numero de punts de tall Discretitzacio : %g\n",sum/count);
}

void mdlpDiscretizer::readFile(char *datafile)
{
	char buffer[10000];
	int i,j;

	realAttributes=new bool[numAttributes];
	cutPoints = new double *[numAttributes];
	numCutPoints = new int[numAttributes];
	int attributeCount=0;
	for(i=0;i<numAttributes;i++) {
		cutPoints[i]=NULL;
	}
	FILE *fp=fopen(datafile,"r");

	fgets(buffer,9999,fp);
	while(!feof(fp) && strncasecmp(buffer,"@data",5)) {
		if(!strncasecmp(buffer,"@attribute ",10)) {
			char *ptr=buffer+10;
			if(attributeCount==numAttributes) {
				while(*ptr==' ' || *ptr=='\t') ptr++;
				while(*ptr!=' ' && *ptr!='\t') ptr++;
				while(*ptr==' ' || *ptr=='\t') ptr++;
				if(*ptr!='{') {
					fprintf(stderr,"Wrong format !!\n");
					exit(1);
				}
				ptr++;
				numClasses=0;
				for(i=0;i<strlen(ptr);i++) {
					if(ptr[i]==',' || ptr[i]=='}') numClasses++;
				}
				classes=new char *[numClasses];
				cd1=new int[numClasses];
				cd2=new int[numClasses];
				cd=new int[numClasses];
				classCount=new int[numClasses];
				for(i=0;i<numClasses;i++) {
					char tmp[1000];
					int index=0;
					while(*ptr!=',' && *ptr!='}') {
						tmp[index++]=*ptr++;
					}
					tmp[index]=0;
					classes[i]=new char[strlen(tmp)+1];
					strcpy(classes[i],tmp);
					ptr++;
				}
			} else {
				while(*ptr==' ' || *ptr=='\t') ptr++;
				while(*ptr!=' ' && *ptr!='\t') ptr++;
				while(*ptr==' ' || *ptr=='\t') ptr++;
				if(!strncasecmp(ptr,"real",4) || !strncasecmp(ptr,"integer",7)) {
					realAttributes[attributeCount]=true;
				} else {
					realAttributes[attributeCount]=false;
				}
			}
			attributeCount++;
		}
		fgets(buffer,9999,fp);
	}

	int pos=ftell(fp);
	numInstances=0;
	fgets(buffer,9999,fp);
	while(!feof(fp)) {
		numInstances++;
		fgets(buffer,9999,fp);
	}
	fseek(fp,pos,SEEK_SET);

	missing = new bool *[numInstances];
	realValues = new double *[numInstances];
	classOfInstances = new int[numInstances];
	candidateCutPoints = new int[numInstances];

	fgets(buffer,9999,fp);
	int instanceCount=0;
	while(!feof(fp)) {
		buffer[strlen(buffer)-1]=0;
		if(buffer[0] && buffer[0]!='%') {		
			missing[instanceCount] = new bool[numAttributes];
			realValues[instanceCount] = new double[numAttributes];
			char *ptr=buffer;
			for(j=0;j<numAttributes;j++) {
				char tmp[1000];
				int index=0;
				while(*ptr!=',') {
					tmp[index++]=*ptr++;
				}
				tmp[index]=0;
				if(realAttributes[j]) {
					if(tmp[0]=='?') {
						missing[instanceCount][j]=true;
					} else {
						missing[instanceCount][j]=false;
						float tempFloat;
						if(sscanf(tmp,"%f",&tempFloat)!=1) {
							fprintf(stderr,"Wrong format 2 %s!!!\n",ptr);
							exit(1);
						}	
						realValues[instanceCount][j]=tempFloat;
					}
				}
				ptr++;
			}
			bool found=false;
			for(j=0;j<numClasses && !found;j++) {
				if(!strcasecmp(classes[j],ptr)) {
					classOfInstances[instanceCount]=j;
					found=true;
				}
			}
			if(!found) {
				fprintf(stderr,"Wrong format 3 !!!\n");
				exit(1);
			}
			instanceCount++;
		}
		fgets(buffer,9999,fp);
	}

	fclose(fp);
}
	
	
void mdlpDiscretizer::doDiscretize() 
{
	int i,j;
	int points[numInstances];
	double cp[numInstances];

	for(i=0;i<numAttributes;i++) {
		if(realAttributes[i]) {
			int numPoints=0;
			for(j=0;j<numInstances;j++) {
				if(!missing[j][i]) {
					points[numPoints++]=j;
				}
			}

			if(numPoints) {
				sortValues(i,points,0,numPoints-1);
	
				int numCP=0;
				discretizeAttribute(i,points,0,numPoints-1,cp,numCP);
				numCutPoints[i]=numCP;
				//printf("Attribute %d, num cut points : %d\n",i,numCutPoints[i]);
				if(numCP>0) {
					cutPoints[i]=new double[numCP];
					for(j=0;j<numCP;j++) {
						cutPoints[i][j]=cp[j];
						//printf("Attribute %d, cut point %d : %f\n",i,j,cutPoints[i][j]);
					}
				}
			} else {
				numCutPoints[i]=0;
			}
		} else {
			numCutPoints[i]=0;
		}
	}
}

void mdlpDiscretizer::sortValues(int attribute,int *values,int begin,int end) 
{
        double pivot;
        int temp;
        int i,j;

        i=begin;j=end;
        pivot=realValues[values[(i+j)/2]][attribute];
        do {
                while(realValues[values[i]][attribute]<pivot) i++;
                while(realValues[values[j]][attribute]>pivot) j--;
                if(i<=j) {
                        if(i<j) {
                                temp=values[i];
                                values[i]=values[j];
                                values[j]=temp;
                        }
                        i++; j--;
                }
        } while(i<=j);
        if(begin<j) sortValues(attribute,values,begin,j);
        if(i<end) sortValues(attribute,values,i,end);
}

void mdlpDiscretizer::discretizeAttribute(int attribute,int *values,int begin,int end,double *cp,int &numCP) {
	int i;
        
	if(partitionPure(attribute,values,begin,end)) return;
	getCandidateCutPoints(attribute,values,begin,end);
        if(numCandidates==0) return;

        int posMin=candidateCutPoints[0];
        double entMin=computePartitionEntropy(attribute,values,begin,posMin,end);
        for(i=1;i<numCandidates;i++) {
                int pos=candidateCutPoints[i];
                double ent=computePartitionEntropy(attribute,values,begin,pos,end);
		if(ent-entMin<-0.0000001) {
                        entMin=ent;
                        posMin=pos;
                }
        }

        if(doCutPoint(attribute,values,begin,posMin,end)) {
                discretizeAttribute(attribute,values,begin,posMin-1,cp,numCP);
                double cutPoint=(realValues[values[posMin-1]][attribute]+realValues[values[posMin]][attribute])/2.0;
		cp[numCP++]=cutPoint;
                discretizeAttribute(attribute,values,posMin,end,cp,numCP);
        } 
}	
		
double mdlpDiscretizer::computePartitionEntropy(int attribute,int *values,int begin,int midPoint,int end) {
        classDistribution(attribute,values,begin,midPoint-1,cd1,numC1,numV1);
        classDistribution(attribute,values,midPoint,end,cd2,numC2,numV2);

        double ent1=computeEntropy(cd1,numC1,numV1);
        double ent2=computeEntropy(cd2,numC2,numV2);
	return ((double)numV1*ent1+(double)numV2*ent2)/(double)(numV1+numV2);
}

double mdlpDiscretizer::log2(double val)
{
	return log(val)/log(2.0);
}

int mdlpDiscretizer::doCutPoint(int attribute,int *values,int begin,int midPoint,int end) {
        classDistribution(attribute,values,begin,midPoint-1,cd1,numC1,numV1);
        classDistribution(attribute,values,midPoint,end,cd2,numC2,numV2);
        classDistribution(attribute,values,begin,end,cd,numC,numV);

        double entAll=computeEntropy(cd,numC,numV);
        double ent1=computeEntropy(cd1,numC1,numV1);
        double ent2=computeEntropy(cd2,numC2,numV2);
	double entPart=((double)numV1*ent1+(double)numV2*ent2)/(double)numV;

	double gain=entAll-entPart;
	double incPart=log2(pow((double)3,(double)numC)-2)-(numC*entAll-numC1*ent1-numC2*ent2);
	int res;
	if(gain<(log2(numV-1)+incPart)/(double)numV) {
		return 0;
	} 
	return 1;
}

double mdlpDiscretizer::computeEntropy(int *cd,int numC,int numV) {
        double ent=0;

        for(int i=0;i<numC;i++) {
                double prob=cd[i]/(double)numV;
                ent+=prob*log(prob)/log(2.0);
        }
        return -ent;
}

void mdlpDiscretizer::getCandidateCutPoints(int attribute,int *values,int begin,int end) {
	numCandidates=0;
        double valueAnt=realValues[values[begin]][attribute];

        for(int i=begin;i<=end;i++) {
                double val=realValues[values[i]][attribute];
                if(val!=valueAnt) candidateCutPoints[numCandidates++]=i;
                valueAnt=val;
        }
}

void mdlpDiscretizer::classDistribution(int attribute,int *values,int begin,int end,int *cd,int &numC,int &numV) {
	int i;

        for(i=0;i<numClasses;i++) classCount[i]=0;
        for(i=begin;i<=end;i++) classCount[classOfInstances[values[i]]]++;

	numV=numC=0;
        for(i=0;i<numClasses;i++) {
                if(classCount[i]>0) {
			cd[numC]=classCount[i];
			numV+=classCount[i];
			numC++;
		}
        }
}

bool mdlpDiscretizer::partitionPure(int attribute,int *values,int begin,int end) {
	int i;

        for(i=0;i<numClasses;i++) classCount[i]=0;
        for(i=begin;i<=end;i++) classCount[classOfInstances[values[i]]]++;

	int numC=0;
        for(i=0;i<numClasses;i++) {
                if(classCount[i]>0) {
			numC++;
		}
		if(numC>1) return false;
        }
	return true;
}


mdlpDiscretizer::mdlpDiscretizer(int attributes)
{ 
	numAttributes = attributes;                             // Nombre d'atributs
	demanaVectorDouble(minDomain,numAttributes);		// Valor mínim
	demanaVectorDouble(maxDomain,numAttributes);		// Valor màxim

}

void mdlpDiscretizer::setBounds(double *min, double *max)
{ 
	for (int i = 0; i < numAttributes; i++)
	{
		minDomain[i] = min[i];
		maxDomain[i] = max[i];
	}
}

void mdlpDiscretizer::updateDiscretizer(char *dataFile)
{
	readFile(dataFile);
	doDiscretize();
	cleanMemory();
	results();
		
}
int mdlpDiscretizer::getNumIntervals(int attr)
{
	return numCutPoints[attr]+1;
}

int mdlpDiscretizer::discretize(int attribute, double value)
{
	if (numCutPoints[attribute] == 0)
		return 0;
	for(int i=0;i<numCutPoints[attribute];i++) {
		if(value<cutPoints[attribute][i]) return i;
	}
	return numCutPoints[attribute];
}

void mdlpDiscretizer::cleanMemory()
{
	int i;

	for(i=0;i<numInstances;i++) {
		delete missing[i];
		delete realValues[i];
	}
	delete missing;
	delete realValues;
	delete classOfInstances;
	delete candidateCutPoints;

	for(i=0;i<numClasses;i++) {
		delete classes[i];
	}
	delete classes;	
	delete cd1;
	delete cd2;
	delete cd;
	delete classCount;
}

mdlpDiscretizer::~mdlpDiscretizer()
{
	int i;

	for(i=0;i<numAttributes;i++) {
		if(cutPoints[i]) delete cutPoints[i];
	}
	delete realAttributes;
	delete cutPoints;
}
