#include "populationWrapper.h"
#include <stdio.h>

populationWrapper::populationWrapper(int pPopSize)
{
	cf=new classifierFactory;
	ga=new geneticAlgorithm(cf);
	popSize = pPopSize;
}

populationWrapper::~populationWrapper()
{
	delete ga;
	delete cf;
}

void populationWrapper::activateModifiedFlag()
{
	int i;

	rank *rk=ga->getPopulationRank();
	for(i=0;i<popSize;i++) rk[i].ind->activateModified();
	ga->resetBest();
}

rank *populationWrapper::getPopulationRank()
{
	int i;

	return ga->getPopulationRank();
}


void populationWrapper::gaIteration()
{
	ga->doIterations(1);
}

void populationWrapper::releasePopulation()
{
	ga->destroyPopulation();
}

classifier *populationWrapper::getBestOverall()
{
	return (classifier *)ga->getBest();
}

classifier **populationWrapper::getPopulation()
{
	return (classifier **)ga->getPopulation();
}


classifier *populationWrapper::getBestPopulation()
{
	rank *rk=ga->getPopulationRank();
	return (classifier *)rk[0].ind;
}

classifier *populationWrapper::getWorstPopulation()
{
	rank *rk=ga->getPopulationRank();
	return (classifier *)rk[popSize - 1].ind;
}

double populationWrapper::getAverageElements()
{
	int i;
	double ave=0;

	rank *rk=ga->getPopulationRank();

	for(i=0;i<popSize;i++) 
		ave+=((classifier *)rk[i].ind)->getNumClassifiers();

	return ave/(double)popSize;
}

double populationWrapper::getAverageAliveElements()
{
	int i;
	double ave=0;

	rank *rk=ga->getPopulationRank();

	for(i=0;i<popSize;i++) 
		ave+=((classifier *)rk[i].ind)->getAliveClassifiers();

	return ave/(double)popSize;
}


void populationWrapper::getAverageDevAccuracy(double &ave,double &dev)
{
	int i;
	ave=0;
	dev=0;

	rank *rk=ga->getPopulationRank();

	for(i=0;i<popSize;i++) {
		double acc=((classifier *)rk[i].ind)->getAccuracy();
		ave+=acc;
		dev+=(acc*acc);
	}
	dev-=(ave*ave)/(double)popSize;
	dev/=(double)(popSize-1);
	dev=sqrt(dev);
	ave/=(double)popSize;
}

double populationWrapper::getMaxAccuracy()
{
	int i;

	rank *rk=ga->getPopulationRank();
	double max=((classifier *)rk[0].ind)->getAccuracy();

	for(i=1;i<popSize;i++) {
		double percen=((classifier *)rk[i].ind)->getAccuracy();
		if(percen>max) max=percen;
	}

	return max;
}

classifier *populationWrapper::cloneClassifier(classifier *orig,int son)
{
	return cf->cloneClassifier(orig,son);
}

classifier *populationWrapper::createClassifier()
{
	return cf->createClassifier();
}

void populationWrapper::destroyClassifier(classifier *orig)
{
	cf->deleteClassifier(orig);
}
