#include "timerCrossover.h"
#include "timerGlobals.h"

extern timerCrossover *tCross;
extern timerGlobals *tGlobals;

void geneticAlgorithm::replacementAlgorithm()
{
	totalReplacement();
	doFitnessComputations();
	createPopulationRank();
	if(tGlobals->elitismEnabled) doElitism();
}

void geneticAlgorithm::doElitism()
{
	int i,j;
	//char phenotype[100000];

	if(tGlobals->elitistLocalSearch) {
		populationRank[0].ind->doLocalSearch();
		populationRank[0].ind->fitnessComputation();
	}

	if(lastIteration) {
		if(flagResetBest) {
			for(i=0;i<numVersions;i++) {
				if(best[i]) {
					best[i]->fitnessComputation();
					//best[i]->dumpPhenotype(phenotype);
					//mb.printf("Phenotype of elite %d\n",i);
					//mb.printf("%s\n",phenotype);
				}
			}
		}

		int numV=numVersions;
		if(tCross->elitismUsesSmartXLastIt) {
			classifier *parents[tCross->numParentsElitismSmartX];
			int numParents=0;
			for(i=0;i<tCross->numParentsElitismSmartX;i++) {
				parents[numParents++]=populationRank[i].ind;
			}
			best[numV] = cf->createClassifier();	
			best[numV]->crossover_smart(parents,numParents);
			best[numV]->fitnessComputation();
			numV++;
		}

		JVector<int> priorities(popSize+numV);
		for(i=0;i<popSize;i++)
			priorities.addElement(populationRank[i].pos);
		for(i=0;i<numV;i++) {
			if(best[i]) {
				int size=priorities.size();
				for(j=0;j<size;j++) {
					classifier *ind;
					int pos=priorities[j];
					if(pos>=popSize) {
						ind=best[pos-popSize];
					} else {
						ind=population[pos];
					}

					if(best[i]->compareToIndividual(ind
						  ,optimizationMethod)>0) {
						priorities.insertElementAt(
							popSize+i,j);
						break;
					}
				}
				if(j==size) {
					priorities.addElement(popSize+i);
				}
			}
		}

		JVector<int> elite;
		for(i=0;i<popSize;i++) {
			if(priorities[i]>=popSize) {
				elite.addElement(priorities[i]-popSize);
				//printf("Elite %d selected\n",priorities[i]-popSize);
			}
		}
		int index=0;
		int size=priorities.size();
		for(i=popSize;i<size;i++) {
			if(priorities[i]<popSize) {
				int pos=priorities[i];
				cf->deleteClassifier(population[pos]);
				population[pos]= cf->cloneClassifier(best[elite[index++]]);
			}
		}
	} else {
		int negOffset=1;

		int previousVerUsed=0;
		int currVer=is->getCurrentVersion();
		if(!best[currVer] && currVer>0) {
			previousVerUsed=1;
			currVer--;
		}

		if(best[currVer]) {
			if(flagResetBest) {
				best[currVer]->fitnessComputation();
			}

			int pos=populationRank[popSize-negOffset++].pos;
			cf->deleteClassifier(population[pos]);
			population[pos]= cf->cloneClassifier(best[currVer]);
		}

		if(!previousVerUsed) {
			int prevVer;
			if(currVer==0) prevVer=numVersions-1;
			else prevVer=currVer-1;

			if(best[prevVer]) {
				if(flagResetBest) {
					best[prevVer]->fitnessComputation();
				}

				int pos=populationRank[popSize-negOffset++].pos;
				cf->deleteClassifier(population[pos]);
				population[pos]= cf->cloneClassifier(best[prevVer]);
			}
		}

		if(tCross->elitismUsesSmartX) {
			classifier *parents[tCross->numParentsElitismSmartX];
			int numParents=0;
			for(i=0;i<tCross->numParentsElitismSmartX;i++) {
				parents[numParents++]=populationRank[i].ind;
			}
			int pos=populationRank[popSize-negOffset++].pos;
			population[pos]->crossover_smart(parents,numParents);
			population[pos]->fitnessComputation();
		}
	}
	flagResetBest=0;
}


void geneticAlgorithm::totalReplacement()
{
	int i;

	for(i=0;i<popSize;i++) {
		cf->deleteClassifier(population[i]);
	}

	classifier **tempPop=population;
	population = offspringPopulation;
	offspringPopulation=tempPop;
}
