#include "ruleOrderAgent.h"
#include "attributesInfo.h"
#include "instanceSet.h"
#include "timerCrossover.h"
#include "messageBuffer.h"

extern messageBuffer mb;
extern instanceSet *is;
extern attributesInfo ai;
extern timerCrossover *tCross;


ruleOrderAgent::ruleOrderAgent(int pDefaultClass,unsigned long long pNumInstances,int capacity)
{
	unsigned long long i;

	defaultClass=pDefaultClass;
	numInstances=pNumInstances;
	numClasses=ai.getNumClasses();

	rules.ensureCapacity(capacity);
	profiles.ensureCapacity(capacity);

	isMatchedOK=new char[numInstances];
	ruleMatching=new int[numInstances];
	ratioOfClass = new double[numClasses];

	numRules=1;
	numOK=0;

	for(i=0;i<numClasses;i++) ratioOfClass[i]=0;

	mpaDR=new matchProfileAgent(numInstances,defaultClass);
	for(i=0;i<numInstances;i++) {
		ruleMatching[i]=0;
		register int cl=is->getInstance(i)->instanceClass;
		if(cl==defaultClass) {
			numOK++;
			isMatchedOK[i]=1;
			mpaDR->addOK(i);
		} else {
			isMatchedOK[i]=0;
			mpaDR->addKO(i);
		}
		ratioOfClass[cl]++;
	}
	mpaDR->generateProfiles();

	for(i=0;i<numClasses;i++) {
		unsigned long long num=(unsigned long long)ratioOfClass[i];
		ratioOfClass[i]*=tCross->filterSmartCrossover;
		if(ratioOfClass[i]<5) ratioOfClass[i]=5;
		if(ratioOfClass[i]>num*0.20) {
			ratioOfClass[i]=num*0.20;
		}
		//mb.printf("Ratio %d:%f\n",i,ratioOfClass[i]);
	}

	accuracy=(double)numOK/(double)numInstances;
}

ruleOrderAgent::~ruleOrderAgent()
{
	delete isMatchedOK;
	delete ruleMatching;
	delete ratioOfClass;
	delete mpaDR;
}

int ruleOrderAgent::insertRule(int whichRule,matchProfileAgent *mpa)
{
	int i;
	int bestPos;
	double bestAcc=0;

	for(i=0;i<numRules;i++) {
		double acc=accuracyOfPosition(i,mpa);
		if(i==0 || acc>bestAcc) {
			bestPos=i;
			bestAcc=acc;
		}
	}

	if(bestAcc>accuracy) {
		addRule(whichRule,mpa,bestPos);
		return bestPos;
	}
	return -1;
}

void ruleOrderAgent::addRule(int whichRule,matchProfileAgent *mpa,int whichPos) 
{
	unsigned long long i;

	if(rules.size()==0) {
		rules.addElement(whichRule);
		profiles.addElement(mpa);
	} else if(rules.size()==whichPos) {
		rules.addElement(whichRule);
		profiles.addElement(mpa);
	} else {
		rules.insertElementAt(whichRule,whichPos);
		profiles.insertElementAt(mpa,whichPos);
	}

	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]>=whichPos) {
			ruleMatching[i]++;
		
			if(mpa->mapOK[i]) {
				ruleMatching[i]=whichPos;
				if(isMatchedOK[i]==0) {
					numOK++;
					isMatchedOK[i]=1;
				}
			}

			if(mpa->mapKO[i]) {
				ruleMatching[i]=whichPos;
				if(isMatchedOK[i]==1) {
					numOK--;
					isMatchedOK[i]=0;
				}
			}
		}
	}
	
	numRules++;
	accuracy=(double)numOK/(double)numInstances;
}

void ruleOrderAgent::updateMPA(int whichPos,matchProfileAgent *mpa) 
{
	unsigned long long i,j;

	profiles[whichPos]=mpa;

	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]>=whichPos) {
			int matched=0;
			if(mpa->mapOK[i]) {
				matched=1;
				ruleMatching[i]=whichPos;
				if(isMatchedOK[i]==0) {
					numOK++;
					isMatchedOK[i]=1;
				}
			}

			if(mpa->mapKO[i]) {
				matched=1;
				ruleMatching[i]=whichPos;
				if(isMatchedOK[i]==1) {
					numOK--;
					isMatchedOK[i]=0;
				}
			}
			if(!matched && ruleMatching[i]==whichPos) {
				int corr=isMatchedOK[i]; 
				int found=0;
				for(j=whichPos+1;j<numRules-1 && !found;j++) {
					if(profiles[j]->mapOK[i]) {
						ruleMatching[i]=j;
						if(!corr) {
							isMatchedOK[i]=1;
							numOK++;
						} 
						found=1;
					} else if(profiles[j]->mapKO[i]) {
						ruleMatching[i]=j;
						if(corr) {
							isMatchedOK[i]=0;
							numOK--;
						}
						found=1;
					}
				}
				if(!found) {
					ruleMatching[i]=numRules-1;
					if(mpaDR->mapOK[i]) {
						if(!corr) {
							isMatchedOK[i]=1;
							numOK++;
						} 
					} else {
						if(corr) {
							isMatchedOK[i]=0;
							numOK--;
						}
					}
				}
			}
		}
	}
	
	accuracy=(double)numOK/(double)numInstances;
}

	

double ruleOrderAgent::accuracyOfPosition(int pos,matchProfileAgent *mpa) 
{
	unsigned long long i;
	unsigned long long newOK=numOK;

	for(i=0;i<mpa->numOK;i++) {
		unsigned long long ex=mpa->listOK[i];
		if(ruleMatching[ex]>=pos && isMatchedOK[ex]==0) {
			newOK++;
		}
	}

	for(i=0;i<mpa->numKO;i++) {
		unsigned long long ex=mpa->listKO[i];
		if(ruleMatching[ex]>=pos && isMatchedOK[ex]==1) {
			newOK--;
		}
	}

	return (double)newOK/(double)numInstances;
}

unsigned long long ruleOrderAgent::ruleNeedsInstances(int rule)
{
	unsigned long long j;
	
	unsigned long long ruleOK=0;
	for(j=0;j<numInstances;j++) {
		if(ruleMatching[j]==rule && isMatchedOK[j]==1) {
			ruleOK++;
		}
	}
	
	if(ruleOK<ratioOfClass[profiles[rule]->ruleClass]) {
		return (unsigned long long)(ratioOfClass[profiles[rule]->ruleClass]-ruleOK+0.5);	
	}
	return 0;
}

void ruleOrderAgent::refineRuleSet()
{
	unsigned long long i,j;
	
	//for(i=numRules-2;i>=0;i--) {
	i=0;
	while(i<numRules-1) {
		int ruleOK=0;
		for(j=0;j<numInstances;j++) {
			if(ruleMatching[j]==i && isMatchedOK[j]==1) {
				ruleOK++;
			}
		}
		
		if(ruleOK<ratioOfClass[profiles[i]->ruleClass]) { 
			//rules.removeElementAt(i);
			//profiles.removeElementAt(i);
			//numRules--;
		//}
			removeRule(i);
		} else {
			i++;
		}
	}
}

void ruleOrderAgent::copyOrder(JVector<int> &finalRules)
{
	int i;

	finalRules.removeAllElements();
	finalRules.ensureCapacity(rules.size());
	for(i=0;i<rules.size();i++) {
		finalRules.addElement(rules[i]);
	}
}

matchProfileAgent *ruleOrderAgent::generateActualMPA(int rule)
{
	unsigned long long i;

	matchProfileAgent *mpa = new matchProfileAgent(numInstances
		,profiles[rule]->ruleClass);

	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]==rule) {
			if(isMatchedOK[i]) {
				mpa->addOK(i);
			} else {
				mpa->addKO(i);
			}
		}
	}
	mpa->generateProfiles();

	return mpa;
}

void ruleOrderAgent::removeMatchesOfRule(int rule,matchProfileAgent *mpa)
{
	unsigned long long i,j;

	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]==rule) {
			if(!mpa->mapOK[i] && !mpa->mapKO[i]) {
				int corr=isMatchedOK[i]; 
				int found=0;
				for(j=rule+1;j<numRules-1 && !found;j++) {
					if(profiles[j]->mapOK[i]) {
						ruleMatching[i]=j;
						if(!corr) {
							isMatchedOK[i]=1;
							numOK++;
						} 
						found=1;
					} else if(profiles[j]->mapKO[i]) {
						ruleMatching[i]=j;
						if(corr) {
							isMatchedOK[i]=0;
							numOK--;
						}
						found=1;
					}
				}
				if(!found) {
					ruleMatching[i]=numRules-1;
					if(mpaDR->mapOK[i]) {
						if(!corr) {
							isMatchedOK[i]=1;
							numOK++;
						} 
					} else {
						if(corr) {
							isMatchedOK[i]=0;
							numOK--;
						}
					}
				}
			}
		}
	}

	accuracy=(double)numOK/(double)numInstances;
} 

void ruleOrderAgent::removeRule(int rule)
{
	unsigned long long i,j;

	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]==rule) {
			int corr=isMatchedOK[i]; 
			int found=0;
			for(j=rule+1;j<numRules-1 && !found;j++) {
				if(profiles[j]->mapOK[i]) {
					ruleMatching[i]=j;
					if(!corr) {
						isMatchedOK[i]=1;
						numOK++;
					} 
					found=1;
				} else if(profiles[j]->mapKO[i]) {
					ruleMatching[i]=j;
					if(corr) {
						isMatchedOK[i]=0;
						numOK--;
					}
					found=1;
				}
			}
			if(!found) {
				ruleMatching[i]=numRules-1;
				if(mpaDR->mapOK[i]) {
					if(!corr) {
						isMatchedOK[i]=1;
						numOK++;
					} 
				} else {
					if(corr) {
						isMatchedOK[i]=0;
						numOK--;
					}
				}
			}
		}

		if(ruleMatching[i]>rule) {
			ruleMatching[i]--;
		}
	}

	rules.removeElementAt(rule);
	profiles.removeElementAt(rule);
	numRules--;
	accuracy=(double)numOK/(double)numInstances;
} 

void ruleOrderAgent::getPossibleInstancesOfRule(int rule,unsigned long long *instances,unsigned long long &numInst)
{
	unsigned long long i;

	numInst=0;
	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]>rule) {
			instances[numInst++]=i;
		}
	}
}

void ruleOrderAgent::getNegativeExamples(int rule,unsigned long long *instances,unsigned long long &numInst)
{
	unsigned long long i;

	for(i=0;i<numInstances;i++) {
		if(ruleMatching[i]==rule && isMatchedOK[i]==0) {
			instances[numInst++]=i;
		}
	}
}


void ruleOrderAgent::addPositiveExamples(int rule,unsigned long long *instances,unsigned long long numInst,
	matchProfileAgent *mpa)
{
	int i;

	profiles[rule]=mpa;

	for(i=0;i<numInst;i++) {
		unsigned long long inst=instances[i];
		ruleMatching[inst]=rule;
		if(isMatchedOK[inst]==0) {
			numOK++;
			isMatchedOK[inst]=1;
		}
	}

	accuracy=(double)numOK/(double)numInstances;
} 

