#include <stdlib.h>
#include <stdio.h>
#include "random.h"
#include "timerEvolutionStats.h"
#include "timerPareto.h"
#include "instanceSet.h"

extern timerEvolutionStats *tEvolStats;
extern timerPareto *tPareto;
extern instanceSet *is;

void geneticAlgorithm::selectionAlgorithm()
{
	scalingAlgorithm();

	switch (selectionAlg) {
	case PARETO_SELECTION:
		ParetoSelection();
		break;
	case TOURNAMENT_WOR_SELECTION:
		TournamentSelectionWOR();
		break;
	case TOURNAMENT_SELECTION:
	default:
		TournamentSelection();
		break;
	}

	for (int i = 0; i < popSize; i++) cf->deleteClassifier(population[i]);
	classifier **tempPop=population;
	population = offspringPopulation;
	offspringPopulation=tempPop;
}

int geneticAlgorithm::selectNicheWOR(int *quotas,int num)
{
	if(num==1) return 0;

	int total=0,i;
	for(i=0;i<num;i++) total+=quotas[i];
	if(total==0) {
		return rnd(0,num-1);
	}
	int pos=rnd(0,total-1);
	total=0;
	for(i=0;i<num;i++) {
		total+=quotas[i];
		if(pos<total) {
			quotas[i]--;
			return i;
		}
	}

	fprintf(stderr,"We should not be here\n");
	exit(1);
	return -1;
}

void geneticAlgorithm::initPool(JVector<int> &pool,int whichNiche)
{
	int i;

	if(tEvolStats->nichingEnabled) {
		for(i=0;i<popSize;i++) {
			if(population[i]->getNiche()==whichNiche) {
				pool.addElement(i);
			}
		}
	} else {
		for(i=0;i<popSize;i++) pool.addElement(i);
	}
}

int geneticAlgorithm::selectCandidateWOR(JVector<int> &pool,int whichNiche)
{
	if(pool.size()==0) {
		initPool(pool,whichNiche);
		if(pool.size()==0) {
			return rnd(0,popSize-1);
		}
	}

	int pos=rnd(0,pool.size()-1);
	int elem=pool[pos];
	pool.removeElementAt(pos);
	return elem;
}

void geneticAlgorithm::TournamentSelectionWOR(void)
{
	int i, j, winner, candidate;

	int numNiches;
	if(tEvolStats->nichingEnabled)
		numNiches=population[0]->getNumNiches();
	else
		numNiches=1;

	JVector<int> pools[numNiches];
	int nicheCounters[numNiches];
	int nicheQuota=popSize/numNiches;
	for (i = 0; i < numNiches; i++) nicheCounters[i]=nicheQuota;

	for (i = 0; i < popSize; i++) {
		//There can be only one
		int niche=selectNicheWOR(nicheCounters,numNiches);
		winner=selectCandidateWOR(pools[niche],niche);
		for (j = 1; j < tournamentSize; j++) {
			candidate=selectCandidateWOR(pools[niche],niche);
			if(population[candidate]->
				compareToIndividual(population[winner],
				optimizationMethod)>0) {
				winner = candidate;
			}
		}
		offspringPopulation[i]=cf->cloneClassifier(population[winner]);
	}
}




int geneticAlgorithm::selectNiche(int *quotas,int num)
{
	if(!tEvolStats->nichingEnabled) return -1;

	int total=0,i;
	for(i=0;i<num;i++) total+=quotas[i];
	if(total==0) return -1;
	int pos=rnd(0,total-1);
	total=0;
	for(i=0;i<num;i++) {
		total+=quotas[i];
		if(pos<total) {
			quotas[i]--;
			return i;
		}
	}

	fprintf(stderr,"We should not be here\n");
	exit(1);
	return -1;
}

int geneticAlgorithm::selectCandidate(int niche)
{
	int countTrial=0;
	int candidate;
	do {
		candidate = rnd (0, popSize - 1);
		countTrial++;
	} while(niche!=-1 && population[candidate]->getNiche()!=niche
		&& countTrial<25);
	return candidate;
}


void geneticAlgorithm::TournamentSelection(void)
{
	int i, j, winner, candidate;

	int numNiches=population[0]->getNumNiches();
	int nicheCounters[numNiches];
	int nicheQuota=popSize/numNiches;
	for (i = 0; i < numNiches; i++) nicheCounters[i]=nicheQuota;

	for (i = 0; i < popSize; i++) {
		//There can be only one
		int niche=selectNiche(nicheCounters,numNiches);
		winner=selectCandidate(niche);
		for (j = 1; j < tournamentSize; j++) {
			candidate=selectCandidate(niche);
			if(population[candidate]->
				compareToIndividual(population[winner],
					optimizationMethod)>0) {
				winner = candidate;
			}
		}
		offspringPopulation[i]=cf->cloneClassifier(population[winner]);
	}
}

double geneticAlgorithm::ParetoOrder(double *a, double *b)
{
	if (a[1] < b[1])
		return -69;
	if (a[1] > b[1])
		return +69;
	if (a[2] < b[2])
		return -69;
	if (a[2] > b[2])
		return +69;
	return 0;
}

void geneticAlgorithm::ParetoSwap(double *a, double *b)
{
	int i;
	double tmp;

	for (i = 0; i <= 2; i++) {
		tmp = a[i];
		a[i] = b[i];
		b[i] = tmp;
	}
}

void geneticAlgorithm::ParetoQSort(double **objectives, int left, int right)
{
	double *y = new double[3];
	int i, j, k;

	i = left;
	j = right;
	for (k = 0; k < 3; k++) y[k] = objectives[(i + j) / 2][k];

	do {
		while (ParetoOrder(objectives[i], y) < 0) i++;
		while (ParetoOrder(objectives[j], y) > 0) j--;
		if (i <= j) {
			ParetoSwap(objectives[i], objectives[j]);
			i++;
			j--;
		}
	} while (i <= j);
	delete y;

	if (left < j) ParetoQSort(objectives, left, j);
	if (i < right) ParetoQSort(objectives, i, right);
}

double geneticAlgorithm::ParetoDistance(double *a, double *b)
{
	double dist = 0;
	int i;
	for (i = 1; i < 3; i++) {
		dist += (a[i] - b[i]) * (a[i] - b[i]);
	}
	return sqrt(dist);
}

double geneticAlgorithm::ParetoSharing(double val)
{
	double res;

	res = 1.0 - (1.0 / 0.1) * val;
	if (res < 0)
		res = 0;

	return res;
}

void geneticAlgorithm::ParetoElitistTournament(double *paretoFitness)
{
	int i, j, winner, candidate;
	int windowingEnabled=is->isWindowingEnabled();

	for (i = 0; i < oldBests.size(); i++) {
		offspringPopulation[i] = oldBests[i];
		if(windowingEnabled) offspringPopulation[i]->activateModified();
	}

	for (j = 0; j < oldFront.size() && i < popSize; j++, i++) {
		offspringPopulation[i] = oldFront[j];
		if(windowingEnabled) offspringPopulation[i]->activateModified();
	}

	for (; i < popSize; i++) {
		//There can be only one
		winner = rnd (0, popSize - 1);
		for (j = 1; j < tournamentSize; j++) {
			candidate = rnd (0, popSize - 1);
			if (paretoFitness[candidate] > paretoFitness[winner]) {
				winner = candidate;
			}
		}

		offspringPopulation[i]=cf->cloneClassifier(population[winner]);
	}
}

double *geneticAlgorithm::ParetoFitness(double **objectives,JVector<int>&elitismGap,JVector<int>&paretoFront,int step)
{
	double *paretoFitness = new double[popSize];
	int i, j, k;
	JVector < JVector < int > >fronts;
	int numFronts;
	double delta=1000;
	double percentageOfElitism = 0.3;

	ParetoQSort(objectives, 0, popSize - 1);

	for(i=0;i<percentageOfElitism*popSize;i++) 
		elitismGap.addElement((int)objectives[i][0]);

	JVector < int >individuals;
	for (i = 0; i < popSize; i++)
		individuals.addElement(i);

	numFronts = 0;
	while (individuals.size() > 0) {
		if (showFronts && step==1) {
			mb.printf("Front %d: ", numFronts);
		}
		JVector < int >front;
		int pos = 0;
		double valueSecondObj;
		while (pos < individuals.size()) {
			int sel = 0;
			int ind = individuals[pos];
			if (front.size() > 0) {
				if (objectives[ind][2] < valueSecondObj) {
					sel = 1;
					valueSecondObj = objectives[ind][2];
				}
			} else {
				sel = 1;
				valueSecondObj = objectives[ind][2];
			}

			if (sel) {
				if (showFronts && step==1) {
					mb.printf("(");
					for (i = 1; i < 3; i++) {
						mb.printf("%f",objectives[ind][i]);
						if (i < 2) mb.printf(",");
					}
					mb.printf(") ");
				}
				front.addElement(ind);
				if (numFronts == 0) 
					paretoFront.addElement(
						(int)objectives[ind][0]);
				individuals.removeElementAt(pos);

				while (pos < individuals.size() && 
					!ParetoOrder(objectives[ind] , objectives[individuals [pos]])) {
					front.addElement(individuals[pos]);
					individuals.removeElementAt(pos);
				}
			} else {
				pos++;
			}
		}
		fronts.addElement(front);
		numFronts++;
		if (showFronts && step==1)
			mb.printf("\n");
	}

	if(step==1) mb.printf("There are %d fronts\n", numFronts);

	for (i = 0; i < fronts.size(); i++) {
		for (j = 0; j < fronts[i].size(); j++) {
			double sharing = 0;
			for (k = 0; k < popSize; k++) {
				sharing +=
				    ParetoSharing(ParetoDistance
						  (objectives[fronts[i][j]]
						   ,
						   objectives[k]));
			}
			paretoFitness[(int) objectives[fronts[i][j]][0]] =
			    delta * (fronts.size() - i - 1) + delta / sharing;
			if(step==1) {
				population[(int) objectives[fronts[i][j]][0]]->setFront(i);
			}
		}
	}

	return paretoFitness;
}

void geneticAlgorithm::ScaleObjectives(double **objectives,int num)
{
	int i;
	double obj1,obj2;

	//obj1=objectives[0][1];
	obj2=objectives[0][2];
	for(i=1;i<num;i++) {
		//if(objectives[i][1]>obj1) obj1=objectives[i][1];
		if(objectives[i][2]>obj2) obj2=objectives[i][2];
	}
	for(i=0;i<num;i++) {
		//objectives[i][1]/=obj1;
		objectives[i][2]/=obj2;
	}
}

void geneticAlgorithm::ParetoSelection(void)
{
	int i,j;
	JVector < int >paretoFront;
	JVector < int >elitismGap;

	if(!tPareto->activated) {
		TournamentSelection();
		return;
	}
	
	double **objectives = new double *[popSize];

	for (i = 0; i < popSize; i++) {
		double *objectivesInd = new double[3];
		objectivesInd[0] = i;
		for (j = 0; j < 2; j++) {
			objectivesInd[j + 1] = population[i]->getObjective(j);
		}
		objectives[i] = objectivesInd;
	}
	ScaleObjectives(objectives,popSize);

	double *paretoFitness=ParetoFitness(objectives,elitismGap,paretoFront,1);
	if(oldBests.size()==0) {
		//First time here
		for(i=0;i<elitismGap.size();i++) 
			oldBests.addElement(cf->cloneClassifier(population[elitismGap[i]]));
		for(i=0;i<paretoFront.size();i++) 
			oldFront.addElement(cf->cloneClassifier(population[paretoFront[i]]));
	}
	
	ParetoElitistTournament(paretoFitness);

	for (i = 0; i < popSize; i++) {
		double *objectivesInd = objectives[i];
		objectivesInd[0] = i;
		for (j = 0; j < 2; j++) {
			objectivesInd[j + 1] = offspringPopulation[i]->getObjective(j);
		}
		objectives[i] = objectivesInd;
	}
	ScaleObjectives(objectives,popSize);

	paretoFront.removeAllElements();
	elitismGap.removeAllElements();
	delete paretoFitness;
	paretoFitness=ParetoFitness(objectives,elitismGap,paretoFront,2);

	oldBests.removeAllElements();
	oldFront.removeAllElements();

	for(i=0;i<elitismGap.size();i++) 
		oldBests.addElement(cf->cloneClassifier(offspringPopulation[elitismGap[i]]));
	for(i=0;i<paretoFront.size();i++) 
		oldFront.addElement(cf->cloneClassifier(offspringPopulation[paretoFront[i]]));

	for (i = 0; i < popSize; i++)
		delete objectives[i];
	delete objectives;
	delete paretoFitness;
}
