#include "classifier_adaptive.h"
#include "timerADI.h"
#include "timerGlobals.h"
#include "discretizationManager.h"
#include "attributesInfo.h"
#include "random.h"
#include "populationWrapper.h"
#include "messageBuffer.h"

extern messageBuffer mb;

extern discretizationManager disc;
extern attributesInfo ai;
extern Random rnd;
extern timerGlobals *tGlobals;

timerADI::timerADI()
{
	int i;

	if (!cm.thereIsParameter(KR_ADI)) {
		enabled = 0;
		return;
	}
	enabled = 1;

	numAttributes = ai.getNumAttributesMC();
	numDiscretizers = disc.getNumDiscretizers();
	probSplitInterval = cm.getParameter(PROB_SPLIT);
	probMergeInterval = cm.getParameter(PROB_MERGE);
	int numIterations = (int)cm.getParameter(ITERATIONS);

	if (cm.thereIsParameter(PROB_REINITIALIZE)) {
		reinitializeEnabled=1;
		double pStart = cm.getParameter(PROB_REINITIALIZE);
		double pEnd;
		if (cm.thereIsParameter(PROB_REINITIALIZE_AT_END)) {
			pEnd = cm.getParameter(PROB_REINITIALIZE_AT_END);
		} else {
			pEnd = pStart;
		}
		probReinitManager = new probabilityManagement(pStart, pEnd, 
			LINEAR);
	} else {
		probReinitializeAttribute = 0;
		reinitializeEnabled = 0;
	}

	if(cm.thereIsParameter(PROB_ONE)) {
		tGlobals->probOne = cm.getParameter(PROB_ONE);
	} else {
		double num=ai.getNumAttributesMC();
		double minR=tGlobals->minClassifiersInit;
		double nc=ai.getNumClasses();
		tGlobals->probOne=pow(1-pow(nc,-1.0/minR),1.0/num);
		if(tGlobals->probOne<0.90) tGlobals->probOne=0.90;
		mb.printf("Probability of ONE set to %f\n",tGlobals->probOne);
	}
	numMaxIntervals = (int) cm.getParameter(MAX_INTERVALS);
	if (numDiscretizers == 0) {
		fprintf(stderr,"No s'ha definit cap tipus de discretitzaci\n");
		exit(1);
	}

	int sum = 2;
	sizeAdaptiveAttribute = new int[numAttributes];
	offsetAdaptiveAttribute = new int[numAttributes];
	typeOfAttribute = new int[numAttributes];
	defaultRuleTheoryLength=0.00000001;
	//defaultRuleTheoryLength=0.001;
	for (i = 0; i < numAttributes; i++) {
		if (ai.getTypeOfAttribute(i) == REAL) {
			typeOfAttribute[i]=REAL;
			sizeAdaptiveAttribute[i] = numMaxIntervals * 2 + 2;
		} else {
			typeOfAttribute[i]=NOMINAL;
			sizeAdaptiveAttribute[i] =
			    ai.getNumValuesAttribute(i) + 2;
		}
		offsetAdaptiveAttribute[i] = sum;
		sum += sizeAdaptiveAttribute[i];
	}
	ruleSize = sum;

	popSize=(int)cm.getParameter(POP_SIZE);
	totalIterations=(int)cm.getParameter(ITERATIONS);

	stageOrder=new int[3];

	if(cm.thereIsParameter(DUMP_DISCRETIZERS)) {
		dumpDiscretizersEnabled=1;
	} else {
		dumpDiscretizersEnabled=0;
	}
}

void timerADI::newIteration(int iteration,int lastIteration)
{
	if (!enabled)
		return;

	numSplit = numMerge = numReinitialize = numMergeFailed = 
		numMergeAttempts = 0;

	stageOrder[0]=MERGE;
	stageOrder[1]=SPLIT;
	stageOrder[2]=REINITIALIZE;
		
	if(reinitializeEnabled) {
		probReinitializeAttribute=probReinitManager->incStep();
		//mb.printf("Probability of attribute reinitialize:%f\n"
		//	,probReinitializeAttribute);
	} 
}


void timerADI::dumpStats(int iteration)
{
	if (!enabled)
		return;

	classifier_adaptive *best=
		(classifier_adaptive *)pw->getBestPopulation();
	double intervalsPerAttribute = best->getIntervalsPerAttribute();
	//mb.printf("Iteration %d,Split:%d Merge:%d MergeFailed:%d MergeAttempts:%d"
	//	" Reinitialize:%d IntervalsPerAtr:%f\n",iteration, numSplit, 
	//	numMerge, numMergeFailed, numMergeAttempts, numReinitialize, 
	//	intervalsPerAttribute);

	if(dumpDiscretizersEnabled) dumpDiscretizersInPopulation(iteration);
}

void timerADI::dumpDiscretizersInPopulation(int iteration)
{
	int i;
	int num=disc.getNumDiscretizers();
	int counters[num];
	classifier **population=pw->getPopulation();
	classifier *best=pw->getBestPopulation();

	int count=0;
	for(i=0;i<num;i++) counters[i]=0;
	for(i=0;i<popSize;i++) count += ((classifier_adaptive *)
		population[i])->countDiscretizers(counters);
	mb.printf("Iteration %d,Proportions of population: ",iteration);
	for(i=0;i<num;i++) mb.printf("%f ",(double)counters[i]/(double)count);
	mb.printf("\n");

	count=0;
	for(i=0;i<num;i++) counters[i]=0;
	count = ((classifier_adaptive *)best)->countDiscretizers(counters);
	mb.printf("Iteration %d,Proportions of best: ",iteration);
	for(i=0;i<num;i++) mb.printf("%f ",(double)counters[i]/(double)count);
	mb.printf("\n");
}


