#include "timerCrossover.h"
#include <math.h>
#include "messageBuffer.h"
#include "attributesInfo.h"

extern messageBuffer mb;
extern attributesInfo ai;

timerCrossover::timerCrossover()
{
	int i;

	crossoverProb = cm.getParameter(PROB_CROSSOVER);
	checkCrossover = cm.thereIsParameter(CHECK_CROSSOVER);
	if(cm.thereIsParameter(PROB_SMART_CROSSOVER)) {
		smartCrossoverProb=cm.getParameter(PROB_SMART_CROSSOVER);
		numParentsSmartX=
			(int)cm.getParameter(NUM_PARENTS_SMART_CROSSOVER);
		repetitionsRuleOrdering = 
			(int)cm.getParameter(REPETITIONS_RULE_ORDERING);
		addRulesSmartX = cm.thereIsParameter(ADD_RULES_SMART_CROSSOVER);
		filterSmartCrossover = 
			cm.getParameter(FILTER_SMART_CROSSOVER);
		rulePruning=cm.thereIsParameter(SMART_CROSSOVER_RULE_PRUNING);
	} else {
		smartCrossoverProb=0;
	}

	elitismUsesSmartX=cm.thereIsParameter(ELITISM_WITH_SMART_CROSSOVER);
	if(elitismUsesSmartX) {
		elitismUsesSmartXLastIt=cm.thereIsParameter(ELITISM_LAST_ITERATION_WITH_SMART_CROSSOVER);
		repetitionsRuleOrdering = 
			(int)cm.getParameter(REPETITIONS_RULE_ORDERING);
		filterSmartCrossover = 
			cm.getParameter(FILTER_SMART_CROSSOVER);
		numParentsElitismSmartX=
			(int)cm.getParameter(NUM_PARENTS_ELITISM_SMART_CROSSOVER);
		rulePruning=cm.thereIsParameter(SMART_CROSSOVER_RULE_PRUNING);
	} else {
		elitismUsesSmartXLastIt=0;
	}

	informedCrossover=cm.thereIsParameter(INFORMED_CROSSOVER);
	if(informedCrossover) {
		int numAtt=ai.getNumAttributes();
		int attMap[numAtt];

		for(i=0;i<numAtt;i++) attMap[i]=0;

		char string[1000];
		JVector<int>sizes;
		JVector<int *>BBs;
		FILE *fp=fopen("cutPoints.dat","r");
		if(fp==NULL) {
			fprintf(stderr,"Cannot open cutPoints.dat\n");
			exit(1);
		}
		fgets(string,999,fp);
		while(!feof(fp)) {
			char *token;
        		token=strtok(string," ");
			JVector<int> varsBB;
			while(token!=NULL) {
				int value=atoi(token);
				if(value<0 || value>=numAtt) {
					fprintf(stderr,"Attribute %d is out of ranges\n",value);
					exit(1);
				}
				if(attMap[value]==1) {
					fprintf(stderr,"Attribute %d has already been used\n",value);
					exit(1);
				}
				attMap[value]=1;
				varsBB.addElement(value);
        			token=strtok(NULL," ");
			}
	
			int size=varsBB.size();
			int *bb=new int[size];	
			for(i=0;i<size;i++) {
				bb[i]=varsBB[i];
			}
	
			sizes.addElement(size);
			BBs.addElement(bb);
			fgets(string,999,fp);
		}

		numBB=sizes.size();
		sizeBBs=new int[numBB];
		defBBs = new int *[numBB];
		for(i=0;i<numBB;i++) {
			sizeBBs[i]=sizes[i];
			defBBs[i]=BBs[i];
		}

		fclose(fp);

		for(i=0;i<numAtt;i++) {
			if(attMap[i]==0) {
				fprintf(stderr,"Attribute %d is not included in any BB\n",i);
				exit(1);
			}
		}
	}
		
}

void timerCrossover::newIteration(int iteration,int lastIteration)
{
	parentsWin=0;
	offspringWin=0;
}

void timerCrossover::dumpStats(int iteration)
{
	if(checkCrossover) {
		mb.printf("Iteration %d,Parents:%d Offpring:%d\n",iteration
			,parentsWin,offspringWin);
	}
}
