#include "populationWrapper.h"
#include "timerEvolutionStats.h"
#include "timerGlobals.h"
#include "messageBuffer.h"
#include "utils.h"

extern int nodeRank;
extern int numTasks;
extern messageBuffer mb;
extern timerGlobals *tGlobals;
extern double percentageOfLearning;

timerEvolutionStats::timerEvolutionStats()
{
	int i;

	iterationsSinceBest = 0;
	globalIterationsSinceBest = 0;
	maxMin = (int) cm.getParameter(MAX_MIN);
	popSize=(int)cm.getParameter(POP_SIZE);
	nichingEnabled=0;
	numExampleEvaluations=0;

	doDumpStats=cm.thereIsParameter(DUMP_EVOLUTION_STATS);

	if(tGlobals->defaultClassPolicy==AUTO) {
		hardNichingDisable=cm.thereIsParameter(HARD_NICHING_DISABLE);
		numNiches=ai.getNumClasses();
		accDefaultRules=new JVector<double>[numNiches];
		nichingEnabled=1;
	}
}

timerEvolutionStats::~timerEvolutionStats()
{
	//if(tGlobals->defaultClassPolicy==AUTO) {
	//	delete accDefaultRules;
	//}
}

int timerEvolutionStats::getGlobalIterationsSinceBest()
{
	return globalIterationsSinceBest;
}
int timerEvolutionStats::getIterationsSinceBest()
{
	return iterationsSinceBest;
}

void timerEvolutionStats::resetBestStats()
{
	iterationsSinceBest=0;
}


void timerEvolutionStats::bestOfIteration(int iteration,
					  double iterationBestFitness,
					  double iterationBestAcc)
{
	if (!iterationsSinceBest) {
		globalIterationsSinceBest++;
		bestFitness = iterationBestFitness;
		iterationsSinceBest = 1;
	} else {
		int newBest = 0;
		if (maxMin == MAXIMIZE) {
			if (iterationBestFitness > bestFitness)
				newBest = 1;
		} else {
			if (iterationBestFitness < bestFitness)
				newBest = 1;
		}

		if (newBest) {
			bestFitness = iterationBestFitness;
			iterationsSinceBest = 1;
			globalIterationsSinceBest = 1;
			//mb.printf("Iteration %d, New best fitness\n",
			//       iteration);
		} else {
			iterationsSinceBest++;
			globalIterationsSinceBest++;
		}
	}
}

void timerEvolutionStats::dumpStats(int iteration)
{
	classifier *ind = pw->getBestOverall();
	double aveClassifiers = pw->getAverageElements();
	double aveAliveClassifiers = pw->getAverageAliveElements();
	double aveAcc,devAcc;
	pw->getAverageDevAccuracy(aveAcc,devAcc);

	if(ind->getNumNiches()==1) nichingEnabled=0;

	if(nichingEnabled) {
		int i;
		int counters[numNiches];
		double nicheFit[numNiches];
		for(i=0;i<numNiches;i++) {
			counters[i]=0;
			nicheFit[i]=0;
		}
		classifier **pop=pw->getPopulation();
		for(i=0;i<popSize;i++) {
			int niche=pop[i]->getNiche();
			counters[niche]++;
			double indAcc=pop[i]->getAccuracy();
			if(indAcc>nicheFit[niche]) nicheFit[niche]=indAcc;
		}
		//for(i=0;i<numNiches;i++) {
		//	mb.printf("%d(%f) ",counters[i],nicheFit[i]);
		//}
		//mb.printf("\n");
		if(accDefaultRules[0].size()==15) {
			for(i=0;i<numNiches;i++) accDefaultRules[i].removeElementAt(0);
		}
		for(i=0;i<numNiches;i++) accDefaultRules[i].addElement(nicheFit[i]);
		if(accDefaultRules[0].size()==15) {
			JVector<double> aves;
			for(i=0;i<numNiches;i++) {
				double aveN=getAverage(accDefaultRules[i]);
				aves.addElement(aveN);
				//mb.printf("%f ",aveN);
			}
			//mb.printf("\n");
			double dev=getDeviation(aves);
			//mb.printf("DefRules dev:%f\n",dev);
			if(dev<0.005 || (hardNichingDisable 
				&& percentageOfLearning>0.25)) {
				mb.printf("Iteration %d,niching disabled\n"
					,iteration);
				nichingEnabled=0;
			}
		}
	}

	double numExamEval=numExampleEvaluations;
	if(doDumpStats || iteration==0) {
		mb.printf("It %d,Best ac:%f fi:%f #cl:%d(%d)."
		       " Ave ac:%f,%f #cl:%f(%f)\n", iteration,
		       ind->getAccuracy(), ind->getFitness(),
		       ind->getNumClassifiers(), ind->getAliveClassifiers(),
		       aveAcc,devAcc, aveClassifiers,aveAliveClassifiers);
		//mb.printf("Evaluations %f, Acc %f\n",numExamEval
		//	,ind->getAccuracy());
	}

	bestOfIteration(iteration, ind->getFitness(), ind->getAccuracy());
}
