#include "populationWrapper.h"
#include "timerMDL.h"
#include "attributesInfo.h"
#include "timerEvolutionStats.h"
#include "timerPruning.h"
#include "instanceSet.h"
#include "utils.h"
#include "messageBuffer.h"

extern messageBuffer mb;

extern timerEvolutionStats *tEvolStats;
extern timerPruning *tPruning;
extern attributesInfo ai;
extern instanceSet *is;

timerMDL::timerMDL()
{
	if (!(cm.getParameter(FITNESS_FUNCTION) == MDL)) {
		mdlAccuracy = 0;
		return;
	}
	mdlAccuracy = 1;

	startIteration = (int) cm.getParameter(MDL_ITERATION);
	if (!startIteration)
		startIteration++;

	mdlWeightRelaxFactor = cm.getParameter(MDL_WEIGHT_RELAX_FACTOR);

	fixedWeight = 0;
	initialTheoryLenghtRatio = cm.getParameter(MDL_INITIAL_TL_RATIO);
	activated = 0;
	iterationMDL = 0;

	if(cm.thereIsParameter(MDL_WEIGHT)) {
		fixedWeight=1;
		mdlWeight=cm.getParameter(MDL_WEIGHT);
	}
}

void timerMDL::newIteration(int iteration, int lastIteration)
{
	if (!mdlAccuracy)
		return;
	int updateWeight = 0;
	iterationMDL++;
	currentIteration = iteration;

	if (iteration == startIteration) {
		mb.printf("Iteration %d:MDL fitness activated\n",
			  iteration);
		activated = 1;
		if(!fixedWeight) {
			classifier *ind1 = pw->getBestPopulation();
			double error = ind1->getExceptionsLength();
			double theoryLength = ind1->getTheoryLength();
			theoryLength *= (double) ai.getNumClasses();
			theoryLength /= (double) ind1->getAliveClassifiers();
			mdlWeight =
			    (initialTheoryLenghtRatio /
			     (1 - initialTheoryLenghtRatio))
			    * (error / theoryLength);
		}
		updateWeight = 1;
	}

	if (activated && !fixedWeight) {
		if (pw->getBestPopulation()->getAccuracy() != 1.0) {
			if (tEvolStats->getIterationsSinceBest()
			    == 10) {
				mdlWeight *= mdlWeightRelaxFactor;
				updateWeight = 1;
			}
		}
	}

	if (updateWeight) {
		tEvolStats->resetBestStats();
		mb.printf("MDL Theory Length Weight: %.10f (%d)\n", mdlWeight,tEvolStats->getGlobalIterationsSinceBest());
		pw->activateModifiedFlag();
	}
}

void timerMDL::dumpStats(int iteration)
{
	if (mdlAccuracy && activated) {
		//classifier *ind1 = pw->getBestPopulation();
		//mb.printf("Iteration %d,MDL Stats: %f %f %f\n", iteration,
		//       ind1->getTheoryLength() * mdlWeight,
		//       ind1->getExceptionsLength(),
		//       ind1->getTheoryLength() * mdlWeight /
		//       ind1->getFitness());
	}
}

double timerMDL::mdlFitness(classifier & ind, agentPerformanceTraining * ap)
{
	double mdlFitness = 0;
	if (activated) {
		mdlFitness = ind.getTheoryLength() * mdlWeight;
		if (mdlFitness == 0 && ap->getNC() != 1.0) {
			fprintf(stderr, "TL:%f\n",
				ind.computateTheoryLength(ap));
			exit(1);
		}
	}

	double acc = ap->getAccuracy();
	double exceptionsLength = 105.0 - acc * 100.0;
	ind.setExceptionsLength(exceptionsLength);
	mdlFitness += exceptionsLength;

	return mdlFitness;
}
