#include "timerPruning.h"
#include "populationWrapper.h"
#include "random.h"
#include "messageBuffer.h"

extern messageBuffer mb;

extern Random rnd;

timerPruning::timerPruning()
{
	if(!cm.thereIsParameter(PRUNING_ITERATION)) {
		enabled = 0;
		return;
	}
	enabled=1;

	if(cm.thereIsParameter(PRUNING_POLICY)) {
		policy=(int)cm.getParameter(PRUNING_POLICY);
	} else {
		policy=RANDOM;
	}

	minClassifiers = (int) cm.getParameter(PRUNING_MIN_CLASSIFIERS);
	startIteration = (int) cm.getParameter(PRUNING_ITERATION);
	if(!startIteration) pruningActivated = 1;
	else pruningActivated = 0;

	autoThreshold=cm.thereIsParameter(PRUNING_AUTO_THRESHOLD);
	autoThreshold2=cm.thereIsParameter(PRUNING_AUTO_THRESHOLD2);

	if(autoThreshold || autoThreshold2) {
		autoOffset= (int) cm.getParameter(PRUNING_AUTO_OFFSET);
	}
}

void timerPruning::newIteration(int iteration,int finalIteration)
{
	if(!enabled) return;

	if (iteration == startIteration) {
		pruningActivated = 1;
		pw->activateModifiedFlag();
		mb.printf("Iteration %d:Classifier pruning activated\n",
		       iteration);
	}

	if(autoThreshold) {
		minClassifiers=pw->getBestPopulation()->getAliveClassifiers()+autoOffset;
		if(minClassifiers<2) minClassifiers=2;
		mb.printf("Iteration %d:Pruning min classifiers %d\n",
		       iteration,minClassifiers);
		
	}
	if(autoThreshold2) {
		minClassifiers=(int)round(pw->getAverageAliveElements())+autoOffset;
		if(minClassifiers<2) minClassifiers=2;
		mb.printf("Iteration %d:Pruning min classifiers %d\n",
		       iteration,minClassifiers);
	}

	if (finalIteration) {
		minClassifiers = 1;
	}
}

int compare(const void *a,const void *b) {
	int val1=*(int *)a;
	int val2=*(int *)b;

	if(val1>val2) return +69;
	if(val1<val2) return -69;
	return 0;
}

void timerPruning::pruneClassifiers(classifier & ind,
				    agentPerformanceTraining * ap)
{
	if (!enabled || !pruningActivated) {
		return;
	}

	if(ind.getNumClassifiers()<=minClassifiers) return;

	JVector<int> classifiersToPrune;
	int numClassifiers=ind.getNumClassifiers();
	if(tGlobals->defaultClassPolicy!=DISABLED) numClassifiers--;
	int i;

	for (i = 0; i < numClassifiers ; i++) {
		if (!ap->getActivationsOfClassifier(i)) {
			classifiersToPrune.addElement(i);
		}
	}

	if(policy==FTB) frontToBack(ind,classifiersToPrune);
	else if(policy==BTF) backToFront(ind,classifiersToPrune);
	else random(ind,classifiersToPrune);
}

void timerPruning::frontToBack(classifier &ind,JVector<int> &classifiersToPrune)
{
	int numC=ind.getNumClassifiers();
	int selectedClassifiers=classifiersToPrune.size();
	int i;

	if(numC-selectedClassifiers<minClassifiers) {
		selectedClassifiers=numC-minClassifiers;
	}

	int *classifiersPruned=new int[selectedClassifiers];

	for(i=0;i<selectedClassifiers;i++) {
		classifiersPruned[i]=classifiersToPrune[i];
	}
	ind.pruneClassifiers(classifiersPruned, selectedClassifiers);
	delete classifiersPruned;
}

void timerPruning::backToFront(classifier &ind,JVector<int> &classifiersToPrune)
{
	int numC=ind.getNumClassifiers();
	int max=classifiersToPrune.size();
	int selectedClassifiers=max;
	int index;
	int i;

	if(numC-selectedClassifiers<minClassifiers) {
		selectedClassifiers=numC-minClassifiers;
	}

	int offset=max-selectedClassifiers;
	int *classifiersPruned=new int[selectedClassifiers];
	for(index=0,i=offset;i<max;i++,index++) {
		classifiersPruned[index]=classifiersToPrune[i];
	}
	ind.pruneClassifiers(classifiersPruned, selectedClassifiers);
	delete classifiersPruned;
}

void timerPruning::random(classifier &ind,JVector<int> &classifiersToPrune)
{
	int numC=ind.getNumClassifiers();
	int max=classifiersToPrune.size();
	int selectedClassifiers=max;
	int i,j;

	if(numC-selectedClassifiers<minClassifiers) {
		selectedClassifiers=numC-minClassifiers;
		while(selectedClassifiers<max) {
			int pos=rnd(0,max-1);
			classifiersToPrune.removeElementAt(pos);
			max--;
		}
	}

	int *classifiersPruned=new int[selectedClassifiers];
	for(i=0;i<selectedClassifiers;i++) {
		classifiersPruned[i]=classifiersToPrune[i];
	}
	ind.pruneClassifiers(classifiersPruned, selectedClassifiers);
	delete classifiersPruned;
}

