#include "timerRealKR.h"
#include "timerMDL.h"
#include "attributesInfo.h"
#include "random.h"
#include "messageBuffer.h"
#include <math.h>
#include "timerGlobals.h"

extern messageBuffer mb;

extern attributesInfo ai;
extern Random rnd;
extern timerMDL *tMDL;
extern timerGlobals *tGlobals;

timerRealKR::timerRealKR()
{
	if (!cm.thereIsParameter(KR_HYPERRECT) &&
	    !cm.thereIsParameter(KR_INSTANCE_SET)) {
		enabled = 0;
		return;
	}
	enabled = 1;
	int i;

	if (cm.thereIsParameter(KR_HYPERRECT)) {
                if(cm.thereIsParameter(PROB_ONE)) {
                        tGlobals->probOne = cm.getParameter(PROB_ONE);
                } else {
                        int num=ai.getNumAttributesMC();
                        int minR=tGlobals->minClassifiersInit;
                        double nc=ai.getNumClasses();
                        tGlobals->probOne=pow(1-pow(nc,-1.0/minR),1.0/num);
                        if(tGlobals->probOne<0.90) tGlobals->probOne=0.90;
                        mb.printf("Probability of ONE set to %f\n",tGlobals->probOne);
                }

	        int numAtt=ai.getNumAttributesMC();
	        int numExpAtt=(int)cm.getParameter(EXPRESSED_ATT_INIT);
	        if(numExpAtt>numAtt) numExpAtt=numAtt;
	        probIrr= 1-(double)numExpAtt/(double)numAtt;
	        mb.printf("Probability of irrelevant attribute set to %f\n",probIrr);


		attributeSize = new int[ai.getNumAttributes()];
		attributeOffset = new int[ai.getNumAttributes()];

		ruleSize = 0;
		for (i = 0; i < ai.getNumAttributesMC(); i++) {
			if (ai.getTypeOfAttribute(i) == REAL) {
				attributeSize[i] = 2;
			} else {
				attributeSize[i] = ai.getNumValuesAttribute(i);
			}
			attributeOffset[i] = ruleSize;
			ruleSize += attributeSize[i];
		}
		attributeSize[i] = 1;
		attributeOffset[i] = ruleSize;
		ruleSize++;

		defaultRuleTheoryLength=0.001;
			//2*ai.getNumAttributesMC()+tMDL->classLength;
	} else {
		instanceTheoryLength = 10;

		/*for (i = 0; i < ai.getNumAttributesMC(); i++) {
			if (ai.getTypeOfAttribute(i) == REAL) {
				instanceTheoryLength += sizeof(double) * 8;
			} else {
				instanceTheoryLength +=
				    1;
			}
		}
		instanceTheoryLength += ai.getNumAttributesMC()/2.0;*/
	}


	thereIsSpecialCrossover = 0;
	if (cm.thereIsParameter(ALPHA_OF_BLX)) {
		alphaOfBLX = cm.getParameter(ALPHA_OF_BLX);
	} else
		alphaOfBLX = -1;

	if (cm.thereIsParameter(D_OF_FR)) {
		dOfFR = cm.getParameter(D_OF_FR);
	} else
		dOfFR = -1;

	if (cm.thereIsParameter(N_OF_SBX)) {
		nOfSBX = cm.getParameter(N_OF_SBX);
	} else
		nOfSBX = -1;

	if (alphaOfBLX != -1 || dOfFR != -1 || nOfSBX != -1)
		thereIsSpecialCrossover = 1;
}

void timerRealKR::dumpStats(int iteration)
{
	if (!enabled)
		return;
}

void timerRealKR::specialCrossover(double parent1, double parent2,
				   double &son1, double &son2,
				   double minDomain, double maxDomain)
{
	if (alphaOfBLX != -1)
		crossoverBLX(parent1, parent2, son1, son2, minDomain,
			     maxDomain);
	else if (dOfFR != -1)
		crossoverFR(parent1, parent2, son1, son2, minDomain,
			    maxDomain);
	else
		crossoverSBX(parent1, parent2, son1, son2, minDomain,
			     maxDomain);
}

void timerRealKR::crossoverBLX(double parent1, double parent2,
			       double &son1, double &son2,
			       double minDomain, double maxDomain)
{
	double minPare = fmin(parent1, parent2);
	double maxPare = fmax(parent1, parent2);
	double interval = maxPare - minPare;

	double minInterval = minPare - alphaOfBLX * interval;
	if (minInterval < minDomain)
		minInterval = minDomain;

	double maxInterval = maxPare + alphaOfBLX * interval;
	if (maxInterval > maxDomain)
		maxInterval = maxDomain;

	son1 = !rnd * (maxInterval - minInterval) + minInterval;
	son2 = !rnd * (maxInterval - minInterval) + minInterval;
}

void timerRealKR::crossoverSBX(double parent1, double parent2,
			       double &son1, double &son2,
			       double minDomain, double maxDomain)
{
	double u, beta;

	u = !rnd;
	u *= 0.9999999999;
	if (u <= 0.5) {
		beta = pow(2 * u, 1.0 / (nOfSBX + 1));
	} else {
		beta = pow(1.0 / (2.0 * (1 - u)), 1.0 / (nOfSBX + 1));
	}

	son1 = 0.5 * ((1 + beta) * parent1 + (1 - beta) * parent2);
	if (son1 < minDomain)
		son1 = minDomain;
	if (son1 > maxDomain)
		son1 = maxDomain;

	son2 = 0.5 * ((1 - beta) * parent1 + (1 + beta) * parent2);
	if (son2 < minDomain)
		son2 = minDomain;
	if (son2 > maxDomain)
		son2 = maxDomain;
}

void timerRealKR::crossoverFR(double parent1, double parent2, double &son1,
			      double &son2, double minDomain,
			      double maxDomain)
{
	double minPare = fmin(parent1, parent2);
	double maxPare = fmax(parent1, parent2);
	double interval = maxPare - minPare;

	son1 =
	    crossoverFRp(minPare, maxPare, interval, minDomain, maxDomain);
	son2 =
	    crossoverFRp(minPare, maxPare, interval, minDomain, maxDomain);
}

double timerRealKR::crossoverFRp(double minPare, double maxPare,
				 double interval, double minDomain,
				 double maxDomain)
{
	double centre;
	double valor;

	if (!rnd < 0.5) {
		centre = minPare;
	} else {
		centre = maxPare;
	}

	if (!rnd < 0.5) {
		valor = centre + (!rnd - 1) * dOfFR * interval;
	} else {
		valor = centre + (1 - !rnd) * dOfFR * interval;
	}
	if (valor < minDomain)
		valor = minDomain;
	if (valor > maxDomain)
		valor = maxDomain;

	return valor;
}

double timerRealKR::fmin(double a, double b)
{
	if (a < b)
		return a;
	return b;
}
double timerRealKR::fmax(double a, double b)
{
	if (a > b)
		return a;
	return b;
}
