#include "usdDiscretizer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include "messageBuffer.h"

extern messageBuffer mb;

void usdDiscretizer::results()
{
        double numTotal=0;
        double numAttr=0;
        for(int i=0;i<numAttributes;i++)
        { // Per a tots els atributs
		if(numCutPoints[i]!=-1) {
			numAttr++;
			numTotal+=numCutPoints[i];
		}
        }
        mb.printf("Numero de punts de tall Discretitzacio : %g\n",numTotal/numAttr);
}

usdDiscretizer::usdDiscretizer(int attributes)
{
	numAttributes = attributes;
	cutPoints = new double *[numAttributes];
	numCutPoints = new int[numAttributes];

	int i;
	for (i = 0; i < numAttributes; i++)
		numCutPoints[i] = -1;
}

void usdDiscretizer::setBounds(double *min, double *max)
{
}

void usdDiscretizer::parseConfig(char *config)
{
	FILE *fp = fopen(config, "r");
	if (!fp) {
		fprintf(stderr, "Can't open %s\n", config);
		exit(1);
	}

	char buffer[100];

	fgets(buffer, 99, fp);
	while (!feof(fp)) {
		int attribute, numCP;
		if(sscanf(buffer,"Attribute %d : Cut points : %d"
			,&attribute, &numCP) != 2) {
			fprintf(stderr, "Parse error1 at parseConfig:%s\n",
				buffer);
			exit(1);
		}

		if (attribute < 0 || attribute >= numAttributes) {
			fprintf(stderr,
				"Wrong attribute a parseConfig:%s\n",
				buffer);
			exit(1);
		}

		if (numCutPoints[attribute] != -1) {
			fprintf(stderr,
				"Attribute cut points already defined:%s\n",
				buffer);
			exit(1);
		}

		numCutPoints[attribute] = numCP;
		if (numCP) cutPoints[attribute] = new double[numCP];
		int i;
		for (i = 0; i < numCP; i++) {
			fgets(buffer, 99, fp);
			if (feof(fp)) {
				fprintf(stderr,
					"EOF error a parseConfig\n");
				exit(1);
			}
			int atr, val;
			float point;

			if (sscanf
			    (buffer, "Cut point %d of attribute %d : %f",
			     &val, &atr, &point) != 3) {
				fprintf(stderr,
					"Parse error3 a parseConfig:%s\n",
					buffer);
				exit(1);
			}
			if (atr != attribute || val != i) {
				fprintf(stderr,
					"Cut point def does not match atr %s\n",
					buffer);
				exit(1);
			}

			if (i > 0 && point < cutPoints[attribute][i - 1]) {
				fprintf(stderr,
					"Cut point must be ordered %s\n",
					buffer);
				exit(1);
			}

			cutPoints[attribute][i] = point;
		}

		fgets(buffer, 99, fp);
	}

	fclose(fp);
}

void usdDiscretizer::updateDiscretizer(char *dataFile)
{
	char command[1000];
	sprintf(command, "usd.pl %s > tempConf%d",
		dataFile, getpid());
	system(command);
	sprintf(command, "tempConf%d", getpid());
	parseConfig(command);
	unlink(command);
	results();
}

int usdDiscretizer::discretize(int attribute, double value)
{
	int i;

	if (numCutPoints[attribute] == 0)
		return 0;

	for (i = 0;
	     i < numCutPoints[attribute] && value>=cutPoints[attribute][i]; 
	     i++);

	return i;
}
