#include "windowingGWS.h"
#include "configManagement.h"
#include "attributesInfo.h"
#include "random.h"
#include "messageBuffer.h"
#include "sampling.h"
#include <math.h>

extern messageBuffer mb;

extern attributesInfo ai;
extern configManagement cm;
extern Random rnd;

void windowingGWS::setInstances(instance ** pSet, int pHowMuch)
{
	set = pSet;
	howMuch = pHowMuch;

	initializeSampling();
}

void windowingGWS::initializeSampling()
{
	int i,j;

	numClasses = ai.getNumClasses();
	quota = howMuch / (int)cm.getParameter(WINDOWING_GWS);
	sample = new instance *[quota];
	samplingPools = new Sampling *[numClasses];
	instancesOfClass = new int *[numClasses];

	JVector<int> tempInstancesOfClass[numClasses];

	for (i = 0; i < howMuch; i++) {
		tempInstancesOfClass[set[i]->getClass()].addElement(i);
	}

	for (i = 0; i < numClasses; i++) {
		int num=tempInstancesOfClass[i].size();
		instancesOfClass[i] = new int[num];
		for(j=0;j<num;j++) 
			instancesOfClass[i][j]=tempInstancesOfClass[i][j];
		samplingPools[i] = new Sampling(num);
	}
}

void windowingGWS::newIteration(instance ** &selectedInstances,
				int &numSelected)
{
	int i,j;

	numSelected = 0;

	int currentQuota=quota;

	while(currentQuota>=numClasses) {
		for (i = 0; i < numClasses; i++) {
			int pos=samplingPools[i]->getSample();
			sample[numSelected++]=set[instancesOfClass[i][pos]];
			currentQuota--;
		}
	}

	while(currentQuota) {
		int selectedClass=rnd(0,numClasses-1);
		int pos=samplingPools[selectedClass]->getSample();
		sample[numSelected++]=set[instancesOfClass[selectedClass][pos]];
		currentQuota--;
	}

	selectedInstances=sample;
}
