#include "populationWrapper.h"
#include "windowingILAS.h"
#include "configManagement.h"
#include "attributesInfo.h"
#include "random.h"
#include "classifierFitness.h"
#include "messageBuffer.h"
#include "sampling.h"

extern messageBuffer mb;
extern Random rnd;
extern attributesInfo ai;
extern configManagement cm;
extern int nodeRank;

void windowingILAS::setInstances(instance **pSet,int pHowMuch)
{
	int i;

	set=pSet;
	howMuch=pHowMuch;

	numStrata=(int)cm.getParameter(WINDOWING_ILAS);
	strata = new instance **[numStrata];
	strataSizes = new int[numStrata];
	currentIteration=0;

	reorderInstances();
}

void windowingILAS::reorderInstances()
{
	int i,j,k;
	JVector<instance *> tempStrata[numStrata];
	int nc=ai.getNumClasses();

	Sampling *samplings[nc];
	for(i=0;i<nc;i++) {
		samplings[i] = new Sampling(numStrata);
	}

	int tempCapacity=howMuch/numStrata+(howMuch%numStrata)*2;
	for(i=0;i<numStrata;i++) 
		tempStrata[i].ensureCapacity(tempCapacity);

	for(i=0;i<howMuch;i++) {
		int cls=set[i]->getClass();
		tempStrata[samplings[cls]->getSample()].addElement(set[i]);
	}

	for(i=0;i<numStrata;i++) {
		int size=tempStrata[i].size();
		strataSizes[i]=size;
		strata[i]=new instance *[size];
		for(j=0;j<size;j++) {
			strata[i][j]=tempStrata[i][j];
		}
		tempStrata[i].removeAllElements();
		tempStrata[i].trimToSize();
	}

	for(i=0;i<nc;i++) {
		delete samplings[i];
	}
}

void windowingILAS::newIteration(instance **&selectedInstances,int &numSelected) 
{
	stratum=currentIteration%numStrata;
	numSelected=strataSizes[stratum];
	selectedInstances=strata[stratum];
	//bcopy(strata[stratum],selectedInstances,numSelected*sizeof(instance *));
	currentIteration++;
}

