#include "populationWrapper.h"
#include "windowingVirtualILAS.h"
#include "configManagement.h"
#include "attributesInfo.h"
#include "random.h"
#include "classifierFitness.h"
#include "messageBuffer.h"
#include "sampling.h"

extern messageBuffer mb;
extern Random rnd;
extern attributesInfo ai;
extern configManagement cm;
extern int nodeRank;

void windowingVirtualILAS::setInstances(datasetGenerator *pGen)
{
	gen=pGen;
	howMuch=gen->getNumInstances();

	numStrata=(int)cm.getParameter(WINDOWING_ILAS);
	strata = new unsigned long long *[numStrata];
	strataSizes = new int[numStrata];
	currentIteration=0;

	currentWindow=NULL;

	reorderInstances();
}

void windowingVirtualILAS::reorderInstances()
{
	unsigned long long i,j,k;
	JVector<unsigned long long> tempStrata[numStrata];
	int numClasses=ai.getNumClasses();
	Sampling *samp[numClasses];

	for(i=0;i<numClasses;i++) {
		samp[i] = new Sampling(numStrata);
	}

	int tempCapacity=howMuch/numStrata+(howMuch%numStrata)*2;
	for(i=0;i<numStrata;i++) 
		tempStrata[i].ensureCapacity(tempCapacity);

	for(i=0;i<howMuch;i++) {
		int cls=gen->classOfInstance(i);
		tempStrata[samp[cls]->getSample()].addElement(i);
	}

	for(i=0;i<numStrata;i++) {
		int size=tempStrata[i].size();
		strataSizes[i]=size;
		strata[i]=new unsigned long long[size];
		for(j=0;j<size;j++) {
			strata[i][j]=tempStrata[i][j];
		}
		tempStrata[i].removeAllElements();
		tempStrata[i].trimToSize();
	}
}

void windowingVirtualILAS::newIteration(instance **&selectedInstances,int &numSelected) 
{
	int i;

	if(currentWindow) {
		for(i=0;i<currentSize;i++) {
			delete currentWindow[i];
		}
		delete currentWindow;
	}

	stratum=currentIteration%numStrata;
	numSelected=currentSize=strataSizes[stratum];

	currentWindow = new instance *[currentSize];
	if(ai.allAttributesBinary()) {
		for(i=0;i<currentSize;i++) {
			currentWindow[i]=gen->generateInstance(strata[stratum][i]);
			currentWindow[i]->generatePackedRepresentation();
		}
	} else {
		for(i=0;i<currentSize;i++) {
			currentWindow[i]=gen->generateInstance(strata[stratum][i]);
		}
	}
	selectedInstances=currentWindow;
	currentIteration++;
}

