#include "populationWrapper.h"
#include "windowingVirtualSampling.h"
#include "configManagement.h"
#include "attributesInfo.h"
#include "random.h"
#include "classifierFitness.h"
#include "messageBuffer.h"
#include "sampling.h"

extern messageBuffer mb;
extern Random rnd;
extern attributesInfo ai;
extern configManagement cm;
extern int nodeRank;

void windowingVirtualSampling::setInstances(datasetGenerator *pGen)
{
	gen=pGen;
	howMuch=gen->getNumInstances();

	numStrata=cm.getParameter(WINDOWING_ILAS);
	instancesPerStratum=(int)(howMuch/numStrata);
	mb.printf("Instances per stratum in the Sampling Virtual Windowing : %d\n",instancesPerStratum);
	currentIteration=0;
	stratum=0;

	currentWindow=NULL;
}

void windowingVirtualSampling::newIteration(instance **&selectedInstances,int &numSelected) 
{
	int i;

	if(currentWindow) {
		for(i=0;i<instancesPerStratum;i++) {
			delete currentWindow[i];
		}
	} else {
		currentWindow = new instance *[instancesPerStratum];
	}

	numSelected=instancesPerStratum;

	if(ai.allAttributesBinary()) {
		for(i=0;i<instancesPerStratum;i++) {
			currentWindow[i]=gen->generateRandomInstance();
			currentWindow[i]->generatePackedRepresentation();
		}
	} else {
		for(i=0;i<instancesPerStratum;i++) {
			currentWindow[i]=gen->generateRandomInstance();
		}
	}
	selectedInstances=currentWindow;

	currentIteration++;
	stratum++;
	if(stratum>=numStrata) {
		stratum=0;
	}
}

